// DGPSListener.cpp -- listens to a dBlackBox plus DGPS receiver manufactured
// by GPS Gesellschaft fr professionelle Satellitennavigation GmbH

// Written by Walter Piechulla in April 2000

// Based on: Allen Denver. (December 11, 1995). Serial Communications 
// in Win32. Microsoft Windows Developer Support [The Article and code
// is located on the MSDN CDs shipped with MSVCPP Version 6.0]

// This also uses code from Geodesy Foundation Classes and NMEA0183 Parser written by 
// Sam Blackburn http://ourworld.compuserve.com/homepages/Sam_Blackburn/homepage.htm

#include "stdafx.h"
#include "DGPSListener.h"

// Globals originally in settings.c
DCB dcbTemp;

char* szBaud[] = 
{
  "110",
  "300",
  "600",
  "1200",
  "2400", 
	"4800",
  "9600",
  "14400",
  "19200",
	"38400",
  "56000",
  "57600",
  "115200", 
	"128000",
  "256000"
};

DWORD BaudTable[] =  
{
	CBR_110, 
  CBR_300, 
  CBR_600, 
  CBR_1200, 
  CBR_2400,
	CBR_4800, 
  CBR_9600, 
  CBR_14400, 
  CBR_19200, 
  CBR_38400,
	CBR_56000, 
  CBR_57600, 
  CBR_115200, 
  CBR_128000, 
  CBR_256000
};

char* szParity[] =   
{   
  "None", 
  "Even", 
  "Odd", 
  "Mark", 
  "Space" 
};

DWORD ParityTable[] = 
{  
  NOPARITY, 
  EVENPARITY, 
  ODDPARITY, 
  MARKPARITY, 
  SPACEPARITY  
};

char* szStopBits[] =  
{  
  "1", 
  "1.5", 
  "2"  
};

DWORD StopBitsTable[] =  
{ 
  ONESTOPBIT, 
  ONE5STOPBITS, 
  TWOSTOPBITS 
};

char* szDTRControlStrings[] = 
{ 
  "Enable", 
  "Disable", 
  "Handshake" 
};

DWORD DTRControlTable[] = 
{ 
  DTR_CONTROL_ENABLE, 
  DTR_CONTROL_DISABLE, 
  DTR_CONTROL_HANDSHAKE 
};

char* szRTSControlStrings[] = 
{ 
  "Enable", 
  "Disable", 
  "Handshake", 
  "Toggle" 
};

DWORD RTSControlTable[] = 
{   
  RTS_CONTROL_ENABLE, 
  RTS_CONTROL_DISABLE, 
	RTS_CONTROL_HANDSHAKE, 
  RTS_CONTROL_TOGGLE 
};

DWORD EventFlagsTable[] = 
{
	EV_BREAK, 
  EV_CTS, 
  EV_DSR, 
  EV_ERR, 
  EV_RING,
	EV_RLSD, 
  EV_RXCHAR, 
  EV_RXFLAG, 
  EV_TXEMPTY
};

// TimeoutsDefault
// We need ReadIntervalTimeout here to cause the read operations
// that we do to actually timeout and become overlapped.
// Specifying 1 here causes ReadFile to return very quickly
// so that our reader thread will continue execution.
COMMTIMEOUTS gTimeoutsDefault = { 0x01, 0, 0, 0, 0 };
// End Global originally in init.c

// Enabled sentences. DGPSLister disables all sentences on startup and then immediately 
// enables the sentences set to true in this list. The list conforms with factory settinges. 
static BOOL GPALM_status = FALSE; 
static BOOL GPGGA_status = TRUE;
static BOOL GPGSA_status = FALSE;
static BOOL GPGSV_status = FALSE;
static BOOL GPRMC_status = TRUE;
static BOOL GPVTG_status = FALSE;
static BOOL LCGLL_status = FALSE;
static BOOL LCVTG_status = FALSE;
static BOOL PGRME_status = TRUE;
static BOOL PGRMF_status = FALSE;
static BOOL PGRMT_status = FALSE;
static BOOL PGRMV_status = FALSE;


int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow )
{
  MSG msg;

  if (!VersionCheck()) 
	{
    MessageBox(NULL,"DGPSListener can't run on this version of Windows.",NULL,MB_OK);
    return 0;
  }

  if (!InitializeApp(hInstance,nCmdShow)) 
  {
    MessageBox(NULL,"DGPSListener couldn't start!",NULL,MB_OK);
    return 0;
  }

  while (GetMessage(&msg,NULL,0,0)) 
  {
    if (!TranslateAccelerator(ghwndMain,ghAccel,&msg)) 
    {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
  }

  return 1;
}

BOOL VersionCheck()
{
  gOSV.dwOSVersionInfoSize = sizeof(gOSV);

  if (!GetVersionEx(&gOSV))
    return FALSE;

  if (gOSV.dwPlatformId == VER_PLATFORM_WIN32s)
    return FALSE;

  return TRUE ;
}

BOOL InitializeApp(HINSTANCE hInst,int nCmdShow)
{
  WNDCLASS wc = {0};

  GlobalInitialize();     // get all global variables initialized to defaults

  // setup program's main window class
  wc.lpfnWndProc      = (WNDPROC) MTTTYWndProc;
  wc.hInstance        = hInst;
  wc.hIcon            = LoadIcon(hInst,MAKEINTRESOURCE(IDI_APPICON));
  wc.hCursor          = LoadCursor(NULL, IDC_ARROW);
  wc.lpszMenuName     = MAKEINTRESOURCE(IDR_MTTTYMENU);
  wc.hbrBackground    = (HBRUSH) (COLOR_WINDOW + 1) ;
  wc.lpszClassName    = "MTTTYClass";

  if (!RegisterClass(&wc)) 
  {
    GlobalCleanup();
      return FALSE;
  }

  // setup program's tty child window class
  wc.lpfnWndProc      = (WNDPROC) TTYChildProc;
  wc.hInstance        = hInst;
  wc.hCursor          = LoadCursor(NULL, IDC_IBEAM);
  wc.hbrBackground    = (HBRUSH) (COLOR_WINDOW + 1);
  wc.lpszClassName    = "MTTTYChildClass";
  wc.lpszMenuName     = NULL;
  wc.hIcon            = NULL;

  if (!RegisterClass(&wc)) 
  {
    GlobalCleanup();
    return FALSE;
  }

  // create main window
  ghwndMain = CreateWindow("MTTTYClass", 
                           "DGPSListener",
                           WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
                           STARTXWINDOW, STARTYWINDOW,
                           MAXXWINDOW, MAXYWINDOW,
                           NULL, NULL, hInst, NULL);

  if (ghwndMain == NULL) 
  {
    GlobalCleanup();
    return FALSE;
  }

  ShowWindow(ghwndMain,nCmdShow);   
  UpdateWindow(ghwndMain);

  ghInst = hInst;
  ghAccel = LoadAccelerators(hInst,MAKEINTRESOURCE(IDR_MTTTYACCELERATOR));

  BlackBoxDGPSSSetSentencesMenuChecks(ghwndMain);

  return TRUE;
}

void BlackBoxDGPSSSetSentencesMenuChecks(HWND hwnd)
{
  HMENU hMenu = GetMenu(hwnd);

  GPALM_status ? CheckMenuItem(hMenu,ID_BLACKBOX_GPALM_ENABLED,MF_CHECKED) : 
                 CheckMenuItem(hMenu,ID_BLACKBOX_GPALM_ENABLED,MF_UNCHECKED);
  
  GPGGA_status ? CheckMenuItem(hMenu,ID_BLACKBOX_GPGGA_ENABLED,MF_CHECKED) : 
				 CheckMenuItem(hMenu,ID_BLACKBOX_GPGGA_ENABLED,MF_UNCHECKED);
  
  GPGSA_status ? CheckMenuItem(hMenu,ID_BLACKBOX_GPGSA_ENABLED,MF_CHECKED) : 
				 CheckMenuItem(hMenu,ID_BLACKBOX_GPGSA_ENABLED,MF_UNCHECKED);

  GPGSV_status ? CheckMenuItem(hMenu,ID_BLACKBOX_GPGSV_ENABLED,MF_CHECKED) : 
				 CheckMenuItem(hMenu,ID_BLACKBOX_GPGSV_ENABLED,MF_UNCHECKED);

  GPRMC_status ? CheckMenuItem(hMenu,ID_BLACKBOX_GPRMC_ENABLED,MF_CHECKED) : 
				 CheckMenuItem(hMenu,ID_BLACKBOX_GPRMC_ENABLED,MF_UNCHECKED);

  GPVTG_status ? CheckMenuItem(hMenu,ID_BLACKBOX_GPVTG_ENABLED,MF_CHECKED) : 
				 CheckMenuItem(hMenu,ID_BLACKBOX_GPVTG_ENABLED,MF_UNCHECKED);

  LCGLL_status ? CheckMenuItem(hMenu,ID_BLACKBOX_LCGLL_ENABLED,MF_CHECKED) : 
				 CheckMenuItem(hMenu,ID_BLACKBOX_LCGLL_ENABLED,MF_UNCHECKED);

  LCVTG_status ? CheckMenuItem(hMenu,ID_BLACKBOX_LCVTG_ENABLED,MF_CHECKED) : 
				 CheckMenuItem(hMenu,ID_BLACKBOX_LCVTG_ENABLED,MF_UNCHECKED);

  PGRME_status ? CheckMenuItem(hMenu,ID_BLACKBOX_PGRMF_ENABLED,MF_CHECKED) : 
				 CheckMenuItem(hMenu,ID_BLACKBOX_PGRMF_ENABLED,MF_UNCHECKED);

  PGRMF_status ? CheckMenuItem(hMenu,ID_BLACKBOX_PGRMF_ENABLED,MF_CHECKED) : 
				 CheckMenuItem(hMenu,ID_BLACKBOX_PGRMF_ENABLED,MF_UNCHECKED);

  PGRMT_status ? CheckMenuItem(hMenu,ID_BLACKBOX_PGRMT_ENABLED,MF_CHECKED) : 
				 CheckMenuItem(hMenu,ID_BLACKBOX_PGRMT_ENABLED,MF_UNCHECKED);

  PGRMV_status ? CheckMenuItem(hMenu,ID_BLACKBOX_PGRMV_ENABLED,MF_CHECKED) : 
				 CheckMenuItem(hMenu,ID_BLACKBOX_PGRMV_ENABLED,MF_UNCHECKED);
}

int WINAPI MTTTYWndProc(HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam)
{
  switch (message) 
  {
    case WM_CREATE:
         // since main window is created, I can now open all other windows
         InitTTYInfo();
         OpenTTYChildWindow(hwnd);
         OpenSettingsToolbar(hwnd);
         OpenStatusToolbar(hwnd);
         ChangeConnection(hwnd, CONNECTED(TTYInfo));
         break;

    case WM_DESTROY: 
         // since main windows is being destroyed, so same to other windows
         DestroyTTYInfo();
         DestroyWindow(ghWndToolbarDlg);
         DestroyWindow(ghWndStatusDlg);
         DestroyWindow(ghWndTTY);
				
         GlobalCleanup();
         PostQuitMessage(0);
         break;

    case WM_GETMINMAXINFO:
         {
           // make sure that main window doesn't get smaller than
           // the minimum child windows.
           LPMINMAXINFO lpTemp;
           POINT        ptTemp;

           lpTemp = (LPMINMAXINFO) lParam;
      
           ptTemp.x = (long) lpTemp->ptMinTrackSize.x;
           ptTemp.y = (long) gcyMinimumWindowHeight;

           lpTemp->ptMinTrackSize = ptTemp;
         }
         break;

    case WM_SIZE:
         {
           // main window size has changed,
           // so I need to change the positions of child windows
           WORD wTop;
           WORD wHeight;
           WORD wWidth = LOWORD(lParam);

           // put Settings window at top
           wHeight = SETTINGSFACTOR*gwBaseY;
           wTop = 0;
           MoveWindow(ghWndToolbarDlg,0,wTop,wWidth,wHeight,TRUE);

           // put Status window at bottom
           wHeight = STATUSFACTOR*gwBaseY;
           wTop = HIWORD(lParam) - wHeight;
           MoveWindow(ghWndStatusDlg,0,wTop,wWidth,wHeight,TRUE);

           // put DGPSListener window right in the middle
           // height = whole window - height of two previous windows
           wHeight = HIWORD(lParam) - ((STATUSFACTOR + SETTINGSFACTOR)*gwBaseY);
           wTop = SETTINGSFACTOR*gwBaseY;
           MoveWindow(ghWndTTY,0,wTop,wWidth,wHeight,TRUE);
         }
         break;

    case WM_COMMAND:       
         CmdDispatch(LOWORD(wParam),hwnd,lParam);        
         break;

    case WM_CHAR:
         SetFocus(ghWndTTY);
         SendMessage(ghWndTTY,WM_CHAR,wParam,lParam);
         break;

    case WM_CLOSE:
         if (DisconnectOK()) 
         {
           if (CONNECTED(TTYInfo)) 
           {
             if (TRANSFERRING(TTYInfo))
             {
               TransferFileTextEnd();
             }
               
             BreakDownCommPort();
           }
           
           DestroyWindow(hwnd);
         }
         break;

    default:
         return DefWindowProc(hwnd,message,wParam,lParam);
  }

  return 0L;
}

void CmdDispatch(int iMenuChoice,HWND hwnd,LPARAM lParam)
{
  static char szFileName[MAX_PATH] = {0};

  HMENU hMenu = GetMenu(hwnd);

  switch (iMenuChoice) 
  {
    case ID_HELP_ABOUTMTTTY:
         CmdAbout(hwnd);
         break;

    case ID_TRANSFER_SENDFILETEXT:
         {
           char * szFilter = "Text Files\0*.TXT\0";
           OPENFILENAME ofn = {0};

           ofn.lStructSize = sizeof(OPENFILENAME);
           ofn.hwndOwner = hwnd;
           ofn.lpstrFilter = szFilter;
           ofn.lpstrFile = szFileName;
           ofn.nMaxFile = MAX_PATH;
           ofn.lpstrTitle = "Send File";
           ofn.Flags = OFN_FILEMUSTEXIST;

           if (!GetOpenFileName(&ofn))
             break;
                
           if (TRUE)
             TransferFileTextStart(szFileName);
         }
         break;

    case ID_BLACKBOX_CAPTURERAW: // fallthrough
    case ID_TRANSFER_RECEIVEFILETEXT:
         {
           char * szFilter = "Text Files\0*.TXT\0";
           OPENFILENAME ofn = {0};

           ofn.lStructSize = sizeof(OPENFILENAME);
           ofn.hwndOwner = hwnd;
           ofn.lpstrFilter = szFilter;
           ofn.lpstrFile = szFileName;
           ofn.nMaxFile = MAX_PATH;
           ofn.lpstrTitle = "Receive File";
           ofn.Flags = OFN_OVERWRITEPROMPT;

           if (!GetSaveFileName(&ofn))
             break;

           ReceiveFileText(szFileName,false);
         }
         break;

    case ID_BLACKBOX_CAPTURECOOKED:
         {
           char * szFilter = "Text Files\0*.TXT\0";
           OPENFILENAME ofn = {0};

           ofn.lStructSize = sizeof(OPENFILENAME);
           ofn.hwndOwner = hwnd;
           ofn.lpstrFilter = szFilter;
           ofn.lpstrFile = szFileName;
           ofn.nMaxFile = MAX_PATH;
           ofn.lpstrTitle = "Receive File";
           ofn.Flags = OFN_OVERWRITEPROMPT;

           if (!GetSaveFileName(&ofn))
             break;

           ReceiveFileText(szFileName,true);
         }
         break; 

    case ID_TRANSFER_ABORTSENDING:
         // was abort sent from the abort button?
         if (LOWORD(lParam) == IDC_ABORTBTN) 
         {
           if (REPEATING(TTYInfo))  // am I in a transfer repeat?
           {
             TransferRepeatDestroy();
           } 
           else if (gdwReceiveState == RECEIVE_TTY)  // am I in a normal receive state, then stop sending
           {
             TransferFileTextEnd();
           }
           else  // if I am not in a normal receive state, then stop capturing
           {
             gfAbortTransfer = TRUE;
           }
         }
         else  // if I am not in a normal receive state, then stop capturing
         {
           TransferFileTextEnd();
         }
         break;

    case ID_TRANSFER_SENDREPEATEDLY:
         {
           DWORD dwFreq;
           char * szFilter = "Text Files\0*.TXT\0";
           OPENFILENAME ofn = {0};

           ofn.lStructSize = sizeof(OPENFILENAME);
           ofn.hwndOwner = hwnd;
           ofn.lpstrFilter = szFilter;
           ofn.lpstrFile = szFileName;
           ofn.nMaxFile = MAX_PATH;
           ofn.lpstrTitle = "Send File Repeatedly";
           ofn.Flags = OFN_FILEMUSTEXIST;

           if (!GetOpenFileName(&ofn))
             break;

           dwFreq = GetAFrequency();
                
           TransferRepeatCreate(szFileName,dwFreq);
         }
         break;

    case ID_TRANSFER_ABORTREPEATEDSENDING:
         TransferRepeatDestroy();
         break;

    case ID_TTY_CLEAR:
         ClearTTYContents();
         InvalidateRect(ghWndTTY, NULL, TRUE);
         break;

    // The following correspond to menu choices and buttons in the settings dlog
    case IDC_FONTBTN:
    case IDC_COMMEVENTSBTN:
    case IDC_FLOWCONTROLBTN:
    case IDC_TIMEOUTSBTN:
         SendMessage(ghWndToolbarDlg,WM_COMMAND,(WPARAM)iMenuChoice,(LPARAM)GetDlgItem(ghWndToolbarDlg,iMenuChoice));
         break;

    case ID_BLACKBOX_DUMPCOOKED:
         gdwReceiveState = RECEIVE_TTYBLACKBOXCOOKED;

         if (SetupCommPort() != NULL)
         {
           ChangeConnection(hwnd,CONNECTED(TTYInfo));
         }

         EnableMenuItem(hMenu,ID_BLACKBOX_SETBEGINMARKER,MF_ENABLED);
         EnableMenuItem(hMenu,ID_BLACKBOX_SETENDMARKER,MF_ENABLED);
         break; 
    
    case ID_BLACKBOX_DUMPRAW: 
         gdwReceiveState = RECEIVE_TTY; 

         if (SetupCommPort() != NULL)
         {
           ChangeConnection(hwnd,CONNECTED(TTYInfo));
         }
         
         if (!BlackBoxDGPSSelectSentences(S_IGNORED,DESELECTALL))
         {
           MessageBox(hwnd,"DESELECTALL","Error cannot...",MB_OK);
         }
      
         // Settings according to initial settings (checkmarks) already set in menu  
         if (!BlackBoxDGPSSelectSentences(S_GPGGA,SELECT))
         {
           MessageBox(hwnd,"select GPGGA","Error cannot...",MB_OK);
         }

         if (!BlackBoxDGPSSelectSentences(S_GPRMC,SELECT))
         {
           MessageBox(hwnd,"select GPRMC","Error cannot...",MB_OK);
         }

         if (!BlackBoxDGPSSelectSentences(S_PGRME,SELECT))
         {
           MessageBox(hwnd,"select PGRME","Error cannot...",MB_OK);
         }
         break;

    case ID_FILE_CONNECT:
         gdwReceiveState = RECEIVE_TTY; 

         if (SetupCommPort() != NULL)
         {
           ChangeConnection(hwnd,CONNECTED(TTYInfo));
         }
         break;

    case ID_BLACKBOX_STOPSCREENDUMP:
         EnableMenuItem(hMenu,ID_BLACKBOX_SETBEGINMARKER,MF_DISABLED | MF_GRAYED);
         EnableMenuItem(hMenu,ID_BLACKBOX_SETENDMARKER,MF_DISABLED | MF_GRAYED);
         // fallthrough
          
    case ID_FILE_DISCONNECT:
         if (BreakDownCommPort())
         { 
           ChangeConnection(hwnd,CONNECTED(TTYInfo));
         }
         break;

    case ID_FILE_EXIT:
         PostMessage(hwnd,WM_CLOSE,0,0);
         break;

	case ID_BLACKBOX_GPALM_ENABLED:
		   if (GPALM_status)
       {
         if (!BlackBoxDGPSSelectSentences(S_GPALM,DESELECT))
         {
           MessageBox(hwnd,"deselect GPALM","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_UNCHECKED);
         }
       }
       else
       {
         if (!BlackBoxDGPSSelectSentences(S_GPALM,SELECT))
         {
           MessageBox(hwnd,"select GPALM","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }

		   GPALM_status = !GPALM_status;
		   break;	

	case ID_BLACKBOX_GPGGA_ENABLED:
		   if (GPGGA_status)
       {
         if (!BlackBoxDGPSSelectSentences(S_GPGGA,DESELECT))
         {
           MessageBox(hwnd,"deselect GPGGA","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_UNCHECKED);
         }
       }
       else
       {
         if (!BlackBoxDGPSSelectSentences(S_GPGGA,SELECT))
         {
           MessageBox(hwnd,"select GPGGA","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }

       GPGGA_status = !GPGGA_status; 
		   break;

	case ID_BLACKBOX_GPGSA_ENABLED:
		   if (GPGSA_status)
       {
         if (!BlackBoxDGPSSelectSentences(S_GPGSA,DESELECT))
         {
           MessageBox(hwnd,"deselect GPGSA","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_UNCHECKED);
         }
       }
       else
       {
         if (!BlackBoxDGPSSelectSentences(S_GPGSA,SELECT))
         {
           MessageBox(hwnd,"select GPGSA","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }

  		 GPGSA_status = !GPGSA_status;
	  	 break;

	case ID_BLACKBOX_GPGSV_ENABLED:
		   if (GPGSV_status)
       {
         if (!BlackBoxDGPSSelectSentences(S_GPGSV,DESELECT))
         {
           MessageBox(hwnd,"deselect GPGSV","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_UNCHECKED);
         }
       }
       else
       {
         if (!BlackBoxDGPSSelectSentences(S_GPGSV,SELECT))
         {
           MessageBox(hwnd,"select GPGSV","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }

       GPGSV_status = !GPGSV_status; 
		   break;

	case ID_BLACKBOX_GPRMC_ENABLED:
		   if (GPRMC_status)
       {
         if (!BlackBoxDGPSSelectSentences(S_GPRMC,DESELECT))
         {
           MessageBox(hwnd,"deselect GPRMC","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_UNCHECKED);
         } 
       }
       else
       {
         if (!BlackBoxDGPSSelectSentences(S_GPRMC,SELECT))
         {
           MessageBox(hwnd,"select GPRMC","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }

       GPRMC_status = !GPRMC_status;
		   break;

	case ID_BLACKBOX_GPVTG_ENABLED:
		   if (GPVTG_status)
       {
         if (!BlackBoxDGPSSelectSentences(S_GPVTG,DESELECT))
         {
           MessageBox(hwnd,"deselect GPVTG","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_UNCHECKED);
         }
       }
       else
       {
         if (!BlackBoxDGPSSelectSentences(S_GPVTG,SELECT))
         {
           MessageBox(hwnd,"select GPVTG","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }

       GPVTG_status = !GPVTG_status;
		   break;

	case ID_BLACKBOX_LCGLL_ENABLED:
		   if (LCGLL_status)
       {
         if (!BlackBoxDGPSSelectSentences(S_LCGLL,DESELECT))
         {
           MessageBox(hwnd,"deselect LCGLL","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_UNCHECKED);
         }
       }
       else
       {
         if (!BlackBoxDGPSSelectSentences(S_LCGLL,SELECT))
         {
           MessageBox(hwnd,"select LCGLL","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }

       LCGLL_status = !LCGLL_status;
		   break;

	case ID_BLACKBOX_LCVTG_ENABLED:
		   if (LCVTG_status)
       {
         if (!BlackBoxDGPSSelectSentences(S_LCVTG,DESELECT))
         {
           MessageBox(hwnd,"deselect LCVTG","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }
       else
       {
         if (!BlackBoxDGPSSelectSentences(S_LCVTG,SELECT))
         {
           MessageBox(hwnd,"select LCVTG","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }

       LCVTG_status = !LCVTG_status;
		   break;

	case ID_BLACKBOX_PGRME_ENABLED:
		   if (PGRME_status)
       {
         if (!BlackBoxDGPSSelectSentences(S_PGRME,DESELECT))
         {
           MessageBox(hwnd,"deselect PGRME","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_UNCHECKED);
         }         
       }  
       else
       {
         if (!BlackBoxDGPSSelectSentences(S_PGRME,SELECT))
         {
           MessageBox(hwnd,"select PGRME","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }  

       PGRME_status = !PGRME_status;
		   break;

	case ID_BLACKBOX_PGRMF_ENABLED:
		   if (PGRMF_status)
       {
         if (!BlackBoxDGPSSelectSentences(S_PGRMF,DESELECT))
         {
           MessageBox(hwnd,"deselect PGRMF","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_UNCHECKED);
         }         
       }  
       else
       {
         if (!BlackBoxDGPSSelectSentences(S_PGRMF,SELECT))
         {
           MessageBox(hwnd,"select PGRMF","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }  
         
       PGRMF_status = !PGRMF_status;
		   break;

	case ID_BLACKBOX_PGRMT_ENABLED:
		   if (PGRMT_status)
       {
         if (!BlackBoxDGPSSelectSentences(S_PGRMT,DESELECT))
         {
           MessageBox(hwnd,"deselect PGRMT","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_UNCHECKED);
         }         
       }  
       else
       {
         if (!BlackBoxDGPSSelectSentences(S_PGRMT,SELECT))
         {
           MessageBox(hwnd,"select PGRMT","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }  

       PGRMT_status = !PGRMT_status;
		   break;

	case ID_BLACKBOX_PGRMV_ENABLED:
		   if (PGRMV_status)
       {
         if (!BlackBoxDGPSSelectSentences(S_PGRMV,DESELECT))
         {
           MessageBox(hwnd,"deselect PGRMV","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_UNCHECKED);
         }         
       }  
       else
       {
         if (!BlackBoxDGPSSelectSentences(S_PGRMV,SELECT))
         {
           MessageBox(hwnd,"select PGRMV","Error cannot...",MB_OK);
         }
         else
         {
           CheckMenuItem(hMenu,iMenuChoice,MF_CHECKED);
         }
       }  
       
       PGRMV_status = !PGRMV_status;
		   break;

  case ID_BLACKBOX_ENABLE_ALL:
       if (!BlackBoxDGPSSelectSentences(S_IGNORED,SELECTALL))
       {
         MessageBox(hwnd,"select all","Error cannot...",MB_OK);
       }
       else
       {
         GPALM_status = GPGGA_status = GPGSA_status = GPGSV_status = \
         GPRMC_status = GPVTG_status = LCGLL_status = LCVTG_status = \
         PGRME_status = PGRMF_status = PGRMT_status = PGRMV_status = TRUE;

         BlackBoxDGPSSSetSentencesMenuChecks(hwnd);
       }         
       break;

  case ID_BLACKBOX_DISABLE_ALL:
       if (!BlackBoxDGPSSelectSentences(S_IGNORED,DESELECTALL))
       {
         MessageBox(hwnd,"deselect all","Error cannot...",MB_OK);
       }
       else
       {
         GPALM_status = GPGGA_status = GPGSA_status = GPGSV_status = \
         GPRMC_status = GPVTG_status = LCGLL_status = LCVTG_status = \
         PGRME_status = PGRMF_status = PGRMT_status = PGRMV_status = FALSE;

         BlackBoxDGPSSSetSentencesMenuChecks(hwnd);
       }         
       break;

  case ID_BLACKBOX_SETBEGINMARKER:
       BeginMarkerWaitingToBeSet = TRUE;
       break; 

  case ID_BLACKBOX_SETENDMARKER:
       EndMarkerWaitingToBeSet = TRUE;
       break; 
  }
  
  return;
}

void OpenTTYChildWindow(HWND hWnd)
{
  ghWndTTY = CreateWindow("MTTTYChildClass",
                          "TTY Window", 
                          WS_CHILD | WS_VISIBLE | WS_VSCROLL,
                          0,0,
                          0,0,
                          hWnd, 
                          (HMENU)ID_TTYWINDOW, 
                          ghInst, 
                          NULL);

  if (ghWndTTY == NULL)
  {
    ErrorReporter("Can't Create TTY Child Window");
  }

  return;
}

BOOL NEAR ScrollTTYVert(HWND hWnd,WORD wScrollCmd,WORD wScrollPos)
{
  int  nScrollAmt ;

  switch (wScrollCmd)
  {
    case SB_TOP:
         nScrollAmt = -YOFFSET(TTYInfo);
         break;

    case SB_BOTTOM:
         nScrollAmt = YSCROLL(TTYInfo) - YOFFSET(TTYInfo);
         break ;

    case SB_PAGEUP:
         nScrollAmt = -YSIZE(TTYInfo);
         break ;

    case SB_PAGEDOWN:
         nScrollAmt = YSIZE(TTYInfo);
         break ;

    case SB_LINEUP:
         nScrollAmt = -YCHAR(TTYInfo);
         break ;

    case SB_LINEDOWN:
         nScrollAmt = YCHAR(TTYInfo);
         break ;

    case SB_THUMBPOSITION:
         nScrollAmt = wScrollPos - YOFFSET(TTYInfo);
         break ;

    default:
         return (FALSE);
  }

  if ((YOFFSET(TTYInfo) + nScrollAmt) > YSCROLL(TTYInfo))
      nScrollAmt = YSCROLL(TTYInfo) - YOFFSET(TTYInfo);

  if ((YOFFSET(TTYInfo) + nScrollAmt) < 0)
      nScrollAmt = -YOFFSET(TTYInfo);

  ScrollWindowEx(hWnd,0,-nScrollAmt,NULL,NULL,NULL,NULL,SW_INVALIDATE|SW_ERASE);

  YOFFSET(TTYInfo) = YOFFSET(TTYInfo) + nScrollAmt;

  SetScrollPos(hWnd,SB_VERT,YOFFSET(TTYInfo),TRUE);

  return (TRUE);
}

BOOL NEAR ScrollTTYHorz(HWND hWnd,WORD wScrollCmd,WORD wScrollPos)
{
  int nScrollAmt ;

  switch (wScrollCmd)
  {
    case SB_TOP:
         nScrollAmt = -XOFFSET(TTYInfo);
         break ;

    case SB_BOTTOM:
         nScrollAmt = XSCROLL(TTYInfo) - XOFFSET(TTYInfo);
         break ;

    case SB_PAGEUP:
         nScrollAmt = -XSIZE(TTYInfo);
         break ;

    case SB_PAGEDOWN:
         nScrollAmt = XSIZE(TTYInfo);
         break ;

    case SB_LINEUP:
         nScrollAmt = -XCHAR(TTYInfo);
         break ;

    case SB_LINEDOWN:
         nScrollAmt = XCHAR(TTYInfo);
         break ;

    case SB_THUMBPOSITION:
         nScrollAmt = wScrollPos - XOFFSET(TTYInfo);
         break ;

    default:
         return (FALSE);
  }
  
  if ((XOFFSET(TTYInfo) + nScrollAmt) > XSCROLL(TTYInfo))
  {
    nScrollAmt = XSCROLL(TTYInfo) - XOFFSET(TTYInfo);
  }
  
  if ((XOFFSET(TTYInfo) + nScrollAmt) < 0)
  {
    nScrollAmt = -XOFFSET(TTYInfo);
  }
  
  ScrollWindowEx(hWnd,-nScrollAmt,0,NULL,NULL,NULL,NULL,SW_INVALIDATE|SW_ERASE);
  XOFFSET(TTYInfo) = XOFFSET(TTYInfo) + nScrollAmt;
  SetScrollPos(hWnd,SB_HORZ,XOFFSET(TTYInfo),TRUE);

  return (TRUE);
}

BOOL NEAR PaintTTY(HWND hWnd)
{
  PAINTSTRUCT  ps;
  HFONT        hOldFont;
  RECT         rect;
  HDC          hDC;
  int          nRow, nCol, nEndRow, nEndCol;
  int          nCount, nHorzPos, nVertPos;

  hDC = BeginPaint(hWnd,&ps);

  hOldFont = (HFONT)SelectObject(hDC,HTTYFONT(TTYInfo));
  SetTextColor(hDC,FGCOLOR(TTYInfo));
  SetBkColor(hDC,GetSysColor(COLOR_WINDOW));
  rect = ps.rcPaint;
  nRow = min(MAXROWS - 1,max(0,(rect.top + YOFFSET(TTYInfo)) / YCHAR(TTYInfo)));
  nEndRow = min(MAXROWS - 1,((rect.bottom + YOFFSET(TTYInfo) - 1) / YCHAR(TTYInfo)));
  nCol = min(MAXCOLS - 1,max(0,(rect.left + XOFFSET(TTYInfo)) / XCHAR(TTYInfo)));
  nEndCol = min(MAXCOLS - 1,((rect.right + XOFFSET(TTYInfo) - 1) / XCHAR(TTYInfo)));
  nCount = nEndCol - nCol + 1;

  for (;nRow <= nEndRow;nRow++)
  {
    nVertPos = (nRow * YCHAR(TTYInfo)) - YOFFSET(TTYInfo);
    nHorzPos = (nCol * XCHAR(TTYInfo)) - XOFFSET(TTYInfo);
    rect.top = nVertPos ;
    rect.bottom = nVertPos + YCHAR(TTYInfo);
    rect.left = nHorzPos;
    rect.right = nHorzPos + XCHAR(TTYInfo) * nCount;
    SetBkMode(hDC,OPAQUE);
    ExtTextOut(hDC,nHorzPos,nVertPos,ETO_OPAQUE|ETO_CLIPPED,&rect,
               (LPSTR)(SCREEN(TTYInfo) + nRow * MAXCOLS + nCol),
                nCount,NULL);
  }
  
  SelectObject(hDC,hOldFont);
  EndPaint(hWnd,&ps);
  MoveTTYCursor(hWnd);
  return (TRUE);
}

BOOL NEAR MoveTTYCursor(HWND hWnd)
{
  if (CONNECTED(TTYInfo) && (CURSORSTATE(TTYInfo) & CS_SHOW))
  {
    SetCaretPos((COLUMN(TTYInfo) * XCHAR(TTYInfo)) -
                XOFFSET(TTYInfo),
                (ROW(TTYInfo) * YCHAR(TTYInfo)) -
                YOFFSET(TTYInfo));
  }
   
  return (TRUE);
}

BOOL NEAR SetTTYFocus(HWND hWnd)
{
  if (CONNECTED(TTYInfo) && (CURSORSTATE(TTYInfo) != CS_SHOW))
  {
    CreateCaret(hWnd,NULL,XCHAR(TTYInfo),YCHAR(TTYInfo));
    ShowCaret(hWnd);
    CURSORSTATE(TTYInfo) = CS_SHOW;
  }

  MoveTTYCursor(hWnd);
  return (TRUE);
}

BOOL NEAR KillTTYFocus(HWND hWnd)
{
  if (CURSORSTATE(TTYInfo) != CS_HIDE)
  {
    HideCaret(hWnd);
    DestroyCaret();
    CURSORSTATE(TTYInfo) = CS_HIDE;
  }
  return (TRUE);
}

BOOL NEAR SizeTTY(HWND hWnd,WORD wWidth,WORD wHeight)
{
  int nScrollAmt ;

  // adjust vert settings
  YSIZE(TTYInfo) = (int)wHeight ;
  YSCROLL(TTYInfo) = max(0,(MAXROWS * YCHAR(TTYInfo)) - YSIZE(TTYInfo));
  nScrollAmt = min(YSCROLL(TTYInfo),YOFFSET(TTYInfo)) - YOFFSET(TTYInfo);
  ScrollWindow(hWnd,0,-nScrollAmt,NULL,NULL);

  YOFFSET(TTYInfo) = YOFFSET(TTYInfo) + nScrollAmt;
  SetScrollPos(hWnd,SB_VERT,YOFFSET(TTYInfo),FALSE);
  SetScrollRange(hWnd,SB_VERT,0,YSCROLL(TTYInfo),TRUE);

  // adjust horz settings
  XSIZE(TTYInfo) = (int)wHeight ;
  XSCROLL(TTYInfo) = max(0,(MAXCOLS * XCHAR(TTYInfo)) - XSIZE(TTYInfo));
  nScrollAmt = min(XSCROLL(TTYInfo), XOFFSET(TTYInfo)) - XOFFSET(TTYInfo);
  ScrollWindow(hWnd,nScrollAmt,0,NULL,NULL);
  XOFFSET(TTYInfo) = XOFFSET(TTYInfo) + nScrollAmt;
  SetScrollRange(hWnd,SB_HORZ,0,XSCROLL(TTYInfo),FALSE);
  SetScrollPos(hWnd,SB_HORZ,XOFFSET(TTYInfo),TRUE);

  InvalidateRect(hWnd,NULL,FALSE);   // redraw entire window

  return (TRUE);
}

int WINAPI TTYChildProc(HWND hWnd,UINT uMessage,WPARAM wParam,LPARAM lParam)
{
  switch(uMessage)
  {
    case WM_VSCROLL:
         ScrollTTYVert(hWnd,LOWORD(wParam),HIWORD(wParam));
         break;

    case WM_HSCROLL:
         ScrollTTYHorz(hWnd,LOWORD(wParam),HIWORD(wParam));
         break;

    case WM_SIZE:
         SizeTTY(hWnd,LOWORD(lParam),HIWORD(lParam));
         break;

    case WM_PAINT:
         PaintTTY(hWnd);
         break;

    case WM_CHAR:
         {
           // keyboard activity in TTY Window 
           if (CONNECTED(TTYInfo)) 
           {
             if (!WriterAddNewNode(WRITE_CHAR,0,(char)wParam,NULL,NULL,NULL))
             {
               return FALSE;
             }

             if (LOCALECHO(TTYInfo))
             {
               OutputABufferToWindow(ghWndTTY,(CHAR *)&wParam,1);
             }
           }
         }
         break;

    case WM_SETFOCUS:
         SetTTYFocus(ghWndTTY);
         break;

    case WM_KILLFOCUS:
         KillTTYFocus(ghWndTTY);
         break;

    case WM_MOUSEACTIVATE:
         // If mouse is clicked in me (the tty child window) then I need to get the focus.
         SetFocus(hWnd);
         return MA_ACTIVATE;
         break;

    default:
         return DefWindowProc(hWnd,uMessage,wParam,lParam);
  }
  
  return 0L;
}

void GlobalInitialize()
{
  int cyMenuHeight, cyCaptionHeight, cyFrameHeight;

  // critical sections in status reporting & node management
  InitializeCriticalSection(&gStatusCritical);
  InitializeCriticalSection(&gcsWriterHeap);
  InitializeCriticalSection(&gcsDataHeap);

  // status message event
  ghStatusMessageEvent = CreateEvent(NULL,FALSE,FALSE,NULL);
  
  if (ghStatusMessageEvent == NULL)
  {
    ErrorReporter("CreateEvent(Status message event)");
  }

  // thread exit event
  ghThreadExitEvent = CreateEvent(NULL,TRUE,FALSE,NULL);
    
  if (ghThreadExitEvent == NULL)
  {
    ErrorReporter("CreateEvent(Thread exit event)");
  }  

  // used in file transfer status bar
  InitCommonControls();
  
//  Comment by Walter: 
//  InitCommonControls() is deprecated
//  Nowadays Microsoft recommends the following:
//  {
//    INITCOMMONCONTROLSEX iccEx;
//    iccEx.dwSize = sizeof(INITCOMMONCONTROLSEX);
//    iccEx.dwICC = ICC_WIN95_CLASSES;
//      
//    InitCommonControlsEx(&iccEx); 
//  }

  // font for status reporting control
  ghFontStatus = CreateStatusEditFont();

  // the following are used for sizing the tty window and dialog windows
  gwBaseY = HIWORD(GetDialogBaseUnits());
  cyMenuHeight = GetSystemMetrics(SM_CYMENU);
  cyCaptionHeight = GetSystemMetrics(SM_CYCAPTION);
  cyFrameHeight = GetSystemMetrics(SM_CYFRAME);
  gcyMinimumWindowHeight = cyMenuHeight + \
                           4 * cyCaptionHeight + \
                           2 * cyFrameHeight +
                           (SETTINGSFACTOR + STATUSFACTOR) * gwBaseY;
  return;
}

void GlobalCleanup()
{
  DeleteCriticalSection(&gStatusCritical);
  DeleteCriticalSection(&gcsWriterHeap);
  DeleteCriticalSection(&gcsDataHeap);
  DeleteObject(ghFontStatus);
  CloseHandle(ghStatusMessageEvent);
  CloseHandle(ghThreadExitEvent);
  HeapDestroy(ghStatusMessageHeap);
  return;
}

BOOL ClearTTYContents()
{
  FillMemory(SCREEN(TTYInfo),MAXCOLS*MAXROWS,' ');
  return TRUE;
}

void InitNewFont(LOGFONT LogFont, COLORREF rgbColor)
{
  TEXTMETRIC tm;
  HDC        hDC;

  // if old one exists, then I should delete it
  if (HTTYFONT(TTYInfo))
  {
    DeleteObject(HTTYFONT(TTYInfo));
  }

  LFTTYFONT(TTYInfo) = LogFont;
  HTTYFONT(TTYInfo) = CreateFontIndirect(&(LFTTYFONT(TTYInfo)));
  FGCOLOR(TTYInfo) = rgbColor;

  hDC = GetDC(ghwndMain);
  SelectObject(hDC,HTTYFONT(TTYInfo));
  GetTextMetrics(hDC,&tm);
  ReleaseDC(ghwndMain,hDC);

  // character width and height
  XCHAR(TTYInfo) = tm.tmAveCharWidth;
  YCHAR(TTYInfo) = tm.tmHeight + tm.tmExternalLeading;

  XOFFSET(TTYInfo) = 0;
  YOFFSET(TTYInfo) = YCHAR(TTYInfo) * ROW(TTYInfo);
}

BOOL InitTTYInfo()
{
  // Initialize TTY info for dBlackBox plus (you can look up the DCB structure in help)
  // Many of these values are also very common for other devices
  COMDEV(TTYInfo)          = NULL;
  CONNECTED(TTYInfo)       = FALSE;
  LOCALECHO(TTYInfo)       = FALSE;
  CURSORSTATE(TTYInfo)     = CS_HIDE;
  PORT(TTYInfo)            = 1;
  BAUDRATE(TTYInfo)        = CBR_4800;
  BYTESIZE(TTYInfo)        = 8;
  PARITY(TTYInfo)          = NOPARITY;
  STOPBITS(TTYInfo)        = ONESTOPBIT;
  AUTOWRAP(TTYInfo)        = TRUE;
  NEWLINE(TTYInfo)         = FALSE;
  XSIZE(TTYInfo)           = 0;
  YSIZE(TTYInfo)           = 0;
  XSCROLL(TTYInfo)         = 0;
  YSCROLL(TTYInfo)         = 0;
  COLUMN(TTYInfo)          = 0;
  ROW(TTYInfo)             = MAXROWS - 1;
  DISPLAYERRORS(TTYInfo)   = TRUE;

  // timeouts
  TIMEOUTSNEW( TTYInfo )   = gTimeoutsDefault;

  // read state and status events
  gdwReceiveState          = RECEIVE_TTY;
  EVENTFLAGS(TTYInfo)      = EVENTFLAGS_DEFAULT;
  FLAGCHAR(TTYInfo)        = FLAGCHAR_DEFAULT;

  // Flow Control Settings
  DTRCONTROL(TTYInfo)      = DTR_CONTROL_DISABLE;
  RTSCONTROL(TTYInfo)      = RTS_CONTROL_DISABLE;
  XONCHAR(TTYInfo)         = ASCII_XON;
  XOFFCHAR(TTYInfo)        = ASCII_XOFF;
  XONLIMIT(TTYInfo)        = 0;
  XOFFLIMIT(TTYInfo)       = 0;
  CTSOUTFLOW(TTYInfo)      = FALSE;
  DSROUTFLOW(TTYInfo)      = FALSE;
  DSRINFLOW(TTYInfo)       = FALSE;
  XONXOFFOUTFLOW(TTYInfo)  = FALSE;
  XONXOFFINFLOW(TTYInfo)   = FALSE;
  TXAFTERXOFFSENT(TTYInfo) = FALSE;

  NOREADING(TTYInfo)       = FALSE;
  NOWRITING(TTYInfo)       = FALSE;
  NOEVENTS(TTYInfo)        = FALSE;
  NOSTATUS(TTYInfo)        = FALSE;
  SHOWTIMEOUTS(TTYInfo)    = FALSE;

  // setup default font information
  LFTTYFONT(TTYInfo).lfHeight =         12 ;
  LFTTYFONT(TTYInfo).lfWidth =          0 ;
  LFTTYFONT(TTYInfo).lfEscapement =     0 ;
  LFTTYFONT(TTYInfo).lfOrientation =    0 ;
  LFTTYFONT(TTYInfo).lfWeight =         0 ;
  LFTTYFONT(TTYInfo).lfItalic =         0 ;
  LFTTYFONT(TTYInfo).lfUnderline =      0 ;
  LFTTYFONT(TTYInfo).lfStrikeOut =      0 ;
  LFTTYFONT(TTYInfo).lfCharSet =        OEM_CHARSET ;
  LFTTYFONT(TTYInfo).lfOutPrecision =   OUT_DEFAULT_PRECIS ;
  LFTTYFONT(TTYInfo).lfClipPrecision =  CLIP_DEFAULT_PRECIS ;
  LFTTYFONT(TTYInfo).lfQuality =        DEFAULT_QUALITY ;
  LFTTYFONT(TTYInfo).lfPitchAndFamily = FIXED_PITCH | FF_MODERN ;
  strcpy(LFTTYFONT(TTYInfo).lfFaceName, "FixedSys" ) ;

  InitNewFont(LFTTYFONT(TTYInfo), RGB(0,0,0));

  ClearTTYContents();

  return (TRUE) ;
}

void DestroyTTYInfo()
{
  DeleteObject(HTTYFONT(TTYInfo));
}

void StartThreads(void)
{
  DWORD dwReadStatId;
  DWORD dwWriterId;

  READSTATTHREAD(TTYInfo) = CreateThread(NULL, 
                                         0,
                                         (LPTHREAD_START_ROUTINE) ReaderAndStatusProc,
                                         (LPVOID) ghWndTTY, 
                                         0, 
                                         &dwReadStatId);

  if (READSTATTHREAD(TTYInfo) == NULL)
  {
    ErrorInComm("CreateThread(Reader/Status)");
  }

  WRITERTHREAD(TTYInfo) = CreateThread(NULL, 
                                       0, 
                                       (LPTHREAD_START_ROUTINE) WriterProc, 
                                       (LPVOID) NULL, 
                                       0, 
                                       &dwWriterId);
                   
  if (WRITERTHREAD(TTYInfo) == NULL)
  {
    ErrorInComm("CreateThread(Writer)");
  }

  return;
}

HANDLE SetupCommPort()
{
  // get tty settings from settings dialog
  UpdateTTYInfo();

  // open communication port handle
  COMDEV(TTYInfo) = CreateFile(gszPort,  
                               GENERIC_READ | GENERIC_WRITE, 
                               0, 
                               0, 
                               OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED,
                               0);

  if (COMDEV(TTYInfo) == INVALID_HANDLE_VALUE) 
  {   
    ErrorReporter("CreateFile");
    return NULL;
  }

  // Save original comm timeouts and set new ones
  if (!GetCommTimeouts( COMDEV(TTYInfo),&(TIMEOUTSORIG(TTYInfo))))
  {
    ErrorReporter("GetCommTimeouts");
  }

  // Set port state
  UpdateConnection();

  // set comm buffer sizes
  SetupComm(COMDEV(TTYInfo),MAX_READ_BUFFER,MAX_WRITE_BUFFER);

  // raise DTR
  if (!EscapeCommFunction(COMDEV(TTYInfo),SETDTR))
  {
    ErrorReporter("EscapeCommFunction(SETDTR)");
  }

  // start threads and set initial thread state to not done
  StartThreads();

  // set overall connect flag
  CONNECTED(TTYInfo) = TRUE ;

  return COMDEV(TTYInfo);
}

DWORD WaitForThreads(DWORD dwTimeout)
{	
  HANDLE hThreads[2];
  DWORD  dwRes;

  hThreads[0] = READSTATTHREAD(TTYInfo);
  hThreads[1] = WRITERTHREAD(TTYInfo);

  // set thread exit event here
  SetEvent(ghThreadExitEvent);

  dwRes = WaitForMultipleObjects(2,hThreads,TRUE,dwTimeout);
  
  switch(dwRes)
  {
    case WAIT_OBJECT_0:
    case WAIT_OBJECT_0 + 1: 
         dwRes = WAIT_OBJECT_0;
         break;

    case WAIT_TIMEOUT:
         if (WaitForSingleObject(READSTATTHREAD(TTYInfo),0) == WAIT_TIMEOUT)
         {
           OutputDebugString("Reader/Status Thread didn't exit.\n\r");
         }

         if (WaitForSingleObject(WRITERTHREAD(TTYInfo),0) == WAIT_TIMEOUT)
         {
           OutputDebugString("Writer Thread didn't exit.\n\r");
         }
         break;

    default:
         ErrorReporter("WaitForMultipleObjects");
         break;
  }

  // reset thread exit event here
  ResetEvent(ghThreadExitEvent);

  return dwRes;
}

BOOL BreakDownCommPort()
{
  if (!CONNECTED(TTYInfo))
  {
    return FALSE;
  }

  CONNECTED(TTYInfo) = FALSE;

  // wait for the threads for a small period
  if (WaitForThreads(20000) != WAIT_OBJECT_0)
  { 
    // if threads haven't exited, then they will interfere with a new 
    // connection. I must abort the entire program.
    ErrorHandler("Error closing port.");
  }

  // lower DTR
  if (!EscapeCommFunction(COMDEV(TTYInfo),CLRDTR))
  {
    ErrorReporter("EscapeCommFunction(CLRDTR)");
  }

  // restore original comm timeouts
  if (!SetCommTimeouts(COMDEV(TTYInfo),&(TIMEOUTSORIG(TTYInfo))))
  {
    ErrorReporter("SetCommTimeouts (Restoration to original)");
  }

  // Purge reads/writes, input buffer and output buffer
  if (!PurgeComm(COMDEV(TTYInfo),PURGE_FLAGS))
  {
    ErrorReporter("PurgeComm");
  }

  CloseHandle(COMDEV(TTYInfo));
  CloseHandle(READSTATTHREAD(TTYInfo));
  CloseHandle(WRITERTHREAD(TTYInfo));

  return TRUE;
}

BOOL DisconnectOK()
{
  if (!CONNECTED(TTYInfo))
  {
    return TRUE;
  }
    
  return ((MessageBox(ghwndMain,"OK to Disconnect?",gszPort,MB_YESNO)) == IDYES);
}

void TransferRepeatCreate(LPCTSTR lpszFileName,DWORD dwFrequency)
{
  HMENU hMenu;
  UINT  MenuFlags ;
  DWORD dwFileSize;
  DWORD dwMaxPackets;
  DWORD dwPacketSize;
  DWORD dwRead;

  // open the file
  hFile = OpenTheFile(lpszFileName);
  
  if (hFile == INVALID_HANDLE_VALUE)
  {
    return;
  }

  // modify transfer menu
  hMenu = GetMenu(ghwndMain);
  MenuFlags = MF_DISABLED | MF_GRAYED;
  EnableMenuItem(hMenu,ID_TRANSFER_SENDFILETEXT,MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_SENDREPEATEDLY,MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_ABORTREPEATEDSENDING,MF_ENABLED);
  EnableMenuItem(hMenu,ID_TRANSFER_RECEIVEFILETEXT,MenuFlags);

  // enable abort button and progress bar
  SetWindowText(GetDlgItem(ghWndStatusDlg, IDC_ABORTBTN), "Abort Tx");
  ShowWindow(GetDlgItem(ghWndStatusDlg, IDC_ABORTBTN), SW_SHOW);

  if (!GetTransferSizes(hFile,&dwPacketSize,&dwMaxPackets,&dwFileSize)) 
  {
    TransferRepeatDestroy();
    return;
  }

  // Allocate a buffer
  lpBuf = (char*)HeapAlloc(ghWriterHeap,0,dwFileSize);

  if (lpBuf == NULL) 
  {
    ErrorReporter("HeapAlloc (data block from writer heap).\r\nFile is too large");
    TransferRepeatDestroy();
    return;
  }

  // fill the buffer
  if (!ReadFile(hFile,lpBuf,dwFileSize,&dwRead,NULL)) 
  {
    ErrorReporter("Can't read from file\n");
    TransferRepeatDestroy();
  }

  if (dwRead != dwFileSize)
  {
    ErrorReporter("Didn't read entire file\n");
  }
        
  mmTimer = timeSetEvent((UINT)dwFrequency,10,TransferRepeatDo,dwRead,TIME_PERIODIC);
    
  if (mmTimer == (MMRESULT) NULL) 
  {
    ErrorReporter("Could not create mm timer");
    TransferRepeatDestroy();
  }
  else 
  {
    REPEATING(TTYInfo) = TRUE;
    OutputDebugString("Timer setup.\n");
  }

  return;
}

void TransferRepeatDestroy()
{
  HMENU hMenu;
  DWORD MenuFlags;
  MMRESULT mmRes;

  if (mmTimer != (MMRESULT) NULL) 
  {
    mmRes = timeKillEvent(mmTimer);
    
    if (mmRes != TIMERR_NOERROR)
    {
      ErrorReporter("Can't kill mm timer");
    }
    
    mmTimer = (MMRESULT)NULL;
  }

  // close the file
  CloseHandle(hFile);

  // inform writer to abort all pending write requests
  if (!WriterAddFirstNodeTimeout(WRITE_ABORT,0,0,NULL,NULL,NULL,500))
  {
    ErrorReporter("Couldn't inform writer to abort sending.");
  }

  // free the buffer
  if (!HeapFree(ghWriterHeap, 0, lpBuf))
  {
    ErrorReporter("HeapFree (data block from writer heap)");
  }
    
  REPEATING(TTYInfo) = FALSE;
  OutputDebugString("Repeated transfer destroyed.\r\n");

  // enable transfer menu
  hMenu = GetMenu(ghwndMain);
  MenuFlags = MF_ENABLED;
  EnableMenuItem(hMenu,ID_TRANSFER_ABORTREPEATEDSENDING, MF_DISABLED | MF_GRAYED);
  EnableMenuItem(hMenu,ID_TRANSFER_SENDFILETEXT, MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_SENDREPEATEDLY, MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_RECEIVEFILETEXT, MenuFlags);

  ShowWindow(GetDlgItem(ghWndStatusDlg,IDC_ABORTBTN),SW_HIDE);
    
  return;
}

void CALLBACK TransferRepeatDo(UINT uTimerId,UINT uRes,DWORD dwFileSize,DWORD dwRes1,DWORD dwRes2)
{
  if (!WriterAddNewNodeTimeout(WRITE_BLOCK,dwFileSize,0,lpBuf,0,0,10))
  {
    PostMessage(ghwndMain,WM_COMMAND,ID_TRANSFER_ABORTSENDING,MAKELPARAM(IDC_ABORTBTN,0));
  }    
   
  return;
}

void TransferFileTextStart(LPCTSTR lpstrFileName)
{
  DWORD dwThreadId;
  HMENU hMenu;
  UINT  MenuFlags ;

  // open the file
  hFile = OpenTheFile(lpstrFileName);
  
  if (hFile == INVALID_HANDLE_VALUE)
  {
    return;
  }

  // modify transfer menu
  hMenu = GetMenu(ghwndMain);
  MenuFlags = MF_DISABLED | MF_GRAYED;
  EnableMenuItem(hMenu,ID_TRANSFER_SENDFILETEXT,MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_SENDREPEATEDLY,MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_ABORTSENDING,MF_ENABLED);
  EnableMenuItem(hMenu,ID_TRANSFER_RECEIVEFILETEXT,MenuFlags);
    
  // enable abort button and progress bar
  gfAbortTransfer = FALSE;
  SetWindowText(GetDlgItem(ghWndStatusDlg,IDC_ABORTBTN),"Abort Tx");
  ShowWindow(GetDlgItem(ghWndStatusDlg,IDC_ABORTBTN),SW_SHOW);
  ShowWindow(GetDlgItem(ghWndStatusDlg,IDC_TRANSFERPROGRESS),SW_SHOW);

  // start the transfer thread
  hTransferAbortEvent = CreateEvent(NULL,TRUE,FALSE,NULL);
    
  if (hTransferAbortEvent == NULL)
  {
    ErrorReporter("CreateEvent(Transfer Abort Event)");
  }

  hTransferThread = CreateThread(NULL,0,TransferThreadProc,(LPVOID)hFile,0,&dwThreadId);

  if (hTransferThread == NULL) 
  {
    ErrorReporter("CreateThread (Transfer Thread)");
    TransferFileTextEnd();
  }
  else
  {
    TRANSFERRING(TTYInfo) = TRUE;
  }

  return;
}

void TransferFileTextEnd()
{
  HMENU hMenu;
  UINT MenuFlags ;

  // stop the transfer thread
  SetEvent(hTransferAbortEvent);

  OutputDebugString("Waiting for transfer thread...\n");

  if (WaitForSingleObject(hTransferThread,3000) != WAIT_OBJECT_0) 
  {
    ErrorReporter("TransferThread didn't stop.");
    TerminateThread(hTransferThread, 0);
  }
  else
  {
    OutputDebugString("Transfer thread exited\n");
  }

  CloseHandle(hTransferAbortEvent);
  CloseHandle(hTransferThread);

  TRANSFERRING(TTYInfo) = FALSE;

  // enable transfer menu
  hMenu = GetMenu(ghwndMain);
  MenuFlags = MF_ENABLED;
  EnableMenuItem(hMenu,ID_TRANSFER_SENDFILETEXT,MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_SENDREPEATEDLY,MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_RECEIVEFILETEXT,MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_ABORTSENDING,MF_DISABLED | MF_GRAYED);
    
  // disable abort button and progress bar
  ShowWindow(GetDlgItem(ghWndStatusDlg,IDC_ABORTBTN),SW_HIDE);
  ShowWindow(GetDlgItem(ghWndStatusDlg,IDC_TRANSFERPROGRESS),SW_HIDE);

  // close the file
  CloseHandle(hFile);
}

void ReceiveFileText(LPCTSTR lpstrFileName,BOOL cooked)
{
  HMENU hMenu;
  UINT MenuFlags ;

  // create the file
  ghFileCapture = CreateTheFile(lpstrFileName);
  
  if (ghFileCapture == INVALID_HANDLE_VALUE)
  {
    return;
  }

  // setup transfer, disable file menu
  hMenu = GetMenu(ghwndMain);
  MenuFlags = MF_DISABLED | MF_GRAYED;
  EnableMenuItem(hMenu,ID_FILE_CONNECT,MenuFlags);
  EnableMenuItem(hMenu,ID_FILE_DISCONNECT,MenuFlags);
    
  // disable transfer menu
  EnableMenuItem(hMenu,ID_TRANSFER_SENDFILETEXT,MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_RECEIVEFILETEXT,MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_SENDREPEATEDLY,MenuFlags);

  // disable dBlackBoxPlus menu
  EnableMenuItem(hMenu,ID_BLACKBOX_DUMPRAW,MenuFlags);
  EnableMenuItem(hMenu,ID_BLACKBOX_DUMPCOOKED,MenuFlags);
  EnableMenuItem(hMenu,ID_BLACKBOX_STOPSCREENDUMP,MenuFlags);
  EnableMenuItem(hMenu,ID_BLACKBOX_CAPTURERAW,MenuFlags);
  EnableMenuItem(hMenu,ID_BLACKBOX_CAPTURECOOKED,MenuFlags);
    
  // enable abort button and progress bar
  gfAbortTransfer = FALSE;
  SetWindowText(GetDlgItem(ghWndStatusDlg,IDC_ABORTBTN),"Close Capture");
  ShowWindow(GetDlgItem(ghWndStatusDlg,IDC_ABORTBTN),SW_SHOW);
  ShowWindow(GetDlgItem(ghWndStatusDlg,IDC_TRANSFERPROGRESS),SW_SHOW);

  // send file until done or abort
  CaptureFile(ghFileCapture,GetDlgItem(ghWndStatusDlg,IDC_TRANSFERPROGRESS),cooked);

  // enable menu
  hMenu = GetMenu(ghwndMain);
  MenuFlags = MF_ENABLED;
  ChangeConnection(ghwndMain,CONNECTED(TTYInfo));
    
  // enable transfer menu
  EnableMenuItem(hMenu,ID_TRANSFER_SENDFILETEXT,MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_RECEIVEFILETEXT,MenuFlags);
  EnableMenuItem(hMenu,ID_TRANSFER_SENDREPEATEDLY,MenuFlags);

  // enable/disable dBlackBoxPlus menu
  EnableMenuItem(hMenu,ID_BLACKBOX_DUMPRAW,MF_DISABLED|MF_GRAYED);
  EnableMenuItem(hMenu,ID_BLACKBOX_DUMPCOOKED,MF_DISABLED|MF_GRAYED);
  EnableMenuItem(hMenu,ID_BLACKBOX_STOPSCREENDUMP,MenuFlags);
  EnableMenuItem(hMenu,ID_BLACKBOX_CAPTURERAW,MenuFlags);
  EnableMenuItem(hMenu,ID_BLACKBOX_CAPTURECOOKED,MenuFlags);

  // hide abort button and progress bar
  ShowWindow(GetDlgItem(ghWndStatusDlg,IDC_ABORTBTN),SW_HIDE);
  ShowWindow(GetDlgItem(ghWndStatusDlg,IDC_TRANSFERPROGRESS),SW_HIDE);
    
  gfAbortTransfer = FALSE;

  CloseHandle(ghFileCapture);

  return; // returns when file transfer is complete or aborted
}

HANDLE OpenTheFile(LPCTSTR lpFName)
{
  HANDLE hTemp;

  hTemp = CreateFile(lpFName,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);

  if (hTemp == INVALID_HANDLE_VALUE)
  {
    ErrorReporter("CreateFile");
  }

  return hTemp;
}

HANDLE CreateTheFile(LPCTSTR lpFName)
{
  HANDLE hTemp;

  hTemp = CreateFile(lpFName,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,0,NULL);

  if (hTemp == INVALID_HANDLE_VALUE)
  {
    ErrorReporter("CreateFile");
  }

  return hTemp;
}

BOOL GetTransferSizes(HANDLE hFile,DWORD* pdwDataPacketSize,DWORD* pdwNumPackets,DWORD* pdwFileSize)
{
  BY_HANDLE_FILE_INFORMATION fi;

  if (!GetFileInformationByHandle(hFile,&fi)) 
  {
    ErrorReporter("GetFileInformationByHandle");
    return FALSE;
  }
  else 
  {
    if (fi.nFileSizeHigh) 
    {
      MessageBox(ghwndMain,"File is too large to transfer.","File Transfer Error",MB_OK);
      return FALSE;
    }

    // setup packet size, file size and compute the number of packets
    *pdwDataPacketSize = MAX_WRITE_BUFFER;
    *pdwFileSize = fi.nFileSizeLow;
    *pdwNumPackets = *pdwFileSize / *pdwDataPacketSize;

    if (*pdwNumPackets > 65534) 
    {
      MessageBox(ghwndMain,"File is too large for buffer size.","File Transfer Error",MB_OK);
      return FALSE;
    }
  }

  return TRUE;
}

void ShowTransferStatistics(DWORD dwEnd,DWORD dwStart,DWORD dwBytesTransferred)
{
  char szTemp[100];
  DWORD dwSecs;

  dwSecs = (dwEnd - dwStart) / 1000;

  // display only if dwSecs != 0; if dwSecs == 0, then divide by zero occurs.
  if (dwSecs != 0) 
  {
    wsprintf(szTemp,"Bytes transferred: %d\r\nBytes/Second: %d\r\n",dwBytesTransferred,dwBytesTransferred / dwSecs);
    UpdateStatus(szTemp);
  }

  return;
}

UINT CheckForMessages()
{
  MSG msg;

  if (PeekMessage(&msg,NULL,0,0,PM_REMOVE)) 
  {
    if (msg.message == WM_CLOSE) 
    {
      PostMessage(ghwndMain,WM_CLOSE,0,0);
      return WM_CLOSE;
    }

    if (msg.message == WM_SYSCOMMAND && msg.wParam == SC_CLOSE) 
    {            
      PostMessage(ghwndMain,WM_CLOSE,0,0);
      return WM_CLOSE;
    }

    if (!TranslateAccelerator(ghwndMain,ghAccel,&msg)) 
    {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }

    return TRUE ;
  }
    
  return FALSE ;
}

void CaptureFile(HANDLE hFile,HWND hWndProgress,BOOL cooked)
{
  UINT uMsgResult;
  
  if (cooked)
  {
    gdwReceiveState = RECEIVE_CAPTUREDBLACKBOXCOOKED;
  }
  else
  {
    gdwReceiveState = RECEIVE_CAPTURED;
  }

  while (!gfAbortTransfer) 
  {
    uMsgResult = CheckForMessages();

    // If WM_CLOSE is retrieved, then exit.
    // If no message is retrieved, then sleep a little.
    // If any other message is retrieved, check for another one.
    switch(uMsgResult)
    {
      case WM_CLOSE:  gfAbortTransfer = TRUE; break;
      case FALSE:     Sleep(200);             break;
      case TRUE:                              break;
    }
  }

  gdwReceiveState = RECEIVE_TTY;

  return;
}

DWORD WINAPI TransferThreadProc(LPVOID lpV)
{
  DWORD  dwPacketSize, dwMaxPackets, dwFileSize;
  DWORD  dwStartTime;
  HWND   hWndProgress;
  HANDLE hFileHandle;
  HANDLE hDataHeap;
  BOOL fStarted = TRUE;
  BOOL fAborting = FALSE;

  hFileHandle = (HANDLE) lpV;
  hWndProgress = GetDlgItem(ghWndStatusDlg,IDC_TRANSFERPROGRESS);

  // set up transfer metrics
  if (!GetTransferSizes(hFileHandle,&dwPacketSize,&dwMaxPackets,&dwFileSize))
  {
    fAborting = TRUE;
  }
  else 
  {
    SendMessage(hWndProgress,PBM_SETRANGE,0,MAKELPARAM(0,dwMaxPackets+1));
    SendMessage(hWndProgress,PBM_SETSTEP,(WPARAM)1,0);
    SendMessage(hWndProgress,PBM_SETPOS,0,0);
  }

  // set up transfer heaps
  if (!fAborting) 
  {
    SYSTEM_INFO sysInfo;
    GetSystemInfo(&sysInfo);
    hDataHeap = HeapCreate(0, sysInfo.dwPageSize * 2, sysInfo.dwPageSize * 10);
    
    if (hDataHeap == NULL) 
    {
      ErrorReporter("HeapCreate(Data Heap)");
      fAborting = TRUE;
    }
  }

  // inform writer thread that a file is about to be transferred
  if (!fAborting) 
  {
    if (!WriterAddNewNode(WRITE_FILESTART,dwFileSize,0,NULL,NULL,NULL))
    {
      fAborting = TRUE;
    }
  }

  OutputDebugString("Xfer: About to start sending data\n");

  // Get Transfer Start Time
  dwStartTime = GetTickCount();

  if (WaitForSingleObject(hTransferAbortEvent,0) == WAIT_OBJECT_0)
  {
    fAborting = TRUE;
  }

  while (!fAborting) 
  {
    char * lpDataBuf;
    PWRITEREQUEST pWrite;

    // transfer file, loop until all blocks of file have been read
    lpDataBuf = (char*)HeapAlloc(hDataHeap,0,dwPacketSize);
    pWrite = (WRITEREQUEST*)HeapAlloc(ghWriterHeap,0,sizeof(WRITEREQUEST));
    
    if ((lpDataBuf != NULL) && (pWrite != NULL)) 
    {
      DWORD dwRead;

      // read from file into new buffer
      if (ReadFile(hFileHandle,lpDataBuf,dwPacketSize,&dwRead,NULL)) 
      {
        WriterAddExistingNode(pWrite,WRITE_FILE,dwRead,0,lpDataBuf,hDataHeap,hWndProgress);

        if (dwRead != dwPacketSize) // eof
        {
          break;
        }
      }
    }
    else 
    {
      BOOL fRes;
      // Either the data heap is full, or the writer heap is full.
      // Free any allocated block, wait a little and try again.
      // Waiting lets the writer thread send some blocks and free
      // the data blocks from the data heap and the control
      // blocks from the writer heap.
      if (lpDataBuf) 
      {
        EnterCriticalSection(&gcsDataHeap);
        fRes = HeapFree(hDataHeap,0,lpDataBuf);
        LeaveCriticalSection(&gcsDataHeap);
        
        if (!fRes)
        {
          ErrorReporter("HeapFree(Data block)");
        }
      }

      if (pWrite) 
      {
        EnterCriticalSection(&gcsWriterHeap);
        fRes = HeapFree(ghWriterHeap,0,pWrite);
        LeaveCriticalSection(&gcsWriterHeap);
                
        if (!fRes)
        {
          ErrorReporter("HeapFree(Writer block)");
        }
      }

      OutputDebugString("Xfer: A heap is full.  Waiting...\n");

      // wait a little
      // check for abort during the wait
      if (WaitForSingleObject(hTransferAbortEvent,200) == WAIT_OBJECT_0)
      {
        fAborting = TRUE;
      }
    }

    // has the user aborted?
    if (WaitForSingleObject(hTransferAbortEvent,0) == WAIT_OBJECT_0)
    {
      fAborting = TRUE;
    }
  }

  OutputDebugString("Xfer: Done sending packets.\n");

  if (fAborting) 
  {
    // inform writer that transfer is aborting
    OutputDebugString("Xfer: Sending Abort Packet to writer\n");
    WriterAddFirstNodeTimeout(WRITE_ABORT,dwFileSize,0,NULL,NULL,NULL,500);
  }
  else
  {
    WriterAddNewNodeTimeout(WRITE_FILEEND,dwFileSize,0,NULL,NULL,NULL,500);
  }
        
  {
    // wait til writer thread finishes with all blocks
    HANDLE hEvents[2];
    DWORD dwRes;
    BOOL  fTransferComplete;
        
    hEvents[0] = ghTransferCompleteEvent;
    hEvents[1] = hTransferAbortEvent;

    OutputDebugString("Xfer: Waiting for transfer complete signal from writer\n");
      
    do 
    {
      ResetEvent(hTransferAbortEvent);

      dwRes = WaitForMultipleObjects(2, hEvents, FALSE, INFINITE);
        
      switch(dwRes) 
      {
        case WAIT_OBJECT_0:      
             fTransferComplete = TRUE;   
             OutputDebugString("Transfer complete signal rec'd\n");
             break;

        case WAIT_OBJECT_0 + 1:  
             fAborting = TRUE;           
             OutputDebugString("Transfer abort signal rec'd\n");
             OutputDebugString("Xfer: Sending Abort Packet to writer\n");
               
             if (!WriterAddFirstNodeTimeout(WRITE_ABORT,dwFileSize,0,NULL,NULL,NULL,500))
             {
               ErrorReporter("Can't add abort packet\n");
             } 
             break;

        case WAIT_TIMEOUT:  
             break;
            
        default:
             ErrorReporter("WaitForMultipleObjects(Transfer Complete Event and Transfer Abort Event)");
             fTransferComplete = TRUE;
             break;
      }
    } while (!fTransferComplete);
  }

  OutputDebugString("Xfer: transfer complete\n");

  // report statistics
  if (!fAborting)
  {
    ShowTransferStatistics(GetTickCount(),dwStartTime,dwFileSize);
  }

  // break down metrics
  PostMessage(hWndProgress,PBM_SETPOS,0,0);

  // break down heaps
  if (hDataHeap != NULL) 
  {
    if (!HeapDestroy(hDataHeap))
    {
      ErrorReporter("HeapDestroy(data heap)");
    }
  }

  // If I am done without user intervention, then post the
  // "abort" message myself.  This will cause the main thread to
  // clean up after the file transfer.
  if (!fAborting)
  {
    PostMessage(ghwndMain,WM_COMMAND,ID_TRANSFER_ABORTSENDING,0);
  }

  // exit thread
  return 0;
}

void OpenSettingsToolbar(HWND hWnd)
{
  ghWndToolbarDlg = CreateDialog(ghInst,MAKEINTRESOURCE(IDD_TOOLBARSETTINGS),hWnd,ToolbarProc);

  if (ghWndToolbarDlg == NULL)
  {
	  ErrorReporter("CreateDialog (Toolbar Dialog)");
  }

  return;
}

void ChangeConnection(HWND hwnd,BOOL fConnected)
{
  HMENU hMenu;
  int i;

  if (fConnected) 
  {
	  // The port is connected.  Need to :
	  //  Disable connect menu and enable disconnect menu.
	  //  Enable file transfer menu
	  //  Disable comm port selection box
	  //  Disable no writing, no reading, no events, no status check boxes
	  //  Enable status check boxes
	  //  Set focus to the child tty window
	  hMenu = GetMenu(hwnd);
	  EnableMenuItem(hMenu,ID_FILE_CONNECT,MF_GRAYED|MF_DISABLED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_FILE_DISCONNECT,MF_ENABLED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_TRANSFER_SENDFILETEXT,MF_ENABLED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_TRANSFER_RECEIVEFILETEXT,MF_ENABLED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_TRANSFER_SENDREPEATEDLY,MF_ENABLED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_TRANSFER_ABORTSENDING,MF_DISABLED|MF_GRAYED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_TRANSFER_ABORTREPEATEDSENDING,MF_DISABLED|MF_GRAYED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_BLACKBOX_DUMPRAW,MF_DISABLED|MF_GRAYED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_BLACKBOX_DUMPCOOKED,MF_DISABLED|MF_GRAYED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_BLACKBOX_STOPSCREENDUMP,MF_ENABLED|MF_BYCOMMAND);
    EnableMenuItem(hMenu,ID_BLACKBOX_CAPTURERAW,MF_ENABLED|MF_BYCOMMAND);
    EnableMenuItem(hMenu,ID_BLACKBOX_CAPTURECOOKED,MF_ENABLED|MF_BYCOMMAND);

	  EnableWindow(GetDlgItem(ghWndToolbarDlg,IDC_PORTCOMBO),FALSE);
	  EnableWindow(GetDlgItem(ghWndToolbarDlg,IDC_NOWRITINGCHK),FALSE);
	  EnableWindow(GetDlgItem(ghWndToolbarDlg,IDC_NOREADINGCHK),FALSE);
	  EnableWindow(GetDlgItem(ghWndToolbarDlg,IDC_NOEVENTSCHK),FALSE);
	  EnableWindow(GetDlgItem(ghWndToolbarDlg,IDC_NOSTATUSCHK),FALSE);

	  for (i = IDC_STATCTS;i <= IDC_STATRLSD;i++)
    {
	    EnableWindow(GetDlgItem(ghWndStatusDlg,i),TRUE);
    }

	  for (i = IDC_CTSHOLDCHK;i <= IDC_RXCHAREDIT;i++)
    {
	    EnableWindow(GetDlgItem(ghWndStatusDlg,i),TRUE);
    }
    
	  SetFocus(ghWndTTY);
  }
  else 
  {
	  // Not connected, do opposite of above.
	  hMenu = GetMenu(hwnd);
	  EnableMenuItem(hMenu,ID_FILE_CONNECT,MF_ENABLED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_FILE_DISCONNECT,MF_GRAYED|MF_DISABLED|MF_BYCOMMAND);

	  EnableMenuItem(hMenu,ID_TRANSFER_SENDFILETEXT,MF_DISABLED|MF_GRAYED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_TRANSFER_RECEIVEFILETEXT,MF_DISABLED|MF_GRAYED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_TRANSFER_SENDREPEATEDLY,MF_DISABLED|MF_GRAYED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_TRANSFER_ABORTSENDING,MF_DISABLED|MF_GRAYED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_TRANSFER_ABORTREPEATEDSENDING,MF_DISABLED|MF_GRAYED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_BLACKBOX_DUMPRAW,MF_ENABLED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_BLACKBOX_DUMPCOOKED,MF_ENABLED|MF_BYCOMMAND);
	  EnableMenuItem(hMenu,ID_BLACKBOX_STOPSCREENDUMP,MF_DISABLED|MF_GRAYED|MF_BYCOMMAND);
    EnableMenuItem(hMenu,ID_BLACKBOX_CAPTURERAW,MF_GRAYED|MF_DISABLED|MF_BYCOMMAND);
    EnableMenuItem(hMenu,ID_BLACKBOX_CAPTURECOOKED,MF_GRAYED|MF_DISABLED|MF_BYCOMMAND);

	  EnableWindow(GetDlgItem(ghWndToolbarDlg,IDC_PORTCOMBO),TRUE);
	  EnableWindow(GetDlgItem(ghWndToolbarDlg,IDC_NOWRITINGCHK),TRUE);
	  EnableWindow(GetDlgItem(ghWndToolbarDlg,IDC_NOREADINGCHK),TRUE);
	  EnableWindow(GetDlgItem(ghWndToolbarDlg,IDC_NOEVENTSCHK),TRUE);
	  EnableWindow(GetDlgItem(ghWndToolbarDlg,IDC_NOSTATUSCHK),TRUE);

	  for (i = IDC_STATCTS;i <= IDC_STATRLSD;i++) 
    {
	    CheckDlgButton(ghWndStatusDlg,i,0);
	    EnableWindow(GetDlgItem(ghWndStatusDlg,i),FALSE);
	  }

	  for (i = IDC_CTSHOLDCHK;i <= IDC_RXCHAREDIT;i++) 
    {
	    if (i != IDC_TXCHAREDIT && i != IDC_RXCHAREDIT)
      {
		    CheckDlgButton(ghWndStatusDlg,i,0);
      }
	    else
      {
		    SetDlgItemInt(ghWndStatusDlg,i,0,FALSE);
      }

	    EnableWindow( GetDlgItem(ghWndStatusDlg, i), FALSE);
	  }

	  SetFocus(ghwndMain);
  }

  return;
}

void UpdateTTYInfo()
{
  // update globals from dialog settings
  GetDlgItemText(ghWndToolbarDlg,IDC_PORTCOMBO,gszPort,sizeof(gszPort));

  BAUDRATE(TTYInfo) = GetdwTTYItem(ghWndToolbarDlg, 
					                         IDC_BAUDCOMBO, 
					                         szBaud, 
					                         BaudTable, 
					                         sizeof(BaudTable)/sizeof(BaudTable[0]));

  PARITY(TTYInfo) = GetbTTYItem(ghWndToolbarDlg, 
				                        IDC_PARITYCOMBO, 
				                        szParity, 
				                        ParityTable, 
				                        sizeof(ParityTable)/sizeof(ParityTable[0]));

  STOPBITS(TTYInfo) = GetbTTYItem(ghWndToolbarDlg, 
				                          IDC_STOPBITSCOMBO, 
				                          szStopBits, 
				                          StopBitsTable, 
				                          sizeof(StopBitsTable)/sizeof(StopBitsTable[0]));

  LOCALECHO(TTYInfo) = IsDlgButtonChecked(ghWndToolbarDlg,IDC_LOCALECHOCHK);
  BYTESIZE(TTYInfo) = GetDlgItemInt(ghWndToolbarDlg,IDC_DATABITSCOMBO,NULL,FALSE);
  NEWLINE(TTYInfo) = IsDlgButtonChecked(ghWndToolbarDlg,IDC_LFBTN);
  AUTOWRAP(TTYInfo) = IsDlgButtonChecked(ghWndToolbarDlg,IDC_AUTOWRAPCHK);
  DISPLAYERRORS(TTYInfo) = IsDlgButtonChecked(ghWndToolbarDlg,IDC_DISPLAYERRORSCHK);

  NOREADING(TTYInfo) = IsDlgButtonChecked(ghWndToolbarDlg,IDC_NOREADINGCHK);
  NOWRITING(TTYInfo) = IsDlgButtonChecked(ghWndToolbarDlg,IDC_NOWRITINGCHK);
  NOEVENTS(TTYInfo)  = IsDlgButtonChecked(ghWndToolbarDlg,IDC_NOEVENTSCHK);
  NOSTATUS(TTYInfo)  = IsDlgButtonChecked(ghWndToolbarDlg,IDC_NOSTATUSCHK);

  if (CONNECTED(TTYInfo))      // if connected, then update port state
  {
	  UpdateConnection();
  }

  return;
}

BOOL UpdateConnection()
{
  DCB dcb = {0};
    
  dcb.DCBlength = sizeof(dcb);

  // get current DCB settings
  if (!GetCommState(COMDEV(TTYInfo),&dcb))
  {
	  ErrorReporter("GetCommState");
  }

  // update DCB rate, byte size, parity, and stop bits size
  dcb.BaudRate = BAUDRATE(TTYInfo);
  dcb.ByteSize = BYTESIZE(TTYInfo);
  dcb.Parity   = PARITY(TTYInfo);
  dcb.StopBits = STOPBITS(TTYInfo);

  // update event flags
  if (EVENTFLAGS(TTYInfo) & EV_RXFLAG)
  {
	  dcb.EvtChar = FLAGCHAR(TTYInfo);      
  }
  else
  {
	  dcb.EvtChar = '\0';
  }

  // update flow control settings
  dcb.fDtrControl     = DTRCONTROL(TTYInfo);
  dcb.fRtsControl     = RTSCONTROL(TTYInfo);

  dcb.fOutxCtsFlow    = CTSOUTFLOW(TTYInfo);
  dcb.fOutxDsrFlow    = DSROUTFLOW(TTYInfo);
  dcb.fDsrSensitivity = DSRINFLOW(TTYInfo);
  dcb.fOutX           = XONXOFFOUTFLOW(TTYInfo);
  dcb.fInX            = XONXOFFINFLOW(TTYInfo);
  dcb.fTXContinueOnXoff = TXAFTERXOFFSENT(TTYInfo);
  dcb.XonChar         = XONCHAR(TTYInfo);
  dcb.XoffChar        = XOFFCHAR(TTYInfo);
  dcb.XonLim          = XONLIMIT(TTYInfo);
  dcb.XoffLim         = XOFFLIMIT(TTYInfo);

  // DCB settings not in the user's control
  dcb.fParity = TRUE;

  // set new state
  if (!SetCommState(COMDEV(TTYInfo),&dcb))
  {
	  ErrorReporter("SetCommState");
  }

  // set new timeouts
  if (!SetCommTimeouts(COMDEV(TTYInfo),&(TIMEOUTSNEW(TTYInfo))))
	{
    ErrorReporter("SetCommTimeouts");
  }

  return TRUE;
}

void FillComboBox(HWND hCtrlWnd, 
                  char ** szString,
			            DWORD * npTable, 
                  WORD wTableLen, 
                  DWORD dwCurrentSetting)
{
  WORD wCount, wPosition ;

  for (wCount = 0;wCount < wTableLen;wCount++) 
  {
	  wPosition = LOWORD(SendMessage(hCtrlWnd,CB_ADDSTRING,0,(LPARAM)(LPSTR)szString[wCount]));

	  // use item data to store the actual table value
	  SendMessage(hCtrlWnd,CB_SETITEMDATA,(WPARAM)wPosition,(LPARAM) *(npTable + wCount));

	  // if this is our current setting, select it
	  if (*(npTable + wCount) == dwCurrentSetting)
    {
	    SendMessage(hCtrlWnd,CB_SETCURSEL,(WPARAM)wPosition,0L);
    }
  }
  
  return;
}

void SetComboBox(HWND hCtrlWnd,WORD wTableLen,DWORD dwNewSetting)
{
  WORD wCount, wItemData ;

  for (wCount = 0;wCount < wTableLen;wCount++) 
  {
	  wItemData = LOWORD(SendMessage(hCtrlWnd,CB_GETITEMDATA,(WPARAM)wCount,0L));

	  if (wItemData == dwNewSetting) 
    {
	    SendMessage(hCtrlWnd,CB_SETCURSEL,(WPARAM)wCount,0L);
	    break;
	  }
  }
  
  return;
}

BOOL SettingsDlgInit(HWND hDlg)
{
  char szBuffer[MAXLEN_TEMPSTR],szTemp[MAXLEN_TEMPSTR];
  WORD wCount, wMaxCOM, wPosition;

  wMaxCOM = MAXPORTS;
  strcpy(szTemp,"COM");

  // fill port combo box and make initial selection
  for (wCount = 0;wCount < wMaxCOM;wCount++) 
  {
	  wsprintf(szBuffer,"%s%d",(LPSTR)szTemp,wCount+1);
	  SendDlgItemMessage(hDlg,IDC_PORTCOMBO,CB_ADDSTRING,0,(LPARAM)(LPSTR)szBuffer);
  }

  SendDlgItemMessage(hDlg,IDC_PORTCOMBO,CB_SETCURSEL,(WPARAM)(PORT(TTYInfo)-1),0L);
   
  GetDlgItemText(hDlg,IDC_PORTCOMBO,gszPort,sizeof(gszPort));

  // fill baud combo box and make initial selection
  FillComboBox(GetDlgItem(hDlg,IDC_BAUDCOMBO),
		           szBaud, 
               BaudTable,
		           sizeof(BaudTable) / sizeof(BaudTable[0]),
		           BAUDRATE(TTYInfo));

  // fill data bits combo box and make initial selection
  for (wCount = 5;wCount < 9;wCount++) 
  {
	  wsprintf(szBuffer,"%d",wCount);
	  wPosition = LOWORD(SendDlgItemMessage(hDlg,IDC_DATABITSCOMBO,
						                              CB_ADDSTRING, 0,
						                              (LPARAM)(LPSTR)szBuffer));

	  // if wCount is current selection, tell the combo box
	  if (wCount == BYTESIZE(TTYInfo))
    {
	    SendDlgItemMessage(hDlg,IDC_DATABITSCOMBO,CB_SETCURSEL,(WPARAM)wPosition,0L);
    }
  }

  // fill parity combo box and make initial selection
  FillComboBox(GetDlgItem(hDlg,IDC_PARITYCOMBO),
		           szParity, 
               ParityTable,
		           sizeof(ParityTable) / sizeof(ParityTable[0]),
		           PARITY(TTYInfo));

  // fill stop bits combo box and make initial selection
  FillComboBox(GetDlgItem(hDlg,IDC_STOPBITSCOMBO),
		           szStopBits, 
               StopBitsTable,
		           sizeof(StopBitsTable) / sizeof(StopBitsTable[0]),
		           STOPBITS(TTYInfo));
  
  // set check marks based on TTY data
  CheckDlgButton(hDlg,IDC_LOCALECHOCHK,LOCALECHO(TTYInfo));
  CheckDlgButton(hDlg,IDC_DISPLAYERRORSCHK,DISPLAYERRORS(TTYInfo));
  CheckDlgButton(hDlg,IDC_LFBTN,NEWLINE(TTYInfo));
  CheckDlgButton(hDlg,IDC_AUTOWRAPCHK,AUTOWRAP(TTYInfo));

  CheckDlgButton(hDlg,IDC_NOWRITINGCHK,NOWRITING(TTYInfo));
  CheckDlgButton(hDlg,IDC_NOREADINGCHK,NOREADING(TTYInfo));
  CheckDlgButton(hDlg,IDC_NOSTATUSCHK,NOSTATUS(TTYInfo));
  CheckDlgButton(hDlg,IDC_NOEVENTSCHK,NOEVENTS(TTYInfo));

  return (TRUE);

} // end of SettingsDlgInit()

DWORD GetdwTTYItem(HWND hDlg,int idControl,char ** szString,DWORD * pTable,int iNumItems)
{
  int i;
  char szItem[MAXLEN_TEMPSTR];

  // Get current selection (a string)
  GetDlgItemText(hDlg,idControl,szItem,sizeof(szItem));
    
  // Compare current selection with table to find index of item.
	// If index is found, then return the DWORD item from table.
  for (i = 0;i < iNumItems;i++) 
  {
	  if (strcmp(szString[i],szItem) == 0)
	  {
      return pTable[i];
    }
  }

  return 0;
}

BYTE GetbTTYItem(HWND hDlg, int idControl, char ** szString, DWORD * pTable, int iNumItems)
{
  int i;
  char szItem[MAXLEN_TEMPSTR];

  // Get current selection (a string)
  GetDlgItemText(hDlg,idControl,szItem,sizeof(szItem));
    
  // Compare current selection with table to find index of item.
	// If index is found, then return the BYTE item from table.
   
  for (i = 0;i < iNumItems;i++) 
  {
	  if (strcmp(szString[i],szItem) == 0)
	  {
      return (BYTE)pTable[i];
    }
  }
    
  return 0;
}

BOOL CALLBACK ToolbarProc(HWND hWndDlg,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
  BOOL fRet = FALSE;

  switch(uMsg)
  {
	  case WM_INITDIALOG:     // setup dialog with defaults
	       SettingsDlgInit(hWndDlg);
	       break;

	  case WM_COMMAND: 
	       {
		       switch(LOWORD(wParam))
		       {
		         case IDC_FONTBTN:       // font button pressed
			            {
			              CHOOSEFONT cf = {0};
			              LOGFONT lf;

			              lf = LFTTYFONT(TTYInfo);
			              cf.lStructSize = sizeof(CHOOSEFONT);
			              cf.hwndOwner = hWndDlg;
			              cf.lpLogFont = &lf;
			              cf.rgbColors = FGCOLOR(TTYInfo);
			              cf.Flags = CF_INITTOLOGFONTSTRUCT | CF_SCREENFONTS | CF_FIXEDPITCHONLY | \
				            CF_EFFECTS;

			              if (!ChooseFont(&cf))
                    {
				              break;
                    }

			              InitNewFont(lf,cf.rgbColors);

			              // fix scroll bar sizes since we may have more or less pixels per
			              // character now
			              SizeTTY(ghWndTTY,(WORD)XSIZE(TTYInfo),(WORD)YSIZE(TTYInfo));
			    
			              // repaint screen contents
			              InvalidateRect(ghWndTTY,NULL,TRUE);
			    
			              // kill old cursor
			              KillTTYFocus(ghWndTTY);
			    
			              // create new cursor
			              SetTTYFocus(ghWndTTY);
			            }
			
                  fRet = FALSE;
			            break;

		         case IDC_COMMEVENTSBTN:     // comm events button pressed
			            DialogBox(ghInst,MAKEINTRESOURCE(IDD_COMMEVENTSDLG),ghwndMain,CommEventsProc);
			            fRet = FALSE;
			            break;

		         case IDC_FLOWCONTROLBTN:
			            DialogBox(ghInst,MAKEINTRESOURCE(IDD_FLOWCONTROLDLG),ghwndMain,FlowControlProc);
			            fRet = FALSE;
			            break;

		         case IDC_TIMEOUTSBTN:
			            DialogBox(ghInst,MAKEINTRESOURCE(IDD_TIMEOUTSDLG),ghwndMain,TimeoutsProc);
			            fRet = FALSE;
			            break;

		         default:                    // some other control has been modified
			            if (CONNECTED(TTYInfo))
                  {
			              UpdateTTYInfo();
                  }
			            break;
		       }
	       }
	       break;

	  default:
	       break;
  }    

  return fRet;
}

void InitHexControl(HWND hdlg,WORD wIdNumberBox,WORD wIdCharBox,char chData)
{
  char szFlagText[3] = {0};
  char szFlagChar[2] = {0};
    
  // put character into char edit display control
  szFlagChar[0] = chData;
  SetDlgItemText(hdlg,wIdCharBox,szFlagChar);
    
  // put flag character into hex numeric edit control
  wsprintf(szFlagText,"%02x",0x000000FF & chData);
  SetDlgItemText(hdlg,wIdNumberBox,szFlagText);

  return;
}

char GetHexControl(HWND hdlg,WORD wIdNumberBox,WORD wIdCharBox)
{
  UINT uFlagValue;
  char chFlagEntered[3] = {0};
  char chFlag[2] = {0};
    
  // get numeric value from control
  if (0 == GetDlgItemText(hdlg,wIdNumberBox,chFlagEntered,3))
	{
    return 0;
  }

  sscanf(chFlagEntered,"%x",&uFlagValue);

  chFlag[0] = (char)uFlagValue;
  SetDlgItemText(hdlg,wIdCharBox,chFlag); // display character

  return chFlag[0];
}

void InitCommEventsDlg(HWND hdlg, DWORD dwEventFlags)
{
  int i,j;

  for (i = IDC_EVBREAKBTN,j = 0;i <= IDC_EVTXEMPTYBTN;i++,j++)
  {
	  CheckDlgButton(hdlg,i,dwEventFlags & EventFlagsTable[j]);
  }

  InitHexControl(hdlg,IDC_FLAGEDIT,IDC_FLAGCHAR,FLAGCHAR(TTYInfo));

  EnableWindow(GetDlgItem(hdlg,IDC_FLAGEDIT),dwEventFlags & EV_RXFLAG);
  EnableWindow(GetDlgItem(hdlg,IDC_FLAGCHAR),dwEventFlags & EV_RXFLAG);

  return;
}

void SaveCommEventsDlg(HWND hdlg)
{
  int i,j;
  DWORD dwNew = {0};
  char chNewFlag = '\0';
  BOOL fChangingRXFLAG;

  // create a flag out of dialog selections
  for (i = IDC_EVBREAKBTN,j = 0;i <= IDC_EVTXEMPTYBTN;i++,j++) 
  {
	  if (IsDlgButtonChecked(hdlg,i))
    {
	    dwNew |= EventFlagsTable[j];
    }
  }

  // get current flag character from dialog
  chNewFlag = GetHexControl(hdlg,IDC_FLAGEDIT,IDC_FLAGCHAR);
  fChangingRXFLAG = (EVENTFLAGS(TTYInfo) & EV_RXFLAG) != (dwNew & EV_RXFLAG);
  
  if (chNewFlag != FLAGCHAR(TTYInfo) || fChangingRXFLAG) 
  {
	  FLAGCHAR(TTYInfo) = chNewFlag;
	  UpdateTTYInfo();
  }

  // if new flags have been selected, or
  if (dwNew != EVENTFLAGS(TTYInfo))
	{
    EVENTFLAGS(TTYInfo) = dwNew;
  }

  return;
}

BOOL CALLBACK CommEventsProc(HWND hdlg,UINT uMessage,WPARAM wparam,LPARAM lparam)
{
  switch(uMessage)
  {
	  case WM_INITDIALOG:     // init controls
	       InitCommEventsDlg(hdlg,EVENTFLAGS(TTYInfo));
	       break;

	  case WM_COMMAND: 
	       switch(LOWORD(wparam))
	       {
		       case IDOK:
		            SaveCommEventsDlg(hdlg);
		    
		    		    // FALL THROUGH
		    
		       case IDCANCEL:
		            EndDialog(hdlg,LOWORD(wparam));
		            return TRUE;

		       case IDC_DEFAULTSBTN:
		            InitCommEventsDlg(hdlg,EVENTFLAGS_DEFAULT);
		            return TRUE;

		       case IDC_EVRXFLAGBTN:
		            EnableWindow(GetDlgItem(hdlg,IDC_FLAGEDIT),IsDlgButtonChecked(hdlg,IDC_EVRXFLAGBTN));
		            EnableWindow(GetDlgItem(hdlg,IDC_FLAGCHAR),IsDlgButtonChecked(hdlg,IDC_EVRXFLAGBTN));
		            return TRUE;

		       case IDC_FLAGEDIT:
		            GetHexControl(hdlg,IDC_FLAGEDIT,IDC_FLAGCHAR);
		            return TRUE;
	       }
	       break;
  }

  return FALSE;
}

void SaveFlowControlDlg(HWND hdlg)
{
  BOOL    fNewCTSOut, fNewDSROut, fNewDSRIn, fNewXOut, fNewXIn, fNewTXafterXoffSent;
  DWORD   dwNewDTRControl, dwNewRTSControl;
  WORD    wNewXONLimit, wNewXOFFLimit;
  char    chNewXON, chNewXOFF;
  BOOL    fSuccess;

  BOOL fUpdateDCB = FALSE;

  // update DTR and RTS control if needed
  dwNewDTRControl = GetdwTTYItem(hdlg,
                                 IDC_DTRCONTROLCOMBO, 
		                             szDTRControlStrings, 
                                 DTRControlTable, 
				                         sizeof(DTRControlTable)/sizeof(DTRControlTable[0]));

  dwNewRTSControl = GetdwTTYItem(hdlg, 
                                 IDC_RTSCONTROLCOMBO, 
				                         szRTSControlStrings, 
                                 RTSControlTable, 
				                         sizeof(RTSControlTable)/sizeof(RTSControlTable[0]));

  if (dwNewRTSControl != RTSCONTROL(TTYInfo) || dwNewDTRControl != DTRCONTROL(TTYInfo)) 
  {
	  RTSCONTROL(TTYInfo) = dwNewRTSControl;
	  DTRCONTROL(TTYInfo) = dwNewDTRControl;
	  fUpdateDCB = TRUE;
  }

  // update XON/XOFF limits if needed
  wNewXONLimit = GetDlgItemInt(hdlg,IDC_XONLIMITEDIT,&fSuccess,FALSE);
  wNewXOFFLimit = GetDlgItemInt(hdlg,IDC_XOFFLIMITEDIT,&fSuccess,FALSE);
  
  if (wNewXOFFLimit != XOFFLIMIT(TTYInfo) || wNewXONLimit != XONLIMIT(TTYInfo)) 
  {
	  XOFFLIMIT(TTYInfo) = wNewXOFFLimit;
	  XONLIMIT(TTYInfo)  = wNewXONLimit;
	  fUpdateDCB = TRUE;
  }

  // update XON/XOFF chars if needed
  chNewXON = GetHexControl(hdlg,IDC_XONCHAREDIT,IDC_XONCHARDISP);
  chNewXOFF = GetHexControl(hdlg,IDC_XOFFCHAREDIT,IDC_XOFFCHARDISP);
  
  if (chNewXOFF != XOFFCHAR(TTYInfo) || chNewXON != XONCHAR(TTYInfo)) 
  {
	  XOFFCHAR(TTYInfo) = chNewXOFF;
	  XONCHAR(TTYInfo)  = chNewXON;
	  fUpdateDCB = TRUE;
  }

  // update booleans from check boxes
  fNewTXafterXoffSent = IsDlgButtonChecked(hdlg,IDC_TXAFTERXOFFSENTCHK);
  fNewCTSOut = IsDlgButtonChecked(hdlg,IDC_CTSOUTCHK);
  fNewDSROut = IsDlgButtonChecked(hdlg,IDC_DSROUTCHK);
  fNewDSRIn  = IsDlgButtonChecked(hdlg,IDC_DSRINCHK);
  fNewXOut   = IsDlgButtonChecked(hdlg,IDC_XONXOFFOUTCHK);
  fNewXIn    = IsDlgButtonChecked(hdlg,IDC_XONXOFFINCHK);

  if (fNewTXafterXoffSent != TXAFTERXOFFSENT(TTYInfo) || 
	                  fNewCTSOut != CTSOUTFLOW(TTYInfo) ||
	                  fNewDSROut != DSROUTFLOW(TTYInfo) ||
	                  fNewDSRIn  != DSRINFLOW(TTYInfo)  ||
	              fNewXOut   != XONXOFFOUTFLOW(TTYInfo) ||
	               fNewXIn    != XONXOFFINFLOW(TTYInfo)) 
  {
	  CTSOUTFLOW(TTYInfo) = fNewCTSOut;
	  DSROUTFLOW(TTYInfo) = fNewDSROut;
	  DSRINFLOW(TTYInfo)  = fNewDSRIn;
	  XONXOFFOUTFLOW(TTYInfo) = fNewXOut;
	  XONXOFFINFLOW(TTYInfo)  = fNewXIn;
	  TXAFTERXOFFSENT(TTYInfo) = fNewTXafterXoffSent;
	  fUpdateDCB = TRUE;
  }    

  // update current settings if they have actually changed
  if (fUpdateDCB)
  {
	  UpdateTTYInfo();
  }

  return;
}

void InitFlowControlDlg(HWND hdlg)
{
  // fill and init DTR control combo
  FillComboBox(GetDlgItem(hdlg,IDC_DTRCONTROLCOMBO),
		           szDTRControlStrings, DTRControlTable,
		           sizeof(DTRControlTable) / sizeof(DTRControlTable[0]),
		           DTRCONTROL(TTYInfo));

  // fill and init RTS control combo
  FillComboBox(GetDlgItem(hdlg,IDC_RTSCONTROLCOMBO),
	             szRTSControlStrings, 
               RTSControlTable,
	             sizeof(RTSControlTable) / sizeof(RTSControlTable[0]),
	             RTSCONTROL(TTYInfo));

  // XON/XOFF characters
  InitHexControl(hdlg,IDC_XONCHAREDIT,IDC_XONCHARDISP,XONCHAR(TTYInfo));   
  InitHexControl(hdlg,IDC_XOFFCHAREDIT,IDC_XOFFCHARDISP,XOFFCHAR(TTYInfo));    

  // XON/XOFF limits
  SetDlgItemInt(hdlg,IDC_XONLIMITEDIT,XONLIMIT(TTYInfo),FALSE);
  SetDlgItemInt(hdlg,IDC_XOFFLIMITEDIT,XOFFLIMIT(TTYInfo),FALSE);

  // check boxes
  CheckDlgButton(hdlg,IDC_CTSOUTCHK,CTSOUTFLOW(TTYInfo));
  CheckDlgButton(hdlg,IDC_DSROUTCHK,DSROUTFLOW(TTYInfo));
  CheckDlgButton(hdlg,IDC_DSRINCHK,DSRINFLOW(TTYInfo));
  CheckDlgButton(hdlg,IDC_XONXOFFOUTCHK,XONXOFFOUTFLOW(TTYInfo));
  CheckDlgButton(hdlg,IDC_XONXOFFINCHK,XONXOFFINFLOW(TTYInfo));
  CheckDlgButton(hdlg,IDC_TXAFTERXOFFSENTCHK,TXAFTERXOFFSENT(TTYInfo));

  return;
}

void FlowDefault(HWND hdlg,WORD wId)
{
  // set dtr control to handshake if using DTR/DSR flow-control 
  SetComboBox(GetDlgItem(hdlg,IDC_DTRCONTROLCOMBO),
		          sizeof(DTRControlTable) / sizeof(DTRControlTable[0]),
		          wId == IDC_DTRDSRBTN ? DTR_CONTROL_HANDSHAKE : DTR_CONTROL_ENABLE);

  // set rts control to handshake if using RTS/CTS flow-control 
  SetComboBox(GetDlgItem(hdlg,IDC_RTSCONTROLCOMBO),
		          sizeof(RTSControlTable) / sizeof(RTSControlTable[0]),
		          wId == IDC_RTSCTSBTN ? RTS_CONTROL_HANDSHAKE : RTS_CONTROL_ENABLE);

  // set check boxes according to wId
  switch(wId)
  {
	  case IDC_RTSCTSBTN:
	  case IDC_DTRDSRBTN:
	       CheckDlgButton(hdlg,IDC_CTSOUTCHK,wId == IDC_RTSCTSBTN);
	       CheckDlgButton(hdlg,IDC_DSROUTCHK,wId == IDC_DTRDSRBTN);
	       CheckDlgButton(hdlg,IDC_XONXOFFOUTCHK,FALSE);
	       CheckDlgButton(hdlg,IDC_XONXOFFINCHK,FALSE);
	       break;

	  case IDC_XOFFXONBTN:
	  case IDC_NONEBTN:
	       CheckDlgButton(hdlg,IDC_CTSOUTCHK,FALSE);
	       CheckDlgButton(hdlg,IDC_DSROUTCHK,FALSE);
	       CheckDlgButton(hdlg,IDC_XONXOFFOUTCHK,wId == IDC_XOFFXONBTN);
	       CheckDlgButton(hdlg,IDC_XONXOFFINCHK,wId == IDC_XOFFXONBTN);
	       break;
  }

  // settings that are disabled when any 'default' flow-control is used
  CheckDlgButton(hdlg,IDC_DSRINCHK, FALSE);
  CheckDlgButton(hdlg,IDC_TXAFTERXOFFSENTCHK,FALSE);
     
  return;
}

BOOL CALLBACK FlowControlProc(HWND hdlg, UINT uMessage, WPARAM wparam, LPARAM lparam)
{
  switch(uMessage)
  {
	  case WM_INITDIALOG:     // init controls
	       InitFlowControlDlg(hdlg);
	       break;

	  case WM_COMMAND: 
	       switch(LOWORD(wparam))
	       {
		       case IDOK:
			          SaveFlowControlDlg(hdlg);
			          // FALL THROUGH
			          		
		       case IDCANCEL:          
			          EndDialog(hdlg,LOWORD(wparam));          
			          return TRUE;

		       case IDC_RTSCTSBTN:
		       case IDC_DTRDSRBTN:
		       case IDC_XOFFXONBTN:
		       case IDC_NONEBTN:
			          FlowDefault(hdlg,LOWORD(wparam));
			          return TRUE;

		       case IDC_XONCHAREDIT:
			          GetHexControl(hdlg,IDC_XONCHAREDIT,IDC_XONCHARDISP);
			          return TRUE;

		       case IDC_XOFFCHAREDIT:
			          GetHexControl(hdlg,IDC_XOFFCHAREDIT,IDC_XOFFCHARDISP);
			          return TRUE;
	       }
	       
         break;
  }

  return FALSE;
}

void InitTimeoutsDlg(HWND hdlg,COMMTIMEOUTS ct)
{
  SetDlgItemInt(hdlg,IDC_READINTERVALEDIT,ct.ReadIntervalTimeout,FALSE);
  SetDlgItemInt(hdlg,IDC_READMULTIPLIEREDIT,ct.ReadTotalTimeoutMultiplier,FALSE);
  SetDlgItemInt(hdlg,IDC_READCONSTANTEDIT,ct.ReadTotalTimeoutConstant,FALSE);
  SetDlgItemInt(hdlg,IDC_WRITEMULTIPLIEREDIT,ct.WriteTotalTimeoutMultiplier,FALSE);
  SetDlgItemInt(hdlg,IDC_WRITECONSTANTEDIT,ct.WriteTotalTimeoutConstant,FALSE);
  CheckDlgButton(hdlg,IDC_DISPLAYTIMEOUTS,SHOWTIMEOUTS(TTYInfo));
  return;
}

void SaveTimeoutsDlg(HWND hdlg)
{
  COMMTIMEOUTS ctNew;

  // get new timeouts from dialog controls
  ctNew.ReadIntervalTimeout         = GetDlgItemInt(hdlg,IDC_READINTERVALEDIT,NULL,FALSE);
  ctNew.ReadTotalTimeoutMultiplier  = GetDlgItemInt(hdlg,IDC_READMULTIPLIEREDIT,NULL,FALSE);
  ctNew.ReadTotalTimeoutConstant    = GetDlgItemInt(hdlg,IDC_READCONSTANTEDIT,NULL,FALSE);
  ctNew.WriteTotalTimeoutMultiplier = GetDlgItemInt(hdlg,IDC_WRITEMULTIPLIEREDIT,NULL,FALSE);
  ctNew.WriteTotalTimeoutConstant   = GetDlgItemInt(hdlg,IDC_WRITECONSTANTEDIT,NULL,FALSE);

  SHOWTIMEOUTS(TTYInfo) = IsDlgButtonChecked(hdlg,IDC_DISPLAYTIMEOUTS);

  // set new timeouts if they are different
  if (memcmp(&ctNew,&(TIMEOUTSNEW(TTYInfo)),sizeof(COMMTIMEOUTS))) 
  {
	  // if connected, set new time outs and purge pending operations
	  if (CONNECTED(TTYInfo)) 
    {
	    if (!SetCommTimeouts(COMDEV(TTYInfo),&ctNew)) 
      {
		    ErrorReporter("SetCommTimeouts");
		    return;
	    }
	    
	    if (!PurgeComm(COMDEV(TTYInfo),PURGE_TXABORT | PURGE_RXABORT))
      {
		    ErrorReporter("PurgeComm");
      }
	  }

	  // save timeouts in the tty info structure
	  TIMEOUTSNEW(TTYInfo) = ctNew;
  }
    
  return;
}

BOOL CALLBACK TimeoutsProc(HWND hdlg,UINT uMessage,WPARAM wparam,LPARAM lparam)
{
  switch(uMessage)
  {
	  case WM_INITDIALOG:     // init controls
	       InitTimeoutsDlg(hdlg,TIMEOUTSNEW(TTYInfo));
	       break;

	  case WM_COMMAND:
	       switch(LOWORD(wparam))
	       {
		       case IDOK:
			          SaveTimeoutsDlg(hdlg);
			
			          // FALL THROUGH
           case IDCANCEL:
			          EndDialog(hdlg,LOWORD(wparam));          
			          return TRUE;

		       case IDC_DEFAULTSBTN:
			          InitTimeoutsDlg(hdlg,gTimeoutsDefault);
			          return TRUE;
	       }

	       break;
  }

  return FALSE;
}

BOOL CALLBACK GetADWORDProc(HWND hDlg,UINT uMessage,WPARAM wParam,LPARAM lParam)
{
  int iRet = 0;

  if (uMessage == WM_COMMAND) 
  {
	  switch(LOWORD(wParam)) 
    {
	    case IDOK:
		       iRet = GetDlgItemInt(hDlg,IDC_DWORDEDIT,NULL,FALSE);
		       
           // FALL THROUGH
		       	
	    case IDCANCEL:
		       EndDialog(hDlg,iRet);          
		       return TRUE;
	  }
  }

  return FALSE;
}

DWORD GetAFrequency()
{
  return ((DWORD)DialogBox(ghInst,MAKEINTRESOURCE(IDD_GETADWORD),ghwndMain,GetADWORDProc));
}

void OpenStatusToolbar(HWND hWnd)
{
  ghWndStatusDlg = CreateDialog(ghInst,MAKEINTRESOURCE(IDD_STATUSDIALOG),hWnd,StatusDlgProc);

  if (ghWndStatusDlg == NULL)
  {
    ErrorReporter("CreateDialog (Status Dialog)");
  }

  return;
}

HFONT CreateStatusEditFont()
{
  LOGFONT lf = {0};
  HFONT   hFont;

  lf.lfHeight         = 14 ;
  lf.lfCharSet        = ANSI_CHARSET ;
  lf.lfOutPrecision   = OUT_DEFAULT_PRECIS ;
  lf.lfClipPrecision  = CLIP_DEFAULT_PRECIS ;
  lf.lfQuality        = DEFAULT_QUALITY ;
  lf.lfPitchAndFamily = VARIABLE_PITCH | FF_SWISS ;

  hFont = CreateFontIndirect(&lf);
  return hFont;
}

BOOL CALLBACK StatusDlgProc(HWND hWndDlg,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
  BOOL  fRet = FALSE;

  switch(uMsg)
  {
    case WM_INITDIALOG:     // setup dialog with defaults
         SendMessage(GetDlgItem(hWndDlg,IDC_STATUSEDIT),WM_SETFONT,(WPARAM)ghFontStatus,0);
         InitStatusMessage();
         break;

    case WM_COMMAND: 
         {
           switch(LOWORD(wParam))
           {
             case IDC_ABORTBTN:
                  SendMessage(ghwndMain,WM_COMMAND,ID_TRANSFER_ABORTSENDING,MAKELPARAM(IDC_ABORTBTN,0));
                  fRet = TRUE;
                  break;

                  case IDC_STATCTS:
                  case IDC_STATDSR:
                  case IDC_STATRING:
                  case IDC_STATRLSD:
                       CheckModemStatus(TRUE);
                       fRet = TRUE;
                       break;

                  case IDC_CTSHOLDCHK:
                  case IDC_DSRHOLDCHK:
                  case IDC_RLSDHOLDCHK:
                  case IDC_XOFFHOLDCHK:
                  case IDC_XOFFSENTCHK:
                  case IDC_EOFSENTCHK:
                  case IDC_TXIMCHK:
                       CheckComStat(TRUE);
                       fRet = TRUE;
                       break;

                  default:
                       break;
           }
         }
         break;

    default:
         break;
  }    

  return fRet;
}

void InitStatusMessage()
{
  SYSTEM_INFO SysInfo;

  GetSystemInfo(&SysInfo);
  ghStatusMessageHeap = HeapCreate(0,SysInfo.dwPageSize,0);
  
  if (ghStatusMessageHeap == NULL)
  {
    ErrorReporter("HeapCreate(Status message)");
  }

  glpStatusMessageHead = (STATUS_MESSAGE *)HeapAlloc(ghStatusMessageHeap,HEAP_ZERO_MEMORY,sizeof(STATUS_MESSAGE));
  glpStatusMessageTail = glpStatusMessageHead;

  gnStatusIndex = 0;

  // Queue the initial status message.
  // This won't show up until the ReaderAndStatusProc function
  // is called when the threads are created after the port is connected.
  
  UpdateStatus("Status messages go here:\r\n");

  return;
}

void StatusMessage()
{
  STATUS_MESSAGE * lpStatusMessage;
  DWORD dwRes;
  HWND hEdit;
  BOOL bRes;

  hEdit = GetDlgItem (ghWndStatusDlg,IDC_STATUSEDIT);
  
  // If status control has a lot of characters in it,
  // then wipe them out and start over
  
  if (gnStatusIndex > MAX_STATUS_BUFFER) 
  {
    SendMessageTimeout(hEdit, 
                       EM_SETSEL, 
                       (WPARAM)(INT)0, 
                       (LPARAM)(INT)-1,
                       SMTO_NORMAL | SMTO_ABORTIFHUNG, 
                       500, 
                       &dwRes);

    SendMessageTimeout(hEdit, 
                       EM_REPLACESEL, 
                       0,
                       (LPARAM) "",
                       SMTO_NORMAL | SMTO_ABORTIFHUNG, 
                       500, 
                       &dwRes);

    gnStatusIndex = 0;
  }

  // Loop through each node in the status message linked list.
  for (lpStatusMessage = glpStatusMessageHead->lpNext;lpStatusMessage;)
  {
    // if global quit event is set, then just exit this loop
    if (WaitForSingleObject(ghThreadExitEvent, 0) == WAIT_OBJECT_0)
      break;

    // Place each status message into the status control.
    SendMessageTimeout(hEdit, 
                       EM_SETSEL, 
                       gnStatusIndex, 
                       gnStatusIndex, 
                       SMTO_NORMAL | SMTO_ABORTIFHUNG, 
                       500, 
                       &dwRes);

    SendMessageTimeout(hEdit, 
                       EM_REPLACESEL, 
                       0, 
                       (LPARAM)((LPSTR) &(lpStatusMessage->chMessageStart)),
                       SMTO_NORMAL | SMTO_ABORTIFHUNG, 
                       500, 
                       &dwRes);

    gnStatusIndex += strlen(&(lpStatusMessage->chMessageStart));

    // Update status message linked list
    EnterCriticalSection(&gStatusCritical);

    bRes = HeapFree(ghStatusMessageHeap,0,glpStatusMessageHead);
        
    glpStatusMessageHead = lpStatusMessage;
    lpStatusMessage = lpStatusMessage->lpNext;

    LeaveCriticalSection(&gStatusCritical);

    if (!bRes)
    {
      ErrorReporter("HeapFree (status message)");
    }
  }

  return;
}

void UpdateStatus(char * szText)
{
  char * szNewMsg;
  DWORD dwSize;
  STATUS_MESSAGE * lpStatusMessage;
  static dwMessageCounter = 0;

  dwMessageCounter++;

  dwSize = strlen(szText) + 30;    // include NULL terminator and space for counter

  EnterCriticalSection(&gStatusCritical);

  szNewMsg = (char *)HeapAlloc(ghStatusMessageHeap,0,dwSize+30);
    
  if (szNewMsg == NULL) 
  {
    LeaveCriticalSection(&gStatusCritical);
    ErrorReporter("HeapAlloc(status message)");
    return;
  }
  else 
  {
    wsprintf(szNewMsg,"%d:%s",dwMessageCounter,szText);
        
    if (strlen(szNewMsg) > dwSize) 
    {
      LeaveCriticalSection(&gStatusCritical);
      ErrorInComm("Heap Corruption in UpdateStatus\n");
    }
  }

  lpStatusMessage = (STATUS_MESSAGE *)HeapAlloc(ghStatusMessageHeap,0,sizeof(LPSTR)+dwSize);
    
  if (lpStatusMessage == NULL) 
  {
    LeaveCriticalSection(&gStatusCritical);
    ErrorReporter("HeapAlloc (status message)");
    return ;
  }
      
  lpStatusMessage->lpNext = NULL;
  glpStatusMessageTail->lpNext = lpStatusMessage;
  glpStatusMessageTail = lpStatusMessage;
  CopyMemory(&(lpStatusMessage->chMessageStart),szNewMsg,dwSize);
  LeaveCriticalSection(&gStatusCritical);

  SetEvent(ghStatusMessageEvent);

  return ;
}

void ReportModemStatus(DWORD dwModemStatus)
{
  BOOL fCTS, fDSR, fRING, fRLSD;
  BOOL fStat[4];
  int  i;

  fCTS = fStat[0] = MS_CTS_ON & dwModemStatus;
  fDSR = fStat[1] = MS_DSR_ON & dwModemStatus;
  fRING = fStat[2] = MS_RING_ON & dwModemStatus;
  fRLSD = fStat[3] = MS_RLSD_ON & dwModemStatus;

  for (i = IDC_STATCTS;i <= IDC_STATRLSD;i++)
  {
    CheckDlgButton(ghWndStatusDlg,i,fStat[i-IDC_STATCTS] ? 1 : 0);
  }

  return;
}

void CheckModemStatus(BOOL bUpdateNow)
{
  // dwOldStatus needs to be static so that it is maintained 
  // between function calls by the same thread.
  // It also needs to be __declspec(thread) so that it is 
  // initialized when a new thread is created.
  __declspec(thread) static DWORD dwOldStatus = 0;

  DWORD dwNewModemStatus;

  if (!GetCommModemStatus(COMDEV(TTYInfo),&dwNewModemStatus))
  {
    ErrorReporter("GetCommModemStatus");
  }
  
  // Report status if bUpdateNow is true or status has changed
  if (bUpdateNow || (dwNewModemStatus != dwOldStatus)) 
  {
    ReportModemStatus(dwNewModemStatus);
    dwOldStatus = dwNewModemStatus;
  }

  return;
}

void ReportComStat(COMSTAT ComStat)
{
  CheckDlgButton(ghWndStatusDlg,IDC_CTSHOLDCHK,ComStat.fCtsHold);
  CheckDlgButton(ghWndStatusDlg,IDC_DSRHOLDCHK,ComStat.fDsrHold);
  CheckDlgButton(ghWndStatusDlg,IDC_RLSDHOLDCHK,ComStat.fRlsdHold);
  CheckDlgButton(ghWndStatusDlg,IDC_XOFFHOLDCHK,ComStat.fXoffHold);
  CheckDlgButton(ghWndStatusDlg,IDC_XOFFSENTCHK,ComStat.fXoffSent);
  CheckDlgButton(ghWndStatusDlg,IDC_EOFSENTCHK,ComStat.fEof);
  CheckDlgButton(ghWndStatusDlg,IDC_TXIMCHK,ComStat.fTxim);

  SetDlgItemInt(ghWndStatusDlg,IDC_TXCHAREDIT,ComStat.cbOutQue,FALSE);
  SetDlgItemInt(ghWndStatusDlg,IDC_RXCHAREDIT,ComStat.cbInQue,FALSE);

  return;
}

void CheckComStat(BOOL bUpdateNow)
{
  COMSTAT ComStatNew;
  DWORD dwErrors;

  __declspec(thread) static COMSTAT ComStatOld = {0};
  __declspec(thread) static DWORD dwErrorsOld = 0;

  BOOL bReport = bUpdateNow;

  if (!ClearCommError(COMDEV(TTYInfo), &dwErrors, &ComStatNew))
  {
    ErrorReporter("ClearCommError");
  }

  if (dwErrors != dwErrorsOld) 
  {
    bReport = TRUE;
    dwErrorsOld = dwErrors;
  }

  if (memcmp(&ComStatOld, &ComStatNew, sizeof(COMSTAT))) 
  {
    bReport = TRUE;
    ComStatOld = ComStatNew;
  }
    
  if (bReport)
  {
    ReportComStat(ComStatNew);
  }

  ComStatOld = ComStatNew;

  return;
}

void ReportCommError()
{
  COMSTAT comStat;
  DWORD   dwErrors;
  BOOL    fOOP, fOVERRUN, fPTO, fRXOVER, fRXPARITY, fTXFULL;
  BOOL    fBREAK, fDNS, fFRAME, fIOE, fMODE;
  char    szMessage[100];

  // Get and clear current errors on the port
  if (!ClearCommError(COMDEV(TTYInfo),&dwErrors,&comStat))
  {
    ErrorReporter("ClearCommError");
  }

  // get error flags
  fDNS = dwErrors & CE_DNS;
  fIOE = dwErrors & CE_IOE;
  fOOP = dwErrors & CE_OOP;
  fPTO = dwErrors & CE_PTO;
  fMODE = dwErrors & CE_MODE;
  fBREAK = dwErrors & CE_BREAK;
  fFRAME = dwErrors & CE_FRAME;
  fRXOVER = dwErrors & CE_RXOVER;
  fTXFULL = dwErrors & CE_TXFULL;
  fOVERRUN = dwErrors & CE_OVERRUN;
  fRXPARITY = dwErrors & CE_RXPARITY;

  // create error string
  strcpy(szMessage,"ERROR: ");
  strcat(szMessage,fDNS ? "DNS " : "");
  strcat(szMessage,fIOE ? "IOE " : "");
  strcat(szMessage,fOOP ? "OOP " : "");
  strcat(szMessage,fPTO ? "PTO " : "");
  strcat(szMessage,fMODE ? "MODE " : "");
  strcat(szMessage,fBREAK ? "BREAK " : "");
  strcat(szMessage,fFRAME ? "FRAME " : "");
  strcat(szMessage,fRXOVER ? "RXOVER " : "");
  strcat(szMessage,fTXFULL ? "TXFULL " : "");
  strcat(szMessage,fOVERRUN ? "OVERRUN " : "");
  strcat(szMessage,fRXPARITY ? "RXPARITY " : "");

  strcat(szMessage,"\r\n");

  // if there really were errors, then report them
  if (dwErrors)
  {
    UpdateStatus(szMessage);
  }
    
  // Report info from the COMSTAT structure
  ReportComStat(comStat);

  // Show COMSTAT structure with the error indicator
  if (comStat.fCtsHold)
    UpdateStatus("Tx waiting for CTS signal.\r\n");

  if (comStat.fDsrHold)
    UpdateStatus("Tx waiting for DSR signal.\r\n");

  if (comStat.fRlsdHold)
    UpdateStatus("Tx waiting for RLSD signal.\r\n");

  if (comStat.fXoffHold)
    UpdateStatus("Tx waiting, XOFF char rec'd.\r\n");

  if (comStat.fXoffSent)
    UpdateStatus("Tx waiting, XOFF char sent.\r\n");
    
  if (comStat.fEof)
    UpdateStatus("EOF character received.\r\n");
    
  if (comStat.fTxim)
    UpdateStatus("Character waiting for Tx.\r\n");

  if (comStat.cbInQue) 
  {
    wsprintf(szMessage,"%d bytes in input buffer.\r\n",comStat.cbInQue);
    UpdateStatus(szMessage);
  }

  if (comStat.cbOutQue) 
  {
    wsprintf(szMessage,"%d bytes in output buffer.\r\n",comStat.cbOutQue);
    UpdateStatus(szMessage);
  }

  return;
}

void ReportStatusEvent(DWORD dwStatus)
{
  BOOL fRING, fRLSD, fRXCHAR, fRXFLAG, fTXEMPTY;
  BOOL fBREAK, fCTS, fDSR, fERR;
  char szMessage[70];

  // Get status event flags.
  fCTS = EV_CTS & dwStatus;
  fDSR = EV_DSR & dwStatus;
  fERR = EV_ERR & dwStatus;
  fRING = EV_RING & dwStatus;
  fRLSD = EV_RLSD & dwStatus;
  fBREAK = EV_BREAK & dwStatus;
  fRXCHAR = EV_RXCHAR & dwStatus;
  fRXFLAG = EV_RXFLAG & dwStatus;
  fTXEMPTY = EV_TXEMPTY & dwStatus;

  // Construct status message indicating the 
  // status event flags that are set.
    
  strcpy(szMessage,"EVENT: ");
  strcat(szMessage,fCTS ? "CTS " : "");
  strcat(szMessage,fDSR ? "DSR " : "");
  strcat(szMessage,fERR ? "ERR " : "");
  strcat(szMessage,fRING ? "RING " : "");
  strcat(szMessage,fRLSD ? "RLSD " : "");
  strcat(szMessage,fBREAK ? "BREAK " : "");
  strcat(szMessage,fRXFLAG ? "RXFLAG " : "");
  strcat(szMessage,fRXCHAR ? "RXCHAR " : "");
  strcat(szMessage,fTXEMPTY ? "TXEMPTY " : "");

  // If dwStatus == NULL, then no status event flags are set.
  // This happens when the event flag is changed with SetCommMask.
    
  if (dwStatus == 0x0000)
  {
    strcat(szMessage,"NULL");
  }

  strcat(szMessage,"\r\n");

  // Queue the status message for the status control
  UpdateStatus(szMessage);

  // If an error flag is set in the event flag, then
  // report the error with the status message
  // If not, then just report the comm status.
    
  if (fERR)
  {
    ReportCommError();
  }
    
  // Might as well check the modem status and comm status now since
  // the event may have been caused by a change in line status.
  // Line status is indicated by the CheckModemStatus function.
    
  CheckModemStatus(FALSE);

  // Since line status can affect sending/receiving when 
  // hardware flow-control is used, ReportComStat should 
  // be called to show comm status.  This is called only if no error
  // was reported in the event flag.  If an error was reported, then
  // ReportCommError was called above and CheckComStat was already called
  // in that function.
    
  if (!fERR)
  {
    CheckComStat(FALSE);
  }

  return;
}

BOOL CmdAbout(HWND hwnd)
{
  DialogBox(ghInst,MAKEINTRESOURCE(IDD_ABOUT),hwnd,AboutDlgProc);
  return 0;
}

UINT InitAboutDlg(HWND hDlg)
{
  UINT uTimer;
  char * szFormat = "Microsoft Windows %s\r\nVersion %d.%d\r\nBuild %d ";
  char szVersion[256];
    
  // create timer and set initial icon id
  
  uTimer = SetTimer(hDlg,1,100,NULL);
  
  if (uTimer == 0)
  {
    ErrorReporter("SetTimer");
  }
    
  wsprintf(szVersion, 
           szFormat, 
           gOSV.dwPlatformId == VER_PLATFORM_WIN32_NT ? "NT" : "95",
           gOSV.dwMajorVersion, 
           gOSV.dwMinorVersion, 
           LOWORD(gOSV.dwBuildNumber));

  if (strlen(gOSV.szCSDVersion))
  {
    strcat(szVersion,gOSV.szCSDVersion);
  }
                            
  SetDlgItemText(hDlg,IDC_OSVERSIONINFO,szVersion);

  return uTimer;
}

BOOL CALLBACK AboutDlgProc(HWND hdlg,UINT uMessage,WPARAM wparam,LPARAM lparam)
{
  static UINT uTimer;
  static WORD wCurrentIconId;

  switch(uMessage)
  {
    case WM_INITDIALOG:
         uTimer = InitAboutDlg(hdlg);
         wCurrentIconId = IDI_APPICON;
         break;

    case WM_TIMER:
         // when timer goes off, then change to next icon
         {
           HICON hIcon;

           switch(wCurrentIconId)
           {
             case IDI_APPICON:  wCurrentIconId = IDI_APPICON2; break;
             case IDI_APPICON2: wCurrentIconId = IDI_APPICON3; break;
             case IDI_APPICON3: wCurrentIconId = IDI_APPICON4; break;
             case IDI_APPICON4: wCurrentIconId = IDI_APPICON;  break;
           }
            
           hIcon = LoadIcon(ghInst,MAKEINTRESOURCE(wCurrentIconId));
           SendMessage(GetDlgItem(hdlg,IDC_PICTURE),STM_SETICON,(WPARAM)hIcon,0);
         }
         break;

    case WM_COMMAND:
         if (LOWORD(wparam) == IDOK) 
         {
           KillTimer(hdlg,uTimer);
           EndDialog(hdlg,TRUE);
           return TRUE;
         }
         break;
  }

  return FALSE;
}

DWORD ErrorExtender(DWORD dwError, char ** szBuffer)
{
  DWORD dwRes = 0;

  dwRes = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER | 80,
                        NULL, 
                        dwError,
                        MAKELANGID(LANG_ENGLISH,SUBLANG_ENGLISH_US),
                        (LPTSTR)szBuffer, 
                        0, 
                        NULL);

  if (dwRes == 0) 
  {
    *szBuffer = (char *)LocalAlloc(LPTR,1);
    return 1;
  }

  return dwRes;
}

void ErrorReporter(char * szMessage)
{
  char * szFormat = "Error %d: %s.\n\r%s\r\n";    // format for wsprintf
  char * szExtended;      // error string translated from error code
  char * szFinal;         // final string to report
  DWORD dwExtSize;
  DWORD dwErr;

  dwErr = GetLastError();

  // Get error string from system
  dwExtSize = ErrorExtender(dwErr, &szExtended);
    
  // allocate buffer for error string from system, passed in string
  // and extra stuff from the szFormat string
    
  szFinal = (char *)LocalAlloc(LPTR, strlen(szMessage) + dwExtSize + 30);

  if (szFinal == NULL)	// if no final buffer, then can't format error
  {
    MessageBox(ghwndMain,"Cannot properly report error.","Fatal Error",MB_OK);
  }
  else 
  {	
    wsprintf(szFinal,szFormat,dwErr,szMessage,szExtended);

    OutputDebugString(szFinal);

    if (DISPLAYERRORS(TTYInfo))
    {
      MessageBox(ghwndMain,szFinal,NULL,MB_OK);
    }

    LocalFree(szFinal);                  // free final buffer
  }

  // free extended string buffer
  LocalFree(szExtended);

  return;
}

void OutputABufferToWindow(HWND hTTY,char * lpBuf,DWORD dwBufLen)
{
  RECT rect;

  // update screen buffer with new buffer
  // need to do a character by character check
  // for special characters

  int i;

  for (i = 0;i < (int) dwBufLen;i++) 
  {
    switch (lpBuf[ i ]) 
    {
      case ASCII_BEL:                // BELL CHAR
           MessageBeep(0);
           break;

      case ASCII_BS:                 // Backspace CHAR
           if (COLUMN(TTYInfo) > 0)
           {
             COLUMN(TTYInfo)--;
           }
           break;

      case ASCII_CR:                 // Carriage Return
           COLUMN(TTYInfo) = 0;
           
           if (!NEWLINE(TTYInfo))
             break;

           // FALL THROUGH
           
      case ASCII_LF:                 // Line Feed
           if (ROW(TTYInfo)++ == MAXROWS - 1)
           {
             MoveMemory((LPSTR)(SCREEN(TTYInfo)),
                        (LPSTR)(SCREEN(TTYInfo)+MAXCOLS),
                        (MAXROWS-1)*MAXCOLS);

             FillMemory((LPSTR)(SCREEN(TTYInfo)+(MAXROWS-1)*MAXCOLS),MAXCOLS,' ');
             InvalidateRect(hTTY,NULL,FALSE);
             ROW(TTYInfo)--;
           }
           break ;

      default:                       // standard character
           SCREENCHAR(TTYInfo,COLUMN(TTYInfo),ROW(TTYInfo)) = lpBuf[i];

           rect.left = (COLUMN(TTYInfo) * XCHAR(TTYInfo)) - XOFFSET(TTYInfo);
           rect.right = rect.left + XCHAR(TTYInfo);
           rect.top = (ROW(TTYInfo) * YCHAR(TTYInfo)) - YOFFSET(TTYInfo);
           rect.bottom = rect.top + YCHAR(TTYInfo);
           InvalidateRect(hTTY,&rect,FALSE);

           // Line wrap
           if (COLUMN(TTYInfo) < MAXCOLS-1)
           {
             COLUMN(TTYInfo)++;
           }
           else if (AUTOWRAP(TTYInfo))
           {
             OutputABufferToWindow(hTTY,"\r\n",2);
           }
           break;
    }
  }

  MoveTTYCursor(hTTY);
  return;
}

void OutputABufferToFile(HANDLE hFile, char * lpBuf, DWORD dwBufLen)
{
  DWORD dwWritten;

  // place buffer into file, report any errors
  if (!WriteFile(hFile,lpBuf,dwBufLen,&dwWritten,NULL))
  {
    ErrorReporter("WriteFile in file capture");
  }

  if (dwBufLen != dwWritten)
  {
    ErrorReporter("WriteFile");
  }
    
  // update transfer progress bar
  PostMessage(GetDlgItem(ghWndStatusDlg,IDC_TRANSFERPROGRESS),PBM_STEPIT,0,0);

  return;
}

void OutputABuffer(HWND hTTY,char * lpBuf,DWORD dwBufLen)
{
  if (dwBufLen == 0) 
  {
    OutputDebugString("NULL Buffer in OutputABuffer\n\r");
    return;
  }

  switch(gdwReceiveState)
  {
    case RECEIVE_TTY:
         OutputABufferToWindow(hTTY,lpBuf,dwBufLen);
         break;

    case RECEIVE_CAPTURED:
         OutputABufferToFile(ghFileCapture,lpBuf,dwBufLen);
         break;

    case RECEIVE_TTYBLACKBOXCOOKED:
         BlackBoxDGPSInterpretedWindowOutput(hTTY,lpBuf,dwBufLen); 
         break;

    case RECEIVE_CAPTUREDBLACKBOXCOOKED:
         BlackBoxDGPSInterpretedFileOutput(ghFileCapture,lpBuf,dwBufLen); 
         break;  

    default:
         OutputDebugString("Unknown receive state!\n\r");
  }

  return;
}

DWORD WINAPI WriterProc(LPVOID lpV)
{
  SYSTEM_INFO sysInfo;
  HANDLE hArray[2];
  DWORD dwRes;
  DWORD dwSize;
  BOOL fDone = FALSE;

  // create a heap for WRITE_REQUEST packets
  GetSystemInfo(&sysInfo);
  ghWriterHeap = HeapCreate(0,sysInfo.dwPageSize*2,sysInfo.dwPageSize*4);
  
  if (ghWriterHeap == NULL)
  {
    ErrorInComm("HeapCreate (write request heap)");
  }

  // create synchronization events for write requests and file transfers
  ghWriterEvent = CreateEvent(NULL,FALSE,FALSE,NULL);
  
  if (ghWriterEvent == NULL)
  {
    ErrorInComm("CreateEvent(writ request event)");
  }

  ghTransferCompleteEvent = CreateEvent(NULL,FALSE,FALSE,NULL);
    
  if (ghTransferCompleteEvent == NULL)
  {
    ErrorInComm("CreateEvent(transfer complete event)");
  }

  // initialize write request linked list
  dwSize = sizeof(WRITEREQUEST);
  gpWriterHead = (WRITEREQUEST *)HeapAlloc(ghWriterHeap,HEAP_ZERO_MEMORY,dwSize);
  gpWriterTail = (WRITEREQUEST *)HeapAlloc(ghWriterHeap,HEAP_ZERO_MEMORY,dwSize);
  gpWriterHead->pNext = gpWriterTail;
  gpWriterTail->pPrev = gpWriterHead;

  hArray[0] = ghWriterEvent;
  hArray[1] = ghThreadExitEvent;
   
  while (!fDone) 
  {
    dwRes = WaitForMultipleObjects(2,hArray,FALSE,WRITE_CHECK_TIMEOUT);
        
    switch(dwRes)
    {
      case WAIT_TIMEOUT:
           break;

      case WAIT_FAILED:
           ErrorReporter("WaitForMultipleObjects(writer proc)");
           break;

      // write request event
      case WAIT_OBJECT_0:
           HandleWriteRequests();
           break;
      
      // thread exit event
      case WAIT_OBJECT_0 + 1:
           fDone = TRUE;
           break;
    }
  }

  CloseHandle(ghTransferCompleteEvent);
  CloseHandle(ghWriterEvent);

  // Destroy WRITE_REQUEST heap
  HeapDestroy(ghWriterHeap);
  return 1;
}

void HandleWriteRequests()
{
  PWRITEREQUEST pWrite;
  BOOL fRes;
    
  pWrite = gpWriterHead->pNext;
    
  while(pWrite != gpWriterTail) 
  {
    switch(pWrite->dwWriteType)
    {
      case WRITE_CHAR:          WriterChar(pWrite);                break;

      case WRITE_FILESTART:     WriterFileStart(pWrite->dwSize);   break;

      case WRITE_FILE:          WriterFile(pWrite);
                                // free data block
                                EnterCriticalSection(&gcsDataHeap);
                                fRes = HeapFree(pWrite->hHeap, 0, pWrite->lpBuf);
                                LeaveCriticalSection(&gcsDataHeap);
                                
                                if (!fRes)
                                {
                                  ErrorReporter("HeapFree(file transfer buffer)");
                                }
                                break;

      case WRITE_FILEEND:       WriterComplete();                 break;

      case WRITE_ABORT:         WriterAbort(pWrite);              break;

      case WRITE_BLOCK:         WriterBlock(pWrite);              break;

      default:                  ErrorReporter("Bad write request"); 
                                break;
    }

    // remove current node and get next node
    pWrite = RemoveFromLinkedList(pWrite);
    pWrite = gpWriterHead->pNext;
  }

  return;
}

void WriterComplete()
{
  if (!SetEvent(ghTransferCompleteEvent))
  {
    ErrorReporter("SetEvent(transfer complete event)");
  }

  return;
}

void WriterAbort(PWRITEREQUEST pAbortNode)
{
  PWRITEREQUEST pCurrent;
  PWRITEREQUEST pNextNode;
  BOOL fRes;
  int i = 0;
  char szMessage[30];

  EnterCriticalSection(&gcsWriterHeap);
  // remove all nodes after me
  pCurrent = pAbortNode->pNext;

  while (pCurrent != gpWriterTail) 
  {
    pNextNode = pCurrent->pNext;
    fRes = HeapFree(ghWriterHeap, 0, pCurrent);
    
    if (!fRes)
      break;
      
    i++;
    pCurrent = pNextNode;
  }

  pAbortNode->pNext = gpWriterTail;
  gpWriterTail->pPrev = pAbortNode;
  LeaveCriticalSection(&gcsWriterHeap);

  wsprintf(szMessage,"%d packets ignored.\n",i);
  OutputDebugString(szMessage);

  if (!fRes)
  {
    ErrorReporter("HeapFree(Writer heap)");
  }

  if (!SetEvent(ghTransferCompleteEvent))
  {
    ErrorReporter("SetEvent(transfer complete event)");
  }

  return;
}

void WriterFile(PWRITEREQUEST pWrite)
{
  WriterGeneric(pWrite->lpBuf,pWrite->dwSize);

  // update progress indicator (even if aborting)
  if (!PostMessage(pWrite->hWndProgress,PBM_STEPIT,0,0))
  {
    ErrorReporter("PostMessage (file transfer status)");
  }

  return;
}

void WriterBlock(PWRITEREQUEST pWrite)
{   
  WriterGeneric(pWrite->lpBuf,pWrite->dwSize);
  return;
}

void WriterFileStart(DWORD dwFileSize)
{
  return;
}

void WriterChar(PWRITEREQUEST pWrite)
{
  WriterGeneric(&(pWrite->ch),1);
  return;
}

void WriterGeneric(char * lpBuf,DWORD dwToWrite)
{
  OVERLAPPED osWrite = {0};
  HANDLE hArray[2];
  DWORD dwWritten;
  DWORD dwRes;

  // If no writing is allowed, then just return
  if (NOWRITING(TTYInfo))
    return ;

  // create this writes overlapped structure hEvent
  osWrite.hEvent = CreateEvent(NULL,TRUE,FALSE,NULL);
    
  if (osWrite.hEvent == NULL)
  {
    ErrorInComm("CreateEvent(overlapped write hEvent)");
  }

  hArray[0] = osWrite.hEvent;
  hArray[1] = ghThreadExitEvent;
    
  // issue write
  
  if (!WriteFile(COMDEV(TTYInfo),lpBuf,dwToWrite,&dwWritten,&osWrite)) 
  {
    if (GetLastError() == ERROR_IO_PENDING) 
    { 
      // write is delayed
      dwRes = WaitForMultipleObjects(2,hArray,FALSE,INFINITE);
      
      switch(dwRes)
      {
        // write event set
        case WAIT_OBJECT_0:
             SetLastError(ERROR_SUCCESS);
             
             if (!GetOverlappedResult(COMDEV(TTYInfo),&osWrite,&dwWritten,FALSE)) 
             {
               if (GetLastError() == ERROR_OPERATION_ABORTED)
               {
                 UpdateStatus("Write aborted\r\n");
               }
               else
               {
                 ErrorInComm("GetOverlappedResult(in Writer)");
               } 
             }
                           
             if (dwWritten != dwToWrite) 
             {
               if ((GetLastError() == ERROR_SUCCESS) && SHOWTIMEOUTS(TTYInfo))
               {
                 UpdateStatus("Write timed out. (overlapped)\r\n");
               }
               else
               {
                 ErrorReporter("Error writing data to port (overlapped)");
               }
             }
             
             break;

        //
        // thread exit event set
        //
        case WAIT_OBJECT_0 + 1:
             break;

        // wait timed out
        case WAIT_TIMEOUT:
             UpdateStatus("Wait Timeout in WriterGeneric.\r\n");
             break;

        case WAIT_FAILED:
        default:    
             ErrorInComm("WaitForMultipleObjects(WriterGeneric)");
             break;
      }
    }
    else    
    {
      // writefile failed, but it isn't delayed
      ErrorInComm("WriteFile (in Writer)");
    } 
  }
  else 
  {
    // writefile returned immediately
    if (dwWritten != dwToWrite)
    {
      UpdateStatus("Write timed out. (immediate)\r\n");
    }
  }

  CloseHandle(osWrite.hEvent);
    
  return;
}

BOOL WriterAddNewNode(DWORD dwRequestType,DWORD dwSize,char ch,char * lpBuf,HANDLE hHeap,HWND hProgress)
{
  PWRITEREQUEST pWrite;

  // allocate new packet
  pWrite = (WRITEREQUEST *)HeapAlloc(ghWriterHeap,0,sizeof(WRITEREQUEST));

  if (pWrite == NULL) 
  {
    ErrorReporter("HeapAlloc (writer packet)");
    return FALSE;
  }

  // assign packet info
  pWrite->dwWriteType  = dwRequestType;
  pWrite->dwSize       = dwSize;
  pWrite->ch           = ch;
  pWrite->lpBuf        = lpBuf;
  pWrite->hHeap        = hHeap;
  pWrite->hWndProgress = hProgress;

  AddToLinkedList(pWrite);
    
  return TRUE;
}

BOOL WriterAddNewNodeTimeout(DWORD dwRequestType,DWORD dwSize,char ch,char * lpBuf, 
                             HANDLE hHeap,HWND hProgress,DWORD dwTimeout)
{
  PWRITEREQUEST pWrite;

  // attempt first allocation
  pWrite = (WRITEREQUEST *)HeapAlloc(ghWriterHeap,0,sizeof(WRITEREQUEST));
    
  if (pWrite == NULL) 
  {
    Sleep(dwTimeout);
    // attempt second allocation
    pWrite = (WRITEREQUEST *)HeapAlloc(ghWriterHeap,0,sizeof(WRITEREQUEST));
        
    if (pWrite == NULL) 
    {
      ErrorReporter("HeapAlloc (writer packet)");
      return FALSE;
    }
  }

  // assign packet info
  pWrite->dwWriteType  = dwRequestType;
  pWrite->dwSize       = dwSize;
  pWrite->ch           = ch;
  pWrite->lpBuf        = lpBuf;
  pWrite->hHeap        = hHeap;
  pWrite->hWndProgress = hProgress;

  AddToLinkedList(pWrite);
    
  return TRUE;
}

BOOL WriterAddFirstNodeTimeout(DWORD dwRequestType,DWORD dwSize,char ch,char * lpBuf, 
                               HANDLE hHeap,HWND hProgress,DWORD dwTimeout)
{
  PWRITEREQUEST pWrite;

  // attempt first allocation
  pWrite = (WRITEREQUEST *)HeapAlloc(ghWriterHeap,0,sizeof(WRITEREQUEST));
  
  if (pWrite == NULL) 
  {
    Sleep(dwTimeout);
    // attempt second allocation
    pWrite = (WRITEREQUEST *)HeapAlloc(ghWriterHeap,0,sizeof(WRITEREQUEST));
        
    if (pWrite == NULL) 
    {
      ErrorReporter("HeapAlloc (writer packet)");
      return FALSE;
    }
  }

  // assign packet info
  pWrite->dwWriteType  = dwRequestType;
  pWrite->dwSize       = dwSize;
  pWrite->ch           = ch;
  pWrite->lpBuf        = lpBuf;
  pWrite->hHeap        = hHeap;
  pWrite->hWndProgress = hProgress;

  AddToFrontOfLinkedList(pWrite);
    
  return TRUE;
}

BOOL WriterAddExistingNode(PWRITEREQUEST pNode,DWORD dwRequestType,DWORD dwSize,char ch, 
                           char * lpBuf,HANDLE hHeap,HWND hProgress)
{
  // assign packet info
  pNode->dwWriteType  = dwRequestType;
  pNode->dwSize       = dwSize;
  pNode->ch           = ch;
  pNode->lpBuf        = lpBuf;
  pNode->hHeap        = hHeap;
  pNode->hWndProgress = hProgress;

  AddToLinkedList(pNode);
 
  return TRUE;
}

void AddToLinkedList(PWRITEREQUEST pNode)
{
  PWRITEREQUEST pOldLast;
  // add node to linked list
  EnterCriticalSection(&gcsWriterHeap);

  pOldLast = gpWriterTail->pPrev;

  pNode->pNext = gpWriterTail;
  pNode->pPrev = pOldLast;

  pOldLast->pNext = pNode;
  gpWriterTail->pPrev = pNode;

  LeaveCriticalSection(&gcsWriterHeap);

  // notify writer thread that a node has been added
  if (!SetEvent(ghWriterEvent))
  {
    ErrorReporter("SetEvent( writer packet )");
  }
    
  return;    
}

void AddToFrontOfLinkedList(PWRITEREQUEST pNode)
{
  PWRITEREQUEST pNextNode;
  // add node to linked list
  EnterCriticalSection(&gcsWriterHeap);

  pNextNode = gpWriterHead->pNext;
    
  pNextNode->pPrev = pNode;
  gpWriterHead->pNext = pNode;
    
  pNode->pNext = pNextNode;
  pNode->pPrev = gpWriterHead;

  LeaveCriticalSection(&gcsWriterHeap);

  // notify writer thread that a node has been added
  if (!SetEvent(ghWriterEvent))
  {
    ErrorReporter("SetEvent( writer packet )");
  }
    
  return;    
}

PWRITEREQUEST RemoveFromLinkedList(PWRITEREQUEST pNode)
{
  PWRITEREQUEST pNextNode;
  PWRITEREQUEST pPrevNode;
  BOOL bRes;

  EnterCriticalSection(&gcsWriterHeap);
    
  pNextNode = pNode->pNext;
  pPrevNode = pNode->pPrev;    
    
  bRes = HeapFree(ghWriterHeap,0,pNode);
    
  pPrevNode->pNext = pNextNode;
  pNextNode->pPrev = pPrevNode;

  LeaveCriticalSection(&gcsWriterHeap);

  if (!bRes)
  {
    ErrorReporter("HeapFree(write request)");
  }

  return pNextNode;     // return the freed node's pNext (maybe the tail)
}

void ErrorInComm(char * szMessage)
{
  ErrorReporter(szMessage);
  BreakDownCommPort();
  ExitProcess(0);
}

DWORD WINAPI ReaderAndStatusProc(LPVOID lpV)
{
  OVERLAPPED osReader = {0};  // overlapped structure for read operations
  OVERLAPPED osStatus = {0};  // overlapped structure for status operations
  HANDLE     hArray[NUM_READSTAT_HANDLES];
  DWORD      dwStoredFlags = 0xFFFFFFFF;      // local copy of event flags
  DWORD      dwCommEvent;     // result from WaitCommEvent
  DWORD      dwOvRes;         // result from GetOverlappedResult
  DWORD 	   dwRead;          // bytes actually read
  DWORD      dwRes;           // result from WaitForSingleObject
  BOOL       fWaitingOnRead = FALSE;
  BOOL       fWaitingOnStat = FALSE;
  BOOL       fThreadDone = FALSE;
  char   	   lpBuf[AMOUNT_TO_READ];
  HWND  	   hTTY;


  hTTY = (HWND)(HANDLE) lpV; // inserted the "(HWND)", Walter

  // create two overlapped structures, one for read events
  // and another for status events
  osReader.hEvent = CreateEvent(NULL,TRUE,FALSE,NULL);
  
  if (osReader.hEvent == NULL)
  {
    ErrorInComm("CreateEvent(Reader Event)");
  }

  osStatus.hEvent = CreateEvent(NULL,TRUE,FALSE,NULL);
    
  if (osStatus.hEvent == NULL)
  {
    ErrorInComm("CreateEvent(Status Event)");
  }

  // We want to detect the following events:
  // Read events (from ReadFile)
  // Status events (from WaitCommEvent)
  // Status message events (from our UpdateStatus)
  // Thread exit events (from our shutdown functions)
  
  hArray[0] = osReader.hEvent;
  hArray[1] = osStatus.hEvent;
  hArray[2] = ghStatusMessageEvent;
  hArray[3] = ghThreadExitEvent;

  // initial check, forces updates
  CheckModemStatus(TRUE);
  CheckComStat(TRUE);

  while (!fThreadDone) 
  {
    // If no reading is allowed, then set flag to
    // make it look like a read is already outstanding.
    if (NOREADING(TTYInfo))
    {
      fWaitingOnRead = TRUE;
    }
        
    // if no read is outstanding, then issue another one
    if (!fWaitingOnRead) 
    {
      if (!ReadFile(COMDEV(TTYInfo),lpBuf,AMOUNT_TO_READ,&dwRead,&osReader)) 
      {
        if (GetLastError() != ERROR_IO_PENDING)	  // read not delayed?
        {
          ErrorInComm("ReadFile in ReaderAndStatusProc");
        }

        fWaitingOnRead = TRUE;
      }
      else 
      {    // read completed immediately
        if ((dwRead != MAX_READ_BUFFER) && SHOWTIMEOUTS(TTYInfo))
        {
          UpdateStatus("Read timed out immediately.\r\n");
        }

        if (dwRead)
        {
          OutputABuffer(hTTY,lpBuf,dwRead);
        }
      }
    }

    // If status flags have changed, then reset comm mask.
    // This will cause a pending WaitCommEvent to complete
    // and the resultant event flag will be NULL.
    if (dwStoredFlags != EVENTFLAGS(TTYInfo)) 
    {
      dwStoredFlags = EVENTFLAGS(TTYInfo);
      
      if (!SetCommMask(COMDEV(TTYInfo),dwStoredFlags))
      {
        ErrorReporter("SetCommMask");
      }
    }

    // If event checks are not allowed, then make it look
    // like an event check operation is outstanding
    if (NOEVENTS(TTYInfo))
    {
      fWaitingOnStat = TRUE;
    }
    
    // if no status check is outstanding, then issue another one
    if (!fWaitingOnStat) 
    {
      if (NOEVENTS(TTYInfo))
      {
        fWaitingOnStat = TRUE;
      }
      else 
      {
        if (!WaitCommEvent(COMDEV(TTYInfo),&dwCommEvent,&osStatus)) 
        {
          if (GetLastError() != ERROR_IO_PENDING)	  // Wait not delayed?
          {
            ErrorReporter("WaitCommEvent");
          }
          else
          {
            fWaitingOnStat = TRUE;
          }
        }
        else
        {
          // WaitCommEvent returned immediately
          ReportStatusEvent(dwCommEvent);
        }
      }
    }

    // wait for pending operations to complete
    if (fWaitingOnStat && fWaitingOnRead) 
    {
      dwRes = WaitForMultipleObjects(NUM_READSTAT_HANDLES,hArray,FALSE,STATUS_CHECK_TIMEOUT);

      switch(dwRes)
      {
        // read completed
        case WAIT_OBJECT_0:
             if (!GetOverlappedResult(COMDEV(TTYInfo),&osReader,&dwRead,FALSE)) 
             {
               if (GetLastError() == ERROR_OPERATION_ABORTED)
               {
                 UpdateStatus("Read aborted\r\n");
               }
               else
               {
                 ErrorInComm("GetOverlappedResult (in Reader)");
               }
             }
             else 
             {      
               // read completed successfully
               if ((dwRead != MAX_READ_BUFFER) && SHOWTIMEOUTS(TTYInfo))
               {
                 UpdateStatus("Read timed out overlapped.\r\n");
               }

               if (dwRead)
               {
                 OutputABuffer(hTTY,lpBuf,dwRead);
               }
             }

             fWaitingOnRead = FALSE;
             break;

             // status completed
        case WAIT_OBJECT_0 + 1: 
             if (!GetOverlappedResult(COMDEV(TTYInfo),&osStatus,&dwOvRes,FALSE)) 
             {
               if (GetLastError() == ERROR_OPERATION_ABORTED)
               {
                 UpdateStatus("WaitCommEvent aborted\r\n");
               }
               else
               {
                 ErrorInComm("GetOverlappedResult (in Reader)");
               }
             }
             else       // status check completed successfully
             {
               ReportStatusEvent(dwCommEvent);
             }

             fWaitingOnStat = FALSE;
             break;

             // status message event
        case WAIT_OBJECT_0 + 2:
             StatusMessage();
             break;

             // thread exit event
        case WAIT_OBJECT_0 + 3:
             fThreadDone = TRUE;
             break;

        case WAIT_TIMEOUT:
             // timeouts are not reported because they happen too often
             // OutputDebugString("Timeout in Reader & Status checking\n\r");
             
             // if status checks are not allowed, then don't issue the
             // modem status check nor the com stat check
             if (!NOSTATUS(TTYInfo)) 
             {
               CheckModemStatus(FALSE);    // take this opportunity to do
               CheckComStat(FALSE);        // a modem status check and
             }                             // a comm status check
             break;                       

        default:
             ErrorReporter("WaitForMultipleObjects(Reader & Status handles)");
             break;
      }
    }
  }

  // close event handles
  CloseHandle(osReader.hEvent);
  CloseHandle(osStatus.hEvent);

  return 1;
}

void ErrorHandler(char * szMessage)
{	
  ErrorReporter(szMessage);
  ExitProcess(0);
}

// Here I do it the primitive way. Using Allen Denver's writer thread function TransferThreadProc 
// would be overkill and I would have to modify it and spoil its beauty. You will find much code  
// in this source code file that is without functionality for dBlackBoxPlus. Try to throw these 
// sections out if you do not like this -- DGPSListener is still in the brainstorming phase. 
// Remember that this program was written to learn a lot of things, not to sell it. Some 
// inefficiency is not harmful at all for this purpose.

BOOL BlackBoxDGPSSelectSentences(BBSENTENCE bbs,BBACTION bba)
{
  static char targets[][6] = {
  "GPALM", "GPGGA", "GPGSA", "GPGSV",
  "GPRMC", "GPVTG", "LCGLL", "LCVTG",
  "PGRME", "PGRMF", "PGRMT", "PGRMV" };

  char command[128];
  LPSTR targetSentence;
  char targetSentenceMode;
  OVERLAPPED osWrite = {0};
  DWORD dwWritten;
  DWORD dwToWrite;
  DWORD dwRes;
  BOOL fRes;

  // build targetSentenceMode 
  switch (bba)
  {
    case SELECTALL:
         targetSentenceMode ='3'; // enable all output sentences (except GPALM)      
         break;

    case DESELECTALL:
         targetSentenceMode ='2'; // disable all output sentences weiter hier
         break;

    case SELECT: 
         targetSentenceMode ='1'; // enable specific sentence
         break;

    case DESELECT:
         targetSentenceMode ='0'; // disable specific sentence
         break;

    default: 
         break; 
  }
  
  // targetSentenceDescription (bbs) is not checked by DBlackBoxPlus if targetSentenceMode (bba) is  
  // 2 (disable all output sentences) or 3 (enable all output sentences except GPALM)

  // build command
  switch (bbs)
  {
    case S_GPALM:
         targetSentence = targets[S_GPALM];
         break;
          
    case S_GPGGA:
         targetSentence = targets[S_GPGGA];
         break;

    case S_GPGSA:
         targetSentence = targets[S_GPGSA];
         break;

    case S_GPGSV:
         targetSentence = targets[S_GPGSV];
         break;

    case S_GPRMC:
         targetSentence = targets[S_GPRMC];
         break;

    case S_GPVTG:
         targetSentence = targets[S_GPVTG];
         break;

    case S_LCGLL:
         targetSentence = targets[S_LCGLL];
         break;

    case S_LCVTG:
         targetSentence = targets[S_LCVTG];
         break;

    case S_PGRME:
         targetSentence = targets[S_PGRME];
         break;

    case S_PGRMF:
         targetSentence = targets[S_PGRMF];
         break;

    case S_PGRMT:
         targetSentence = targets[S_PGRMT];
         break;

    case S_PGRMV:
         targetSentence = targets[S_PGRMV];
         break;

    case S_IGNORED:
         targetSentence = NULL;
         break; 
  }

  sprintf(command,"$PGRMO,%s,%c,%c%c",targetSentence,targetSentenceMode,(char)ASCII_CR,(char)ASCII_LF);
  dwToWrite = strlen(command);
  // send command
  osWrite.hEvent = CreateEvent(NULL,TRUE,FALSE,NULL);

  if (osWrite.hEvent == NULL)
  {
    return FALSE; // Error creating overlapped event handle
  }

  if (!WriteFile(COMDEV(TTYInfo),command,dwToWrite,&dwWritten,&osWrite))
  {
    if (GetLastError() != ERROR_IO_PENDING)
    {
      // WriteFile failed, but it isn't delayed
      fRes = FALSE;
    }
    else
    {
      // Write is pending
      dwRes = WaitForSingleObject(osWrite.hEvent,INFINITE);

      switch (dwRes)
      {
        case WAIT_OBJECT_0: // Overlapped event has been signaled
             if (!GetOverlappedResult(COMDEV(TTYInfo),&osWrite,&dwWritten,FALSE))
             {
               fRes = FALSE;
             }
             else
             {
               if (dwWritten != dwToWrite)
               {
                 // Write operation timed out, abort (development only)
                 fRes = FALSE;
               }
               else
               {
                 // Write operation completed successfully
                 fRes = TRUE;
               }            
             }
             break;

        default:
             // An error has occured in WaitForSingleObject. This usually
             // indicates a problem with the overlapped event handle.
             fRes = FALSE;
             break;
      }
    }
  }
  else
  {
    // WriteFile completed immediately
    if (dwWritten != dwToWrite)
    {
      // Write operation timed out, abort (development only)
      fRes = FALSE;
    }
    else
    {
      fRes = TRUE;
    }
  }

  CloseHandle(osWrite.hEvent);
  return fRes;
}

void BlackBoxDGPSInterpretedWindowOutput(HWND hTTY,char * lpBuf,DWORD dwBufLen)
{
  char nextCmdPart[128];
  memset(nextCmdPart,0x0,128);

  if (BlackBoxEndOfCommand(lpBuf,dwBufLen,CmdBuffer,nextCmdPart))
  {  
    BlackBoxInterpretCmd(CmdBuffer);
    OutputABufferToWindow(hTTY,CmdBuffer,strlen(CmdBuffer)); // output the command
    memset(CmdBuffer,0x0,128);  // reinitialize buffer
    strcat(CmdBuffer,nextCmdPart);
  }
}

void BlackBoxDGPSInterpretedFileOutput(HANDLE hFile,char * lpBuf,DWORD dwBufLen)
{
  char nextCmdPart[128];
  memset(nextCmdPart,0x0,128);

  if (BlackBoxEndOfCommand(lpBuf,dwBufLen,CmdBuffer,nextCmdPart))
  {  
    BlackBoxInterpretCmd(CmdBuffer);
    OutputABufferToFile(hFile,CmdBuffer,strlen(CmdBuffer)); // output the command
    memset(CmdBuffer,0x0,128);  // reinitialize buffer
    strcat(CmdBuffer,nextCmdPart);
  } 
}

BOOL BlackBoxEndOfCommand(char * cmdPart,DWORD dwBufLen,char * target,char * nextCmdPart)
{
  int len = (int)dwBufLen;
  int lf_position = -1;
  char terminatedBuf[128];
  memset(terminatedBuf,0x0,128);

  for (int run=0;run < len;run++)
  {
    if (cmdPart[run] == ASCII_LF) // end of command
    {
      terminatedBuf[run] = cmdPart[run];
      lf_position = run;
    }
    else
    {
      terminatedBuf[run] = cmdPart[run];
    }
  }

  if (lf_position != -1) // end of command detected
  {
    if (lf_position < (len-1)) // parts of next cmds follow
    {
      strcpy(nextCmdPart,terminatedBuf+lf_position+1); // copy begin of the next cmd
      terminatedBuf[lf_position+1] = '\0';             // terminate after this cmd
    }

    strcat(target,terminatedBuf);
    return TRUE;
  }
  else
  {
    strcat(target,terminatedBuf);
    return FALSE;
  }
}

void BlackBoxInterpretCmd(char * rawCmd)
{
  KNOWNSENTENCES sentence = NMEAParserGetSentenceType(rawCmd);

  if (NMEAParserComputeChecksum(rawCmd) != NMEAParserGetContainedChecksum(rawCmd))
  {
    wsprintf(rawCmd,"%d\r\n",CHECKSUM_PARSE_ERROR);
    return;
  }

  switch (sentence)
  {
    case GPGGA: NMEAParserAnalyzeGPGGA(rawCmd); break;
    case GPGSA: wsprintf(rawCmd,"\0"); break;   // Not analyzed in this version
    case GPGSV: wsprintf(rawCmd,"\0"); break;   // Not analyzed in this version
    case GPRMC: wsprintf(rawCmd,"\0"); break;
    case PGRMT: wsprintf(rawCmd,"\0"); break; // Not analyzed in this version
    default: wsprintf(rawCmd,"%d\r\n",UNKNOWNSENTENCE_PARSE_ERROR); break;
  }
}

void NMEAParserAnalyzeGPRMC(char * rawCmd)
{
  CString UTCTime         = NMEAParserField(1,rawCmd);
  CString status          = NMEAParserField(2,rawCmd);
  CString lat             = NMEAParserField(3,rawCmd);
  CString latHemisphere   = NMEAParserField(4,rawCmd);
  CString lon             = NMEAParserField(5,rawCmd);
  CString lonHemisphere   = NMEAParserField(6,rawCmd); 

  if (status.GetAt(0) == 'V') 
  {
    wsprintf(rawCmd,"No position fix available\r\n");   
  }
  else
  {
    CPolarCoordinate polarCoord;
    CEarthCoordinate cartesianCoord;

    double latDegrees = NMEAParserGetDegreesFromNMEALat(latHemisphere.GetAt(0),lat.operator LPCTSTR());
    double latMinutes = NMEAParserGetMinutesFromNMEALat(lat.operator LPCTSTR());
    double lonDegrees = NMEAParserGetDegreesFromNMEALon(lonHemisphere.GetAt(0),lon.operator LPCTSTR());
    double lonMinutes = NMEAParserGetMinutesFromNMEALon(lon.operator LPCTSTR());
    double latDecimal = CMath::ConvertDegreesMinutesSecondsCoordinateToDecimalDegrees(latDegrees,latMinutes,0.0);
    double lonDecimal = CMath::ConvertDegreesMinutesSecondsCoordinateToDecimalDegrees(lonDegrees,lonMinutes,0.0);

    polarCoord.SetUpDownAngleInDegrees(latDecimal);    // aka latitude
    polarCoord.SetLeftRightAngleInDegrees(lonDecimal); // aka longitude
    CEarth earth; 
    earth.Convert(polarCoord,cartesianCoord);

    if (BeginMarkerWaitingToBeSet)
    {
      sprintf(rawCmd,
              "%lf %lf %lf ---------begin of interesting section-----------\r\n",
              cartesianCoord.GetXCoordinateInMeters(),
              cartesianCoord.GetYCoordinateInMeters(),
              cartesianCoord.GetZCoordinateInMeters());
      
      BeginMarkerWaitingToBeSet = FALSE;
    }
    else if (EndMarkerWaitingToBeSet)
    {
      sprintf(rawCmd,
              "%lf %lf %lf ----------end of interesting section----------\r\n",
              cartesianCoord.GetXCoordinateInMeters(),
              cartesianCoord.GetYCoordinateInMeters(),
              cartesianCoord.GetZCoordinateInMeters());
      
      EndMarkerWaitingToBeSet = FALSE;
    }
    else
    {
      sprintf(rawCmd,
              "%lf %lf %lf\r\n",
              cartesianCoord.GetXCoordinateInMeters(),
              cartesianCoord.GetYCoordinateInMeters(),
              cartesianCoord.GetZCoordinateInMeters());
    }
  }
}

void NMEAParserAnalyzeGPGGA(char * rawCmd)
{
  CString UTCTime           = NMEAParserField(1,rawCmd);
  CString lat               = NMEAParserField(2,rawCmd);
  CString latHemisphere     = NMEAParserField(3,rawCmd);
  CString lon               = NMEAParserField(4,rawCmd);
  CString lonEastOrWest     = NMEAParserField(5,rawCmd);
  CString qualityIndicator  = NMEAParserField(6,rawCmd); 
  CString pecisionDilution  = NMEAParserField(8,rawCmd);
  CString antennaAboveSea   = NMEAParserField(9,rawCmd);
  CString geoidalSeparation = NMEAParserField(11,rawCmd);

  if (qualityIndicator.GetAt(0) == '0') 
  {
    wsprintf(rawCmd,"No position fix available\r\n");   
  }
  else
  {
    double latDegrees = NMEAParserGetDegreesFromNMEALat(latHemisphere.GetAt(0),lat.operator LPCTSTR());
    double latMinutes = NMEAParserGetMinutesFromNMEALat(lat.operator LPCTSTR());
    double lonDegrees = NMEAParserGetDegreesFromNMEALon(lonEastOrWest.GetAt(0),lon.operator LPCTSTR());
    double lonMinutes = NMEAParserGetMinutesFromNMEALon(lon.operator LPCTSTR());
    double latDecimal = CMath::ConvertDegreesMinutesSecondsCoordinateToDecimalDegrees(latDegrees,latMinutes,0.0);
    double lonDecimal = CMath::ConvertDegreesMinutesSecondsCoordinateToDecimalDegrees(lonDegrees,lonMinutes,0.0);
    double gaussKruegerR = 0.0;
    double gaussKruegerH = 0.0;
    double lfAntennaAboveSea = atof(antennaAboveSea.operator LPCTSTR());
    double lfGeoidalSeparation = atof(geoidalSeparation.operator LPCTSTR());
    // double height = lfAntennaAboveSea - lfGeoidalSeparation; wrong, fixed Oct 23 2002 
    double height = lfAntennaAboveSea + lfGeoidalSeparation;
    // Since height is rather uncritical in the 7 parameter Helmert transformation, the
    // results will not change significantly.  
    
    WGS84toDHDN Wgs84ToDHDNConverter;
    Wgs84ToDHDNConverter.Convert(latDecimal,lonDecimal,height,gaussKruegerR,gaussKruegerH); 

    if (BeginMarkerWaitingToBeSet)
    {
      sprintf(rawCmd,
              "%3.7lf\t%3.7lf\t%.3lf\t%.3lf\t%s\r\n----------begin of interesting section----------\r\n",
              latDecimal,
              lonDecimal,
              gaussKruegerR,
              gaussKruegerH,
              pecisionDilution.operator LPCTSTR());

      MessageBeep(MB_ICONEXCLAMATION);
      BeginMarkerWaitingToBeSet = FALSE;
    }
    else if (EndMarkerWaitingToBeSet)
    {
      sprintf(rawCmd,
              "%3.7lf\t%3.7lf\t%.3lf\t%.3lf\t%s\r\n----------end of interesting section----------\r\n",
              latDecimal,
              lonDecimal,
              gaussKruegerR,
              gaussKruegerH,
              pecisionDilution.operator LPCTSTR());

      MessageBeep(MB_ICONEXCLAMATION);
      EndMarkerWaitingToBeSet = FALSE;
    }
    else
    {
      sprintf(rawCmd,
              "%3.7lf\t%3.7lf\t%.3lf\t%.3lf\t%s\r\n",
              latDecimal,
              lonDecimal,
              gaussKruegerR,
              gaussKruegerH,
              pecisionDilution.operator LPCTSTR());
      
      MessageBeep(MB_OK);
    }
  }
}

KNOWNSENTENCES NMEAParserGetSentenceType(const char *sentence)
{
  // snt[0] is '$'

  switch (sentence[1])
  {
    case 'G':
          switch (sentence[2])
          {
            case 'P':
                 switch(sentence[3])
                 {
                   case 'G':
                        switch(sentence[4])
                        {
                          case 'G':
                               switch(sentence[5])
                               {
                                 case 'A':
                                      return GPGGA;
                                 
                                 default:
                                      return UNIMPLEMENTED;
                               }
                               break;
                          
                          case 'S':
                               switch(sentence[5])
                               {
                                 case 'A':
                                      return GPGSA;

                                 case 'V':
                                      return GPGSV;

                                 default:
                                      return UNIMPLEMENTED;
                               } 
                               break;

                          default:
                               return UNIMPLEMENTED; 
                        }
                        break;
                   
                   case 'R': 
                        switch(sentence[4])
                        {
                          case 'M':
                               switch(sentence[5])
                               {
                                 case 'C':
                                      return GPRMC;
                                 
                                 default:
                                      return UNIMPLEMENTED;
                               }
                               break; 
                          
                          default:
                               return UNIMPLEMENTED;
                        }
                   
                   default:
                        return UNIMPLEMENTED; 
                 }
                 break;
        
            default:
                 return UNIMPLEMENTED; 
          }
          break;

    case 'P':
          switch (sentence[2])
          {
            case 'G': 
                 return PGRMT;
            
            default:                
                 return UNIMPLEMENTED;
          }
          break;

    default:
          return UNIMPLEMENTED;
  }
} 

// Taken from Blackburn's NMEA Parser
DWORD NMEAParserComputeChecksum(char * sentence)
{
  BYTE checksum_value = 0;

  int string_length = strlen(sentence);
  int index = 1; // Skip over the $ at the begining of the sentence

  while(index < string_length && 
        sentence[index] != '*' && 
        sentence[index] != CARRIAGE_RETURN && 
        sentence[index] != LINE_FEED)
  {
    checksum_value ^= sentence[index];
    index++;
  }

  return((DWORD)checksum_value);
}

DWORD NMEAParserGetContainedChecksum(char * sentence)
{
  int idx = 0;
  char stringRepresentation[3];

  while(sentence[idx] != '*')
  {
    idx++;
  }

  idx++;

  stringRepresentation[0] = sentence[idx];
  idx++;
  stringRepresentation[1] = sentence[idx];;
  stringRepresentation[2] = '\0';

  return(NMEAParserHexValue(stringRepresentation));
}

// Taken from Blackburn's NMEA Parser
DWORD NMEAParserHexValue(const char * hex_string)
{
  DWORD return_value = 0;

  sscanf(hex_string,"%lx",(DWORD *)&return_value );

  return(return_value);
}

// Taken from Blackburn's NMEA Parser 
CString NMEAParserField(int desired_field_number,char * sentence)
{
  CString return_string;

  return_string.Empty();

  int index                = 1; // Skip over the $ at the begining of the sentence
  int current_field_number = 0;
  int string_length        = 0;

  string_length = strlen(sentence);

  while(current_field_number < desired_field_number && index < string_length)
  {
    if (sentence[index] == ',' || sentence[index] == '*')
    {
      current_field_number++;
    }

    index++;
  }

  if (current_field_number == desired_field_number)
  {
    while(index < string_length &&
          sentence[index] != ',' &&
          sentence[index] != '*' &&
          sentence[index] != 0x00)
    {
      return_string += sentence[index];
      index++;
    }
  }

  return(return_string);
}

// These are the Geodesy Fundation Classes (Sam Blackburn)

CEarthCoordinate::CEarthCoordinate(void)
{
  m_X_CoordinateInMeters = 0.0;
  m_Y_CoordinateInMeters = 0.0;
  m_Z_CoordinateInMeters = 0.0;
}

CEarthCoordinate::CEarthCoordinate(const CEarthCoordinate& source)
{
  Copy(source);
}

CEarthCoordinate::~CEarthCoordinate(void)
{
  m_X_CoordinateInMeters = 0.0;
  m_Y_CoordinateInMeters = 0.0;
  m_Z_CoordinateInMeters = 0.0;
}

void CEarthCoordinate::Copy(const CEarthCoordinate& source)
{
  m_X_CoordinateInMeters = source.m_X_CoordinateInMeters;
  m_Y_CoordinateInMeters = source.m_Y_CoordinateInMeters;
  m_Z_CoordinateInMeters = source.m_Z_CoordinateInMeters;
}

void CEarthCoordinate::Get(double& x_coordinate,double& y_coordinate,double& z_coordinate) const
{
  x_coordinate = m_X_CoordinateInMeters;
  y_coordinate = m_Y_CoordinateInMeters;
  z_coordinate = m_Z_CoordinateInMeters;
}

double CEarthCoordinate::GetXCoordinateInMeters(void) const
{
  return( m_X_CoordinateInMeters );
}

double CEarthCoordinate::GetYCoordinateInMeters(void) const
{
  return( m_Y_CoordinateInMeters );
}

double CEarthCoordinate::GetZCoordinateInMeters(void) const
{
  return( m_Z_CoordinateInMeters );
}

void CEarthCoordinate::Set(double x_coordinate,double y_coordinate,double z_coordinate)
{
  m_X_CoordinateInMeters = x_coordinate;
  m_Y_CoordinateInMeters = y_coordinate;
  m_Z_CoordinateInMeters = z_coordinate;
}

void CEarthCoordinate::SetXCoordinateInMeters(double x_coordinate)
{
  m_X_CoordinateInMeters = x_coordinate;
}

void CEarthCoordinate::SetYCoordinateInMeters(double y_coordinate)
{
  m_Y_CoordinateInMeters = y_coordinate;
}

void CEarthCoordinate::SetZCoordinateInMeters(double z_coordinate)
{
  m_Z_CoordinateInMeters = z_coordinate;
}

CEarthCoordinate& CEarthCoordinate::operator=(const CEarthCoordinate& source)
{
  Copy(source);
  return(*this);
}

CPolarCoordinate::CPolarCoordinate(void)
{
  m_UpDownAngleInDegrees        = 0.0;
  m_LeftRightAngleInDegrees     = 0.0;
  m_DistanceFromSurfaceInMeters = 0.0;
}

CPolarCoordinate::CPolarCoordinate(const CPolarCoordinate& source)
{
  Copy(source);
}

CPolarCoordinate::~CPolarCoordinate(void)
{
  m_UpDownAngleInDegrees        = 0.0;
  m_LeftRightAngleInDegrees     = 0.0;
  m_DistanceFromSurfaceInMeters = 0.0;
}

void CPolarCoordinate::Copy(const CPolarCoordinate& source)
{
  m_UpDownAngleInDegrees        = source.m_UpDownAngleInDegrees;
  m_LeftRightAngleInDegrees     = source.m_LeftRightAngleInDegrees;
  m_DistanceFromSurfaceInMeters = source.m_DistanceFromSurfaceInMeters;
}

void CPolarCoordinate::Get(double& up_down_angle,double& left_right_angle,double& length) const
{
  up_down_angle    = m_UpDownAngleInDegrees;
  left_right_angle = m_LeftRightAngleInDegrees;
  length           = m_DistanceFromSurfaceInMeters;
}

double CPolarCoordinate::GetUpDownAngleInDegrees(void) const
{
  return(m_UpDownAngleInDegrees);
}

double CPolarCoordinate::GetLeftRightAngleInDegrees(void) const
{
  return(m_LeftRightAngleInDegrees);
}

double CPolarCoordinate::GetDistanceFromSurfaceInMeters(void) const
{
  return(m_DistanceFromSurfaceInMeters);
}

void CPolarCoordinate::Set(double up_down_angle,double left_right_angle,double length)
{
  m_UpDownAngleInDegrees        = up_down_angle;
  m_LeftRightAngleInDegrees     = left_right_angle;
  m_DistanceFromSurfaceInMeters = length;
}

void CPolarCoordinate::SetUpDownAngleInDegrees(double up_down_angle)
{
  m_UpDownAngleInDegrees = up_down_angle;
}

void CPolarCoordinate::SetLeftRightAngleInDegrees(double left_right_angle)
{
  m_LeftRightAngleInDegrees = left_right_angle;
}

void CPolarCoordinate::SetDistanceFromSurfaceInMeters(double distance_from_surface)
{
  m_DistanceFromSurfaceInMeters = distance_from_surface;
}

CPolarCoordinate& CPolarCoordinate::operator=(const CPolarCoordinate& source)
{
  Copy(source);
  return(*this);
}

CEarth::CEarth(int ellipsoid_identifier)
{
  m_Initialize();
  SetEllipsoid(ellipsoid_identifier);
}

CEarth::~CEarth()
{
  m_Initialize();
}

void CEarth::AddLineOfSightDistanceAndDirectionToCoordinate(const CPolarCoordinate& point_1,
                                                            double distance,
                                                            double direction,
                                                            CPolarCoordinate& point_2,
                                                            double height_above_surface_of_point_2)
{
  // The method used here is to convert the straight (line-of-sight) distance to
  // a surface distance and then find out the position using the surface distance.
  // This is a translation of the MMDIST routine found in the FORWRD3D program at
  // ftp://ftp.ngs.noaa.gov/pub/pcsoft/for_inv.3d/source/forwrd3d.for

  double c                                           = 0.0;
  double cosine_of_latitude_squared                  = 0.0;
  double cosine_of_direction_squared                 = 0.0;
  double cosine_of_point_1_latitude                  = 0.0;
  double difference_in_height                        = 0.0;
  double direction_in_radians                        = 0.0;
  double distance_adjusted_for_differences_in_height = 0.0;
  double height_above_surface_of_point_1             = 0.0;
  double n                                           = 0.0;
  double point_1_latitude_in_radians                 = 0.0;
  double polar_eccentricity_squared                  = 0.0;
  double polar_flattening                            = 0.0;
  double r                                           = 0.0;
  double sine_of_point_1_latitude                    = 0.0;
  double surface_distance                            = 0.0;
  double term_1                                      = 0.0;
  double term_2                                      = 0.0;
  double term_3                                      = 0.0;
  double two_r                                       = 0.0;
  double x                                           = 0.0;

  // Many thanks to Peter Dana (pdana@mail.utexas.edu) for educating me
  // on the finer points of Geodesy, one of which was how to compute
  // "second eccentricity squared"

  polar_eccentricity_squared = ((m_EquatorialRadiusInMeters * m_EquatorialRadiusInMeters) - (m_PolarRadiusInMeters * m_PolarRadiusInMeters )) / ( m_PolarRadiusInMeters * m_PolarRadiusInMeters);

  //printf( "polar_eccentricity_squared is %.23lf\n", polar_eccentricity_squared );

  point_1_latitude_in_radians = CMath::ConvertDegreesToRadians(point_1.GetUpDownAngleInDegrees());
  direction_in_radians        = CMath::ConvertDegreesToRadians(direction);

  cosine_of_point_1_latitude = CMath::Cosine(point_1_latitude_in_radians);
  cosine_of_latitude_squared = cosine_of_point_1_latitude * cosine_of_point_1_latitude;

  cosine_of_direction_squared = CMath::Cosine(direction_in_radians) * CMath::Cosine(direction_in_radians);

  c = (m_EquatorialRadiusInMeters * m_EquatorialRadiusInMeters) / m_PolarRadiusInMeters;

  n = c / CMath::SquareRoot(1.0 + (polar_eccentricity_squared * cosine_of_latitude_squared));

  r = n / (1.0 + (polar_eccentricity_squared * cosine_of_latitude_squared * cosine_of_direction_squared));

  height_above_surface_of_point_1 = point_1.GetDistanceFromSurfaceInMeters();

  difference_in_height = height_above_surface_of_point_2 - height_above_surface_of_point_1;

  term_1 = (distance * distance) - (difference_in_height * difference_in_height);
  term_2 = 1.0 + (height_above_surface_of_point_1 / r);
  term_3 = 1.0 + (height_above_surface_of_point_2 / r);
  distance_adjusted_for_differences_in_height = CMath::SquareRoot(term_1 / (term_2 * term_3));

  // printf( "distance_adjusted_for_differences_in_height is %.11lf\n", distance_adjusted_for_differences_in_height );

  two_r = 2.0 * r;

  surface_distance = two_r * CMath::ArcSine(distance_adjusted_for_differences_in_height / two_r);

  // printf( "surface_distance is %.11lf\n", surface_distance );

  AddSurfaceDistanceAndDirectionToCoordinate(point_1,surface_distance,direction,point_2);
}

void CEarth::AddSurfaceDistanceAndDirectionToCoordinate(const CEarthCoordinate& point_1,
                                                        double distance, 
                                                        double direction, 
                                                        CPolarCoordinate& point_2)
{
  CPolarCoordinate polar_point_1;

  Convert(point_1, polar_point_1);

  AddSurfaceDistanceAndDirectionToCoordinate(polar_point_1,distance,direction,point_2);
}

void CEarth::AddSurfaceDistanceAndDirectionToCoordinate(const CEarthCoordinate& point_1, 
                                                        double distance, 
                                                        double direction, 
                                                        CEarthCoordinate& point_2)
{
  CPolarCoordinate polar_point_1;
  CPolarCoordinate polar_point_2;

  Convert(point_1,polar_point_1);

  AddSurfaceDistanceAndDirectionToCoordinate(polar_point_1,distance,direction,polar_point_2);

  Convert(polar_point_2,point_2);
}

void CEarth::AddSurfaceDistanceAndDirectionToCoordinate(const CPolarCoordinate& point_1, 
                                                        double distance, 
                                                        double direction, 
                                                        CEarthCoordinate& point_2)
{
  CPolarCoordinate polar_coordinate;

  AddSurfaceDistanceAndDirectionToCoordinate(point_1,distance,direction,polar_coordinate);

  Convert(polar_coordinate,point_2);
}

void CEarth::AddSurfaceDistanceAndDirectionToCoordinate(const CPolarCoordinate& point_1, 
                                                        double distance, 
                                                        double direction, 
                                                        CPolarCoordinate& point_2)
{
  // This is a translation of the Fortran routine DIRCT1 found in the
  // FORWRD3D program at:
  // ftp://ftp.ngs.noaa.gov/pub/pcsoft/for_inv.3d/source/forwrd3d.for

  double c                                          = 0.0;
  double c2a                                        = 0.0;
  double cosine_of_direction                        = 0.0;
  double cosine_of_y                                = 0.0;
  double cu                                         = 0.0;
  double cz                                         = 0.0;
  double d                                          = 0.0;
  double e                                          = 0.0;
  double direction_in_radians                       = 0.0;
  double eps                                        = 0.0;
  double heading_from_point_2_to_point_1_in_radians = 0.0;
  double point_1_latitude_in_radians                = 0.0;
  double point_1_longitude_in_radians               = 0.0;
  double point_2_latitude_in_radians                = 0.0;
  double point_2_longitude_in_radians               = 0.0;
  double r                                          = 0.0;
  double sa                                         = 0.0;
  double sine_of_direction                          = 0.0;
  double sine_of_y                                  = 0.0;
  double su                                         = 0.0;
  double tangent_u                                  = 0.0;
  double term_1                                     = 0.0;
  double term_2                                     = 0.0;
  double term_3                                     = 0.0;
  double x                                          = 0.0;
  double y                                          = 0.0;

  direction_in_radians = CMath::ConvertDegreesToRadians(direction);

  eps = 0.000000000000005;

  r = 1.0 - m_Flattening;

  point_1_latitude_in_radians  = CMath::ConvertDegreesToRadians(point_1.GetUpDownAngleInDegrees());
  point_1_longitude_in_radians = CMath::ConvertDegreesToRadians(point_1.GetLeftRightAngleInDegrees());

  tangent_u = (r * CMath::Sine(point_1_latitude_in_radians)) / CMath::Cosine(point_1_latitude_in_radians);

  sine_of_direction = CMath::Sine(direction_in_radians);

  cosine_of_direction = CMath::Cosine(direction_in_radians);

  heading_from_point_2_to_point_1_in_radians = 0.0;

  if (cosine_of_direction != 0.0)
  {
    heading_from_point_2_to_point_1_in_radians = CMath::ArcTangentOfYOverX(tangent_u, cosine_of_direction) * 2.0;
  }

  cu = 1.0 / CMath::SquareRoot((tangent_u * tangent_u) + 1.0);
  su = tangent_u * cu;
  sa = cu * sine_of_direction;
  c2a = ((-sa) * sa) + 1.0;
  x = CMath::SquareRoot((((1.0 / r / r) - 1.0) * c2a) + 1.0) + 1.0;
  x = (x - 2.0) / x;
  c = 1.0 - x;
  c = (((x * x) / 4.0) + 1.0) / c;
  d = ((0.375 * (x * x)) -1.0) * x;

  tangent_u = distance / r / m_EquatorialRadiusInMeters / c;

  y = tangent_u;

  bool exit_loop = false;

  while(exit_loop != true)
  {
    sine_of_y = CMath::Sine(y);
    cosine_of_y = CMath::Cosine(y);
    cz = CMath::Cosine(heading_from_point_2_to_point_1_in_radians + y);
    e = (cz * cz * 2.0) - 1.0;
    c = y;
    x = e * cosine_of_y;
    y = (e + e) - 1.0;

    term_1 = (sine_of_y * sine_of_y * 4.0) - 3.0;
    term_2 = ((term_1 * y * cz * d) / 6.0) + x;
    term_3 = ((term_2 * d) / 4.0) - cz;
    y = (term_3 * sine_of_y * d) + tangent_u;

    if (CMath::AbsoluteValue(y - c) > eps)
    {
      exit_loop = false;
    }
    else
    {
      exit_loop = true;
    }
  }

  heading_from_point_2_to_point_1_in_radians = (cu * cosine_of_y * cosine_of_direction) - (su * sine_of_y);

  c = r * CMath::SquareRoot((sa * sa) + (heading_from_point_2_to_point_1_in_radians * heading_from_point_2_to_point_1_in_radians));
  d = (su * cosine_of_y) + (cu * sine_of_y * cosine_of_direction);

  point_2_latitude_in_radians = CMath::ArcTangentOfYOverX(d,c);

  c = (cu * cosine_of_y) - (su * sine_of_y * cosine_of_direction);
  x = CMath::ArcTangentOfYOverX(sine_of_y * sine_of_direction,c);
  c = (((((-3.0 * c2a) + 4.0) * m_Flattening) + 4.0) * c2a * m_Flattening) / 16.0;
  d = ((((e * cosine_of_y * c) + cz) * sine_of_y * c) + y) * sa;

  point_2_longitude_in_radians = (point_1_longitude_in_radians + x) - ((1.0 - c) * d * m_Flattening);

  heading_from_point_2_to_point_1_in_radians = CMath::ArcTangentOfYOverX(sa,heading_from_point_2_to_point_1_in_radians) + CMath::Pi();

  point_2.SetUpDownAngleInDegrees(CMath::ConvertRadiansToDegrees(point_2_latitude_in_radians));
  point_2.SetLeftRightAngleInDegrees(CMath::ConvertRadiansToDegrees(point_2_longitude_in_radians));
}

void CEarth::Convert(const CEarthCoordinate& cartesian_coordinate,
                     CPolarCoordinate& polar_coordinate) const
{
  // convert from cartesian to polar

  double equatorial_radius_times_eccentricity_squared = 0.0;

  equatorial_radius_times_eccentricity_squared = m_EquatorialRadiusInMeters * m_EccentricitySquared;

  double p = 0.0;

  p = CMath::SquareRoot((cartesian_coordinate.GetXCoordinateInMeters() * cartesian_coordinate.GetXCoordinateInMeters()) + 
                        (cartesian_coordinate.GetYCoordinateInMeters() * cartesian_coordinate.GetYCoordinateInMeters()));

  double temp_latitude = 0.0;
  double z_coordinate  = cartesian_coordinate.GetZCoordinateInMeters(); // for convienance
  double one_minus_eccentricity_squared = 1.0 - m_EccentricitySquared;

  temp_latitude = z_coordinate / p / one_minus_eccentricity_squared;

  double old_value  = 0.0;
  double temp_value = 0.0;
  double part_a     = 0.0;
  double part_b     = 0.0;
  double part_c     = 0.0;

  unsigned long loop_index              = 0;
  unsigned long maximum_number_of_tries = 1024;

  bool convergence_was_acheived = false;

  while(convergence_was_acheived != true && loop_index < maximum_number_of_tries)
  {
    old_value = temp_latitude;

    part_a = one_minus_eccentricity_squared * temp_latitude * temp_latitude;
    part_b = equatorial_radius_times_eccentricity_squared / CMath::SquareRoot(1.0 + part_a);
    part_c = p - part_b;

    temp_latitude = z_coordinate / part_c;

    loop_index++;

    if (CMath::AbsoluteValue(temp_latitude - old_value) > 0.000000000000000000001)
    {
      // Oh well, try again...
    }
    else
    {
      // YIPEE!! We've reached convergence!
      convergence_was_acheived = true;
    }
  }

  if (convergence_was_acheived == true)
  {
    double latitude_angle_in_radians = 0.0;

    // Save the UpDown angle in degrees
      
    latitude_angle_in_radians = CMath::ArcTangent(temp_latitude);

    polar_coordinate.SetUpDownAngleInDegrees(CMath::ConvertRadiansToDegrees(latitude_angle_in_radians)); // Latitude

    double sine_of_latitude_in_radians   = 0.0;
    double cosine_of_latitude_in_radians = 0.0;

    sine_of_latitude_in_radians   = CMath::Sine(latitude_angle_in_radians);
    cosine_of_latitude_in_radians = CMath::Cosine(latitude_angle_in_radians);

    double longitude_in_radians = 0.0;

    longitude_in_radians = CMath::ArcTangentOfYOverX(cartesian_coordinate.GetYCoordinateInMeters(),cartesian_coordinate.GetXCoordinateInMeters());

    polar_coordinate.SetLeftRightAngleInDegrees(CMath::ConvertRadiansToDegrees(longitude_in_radians)); // Longitude

    double w = 0.0;

    w = CMath::SquareRoot(1.0 - (m_EccentricitySquared * sine_of_latitude_in_radians * sine_of_latitude_in_radians));

    double distance_from_center_to_surface_of_the_ellipsoid = 0.0;

    distance_from_center_to_surface_of_the_ellipsoid = m_EquatorialRadiusInMeters / w;

    double distance_from_surface = 0.0;

    if (CMath::AbsoluteValue(latitude_angle_in_radians) < 0.7854)
    {
      //printf( "fabs( %14.10lf ) < 0.7854\n", latitude_angle_in_radians );
      distance_from_surface = (p / cosine_of_latitude_in_radians) - distance_from_center_to_surface_of_the_ellipsoid;
    }
    else
    {
      //printf( "fabs( %14.10lf ) >= 0.7854\n", latitude_angle_in_radians );
      distance_from_surface = (z_coordinate / sine_of_latitude_in_radians) 
                              - distance_from_center_to_surface_of_the_ellipsoid 
                              + (m_EccentricitySquared * distance_from_center_to_surface_of_the_ellipsoid);
    }

    // printf( "CEarth::Convert() : First method produced a length of %14.10lf\n", distance_from_surface );

    polar_coordinate.SetDistanceFromSurfaceInMeters(distance_from_surface);
  }
  else
  {
    // Oh well, we gave it a shot..
    polar_coordinate.Set( 0.0, 0.0, 0.0 );
  }
}

void CEarth::Convert(const CPolarCoordinate& polar_coordinate,
                     CEarthCoordinate& cartesian_coordinate) const
{
  // convert from polar to cartesian

  double up_down_radians    = 0.0; // latitude
  double left_right_radians = 0.0; // longitude angle

  up_down_radians    = CMath::ConvertDegreesToRadians(polar_coordinate.GetUpDownAngleInDegrees()); 
  left_right_radians = CMath::ConvertDegreesToRadians(polar_coordinate.GetLeftRightAngleInDegrees());

  double sine_of_up_down_radians      = 0.0;
  double cosine_of_left_right_radians = 0.0; // cosine_of_longitude
  double cosine_of_up_down_radians    = 0.0; // cosine_of_latitude

  sine_of_up_down_radians      = CMath::Sine(up_down_radians);
  cosine_of_left_right_radians = CMath::Cosine(left_right_radians);
  cosine_of_up_down_radians    = CMath::Cosine(up_down_radians);

  // Now we need to calculate the distance from the center of the ellipsoid to the surface of the ellipsoid
  double w = 0.0;

  w = CMath::SquareRoot(1.0 - (m_EccentricitySquared * sine_of_up_down_radians * sine_of_up_down_radians));

  double distance_from_center_to_surface_of_the_ellipsoid = 0.0;

  distance_from_center_to_surface_of_the_ellipsoid = m_EquatorialRadiusInMeters / w;

  // printf( "en = %.25lf\n", distance_from_center_to_surface_of_the_ellipsoid );

  double coordinate = 0.0;

  coordinate = (distance_from_center_to_surface_of_the_ellipsoid + polar_coordinate.GetDistanceFromSurfaceInMeters()) 
               * cosine_of_up_down_radians * cosine_of_left_right_radians;
  
  cartesian_coordinate.SetXCoordinateInMeters(coordinate);

  coordinate = (distance_from_center_to_surface_of_the_ellipsoid + polar_coordinate.GetDistanceFromSurfaceInMeters()) 
               * cosine_of_up_down_radians * CMath::Sine(left_right_radians);
   
  cartesian_coordinate.SetYCoordinateInMeters(coordinate);

  coordinate = (distance_from_center_to_surface_of_the_ellipsoid * (1.0 - m_EccentricitySquared) 
               + polar_coordinate.GetDistanceFromSurfaceInMeters()) * sine_of_up_down_radians;
   
  cartesian_coordinate.SetZCoordinateInMeters(coordinate);
}

double CEarth::GetDistanceToHorizon(const CEarthCoordinate& point_1) const
{
  CPolarCoordinate polar_coordinate;

  Convert(point_1,polar_coordinate);

  return(GetDistanceToHorizon(polar_coordinate));
}

double CEarth::GetDistanceToHorizon(const CPolarCoordinate& point_1) const
{
  double distance_to_horizon = 0.0;

  // d = ::sqrt( feet ) *  1.144 for nmi
  // optical horizon is 1.317 * sqrt( h );
  // d= ::sqrt( 17 * height_in_meters ); d is in meters

  distance_to_horizon = CMath::SquareRoot(17.0 * point_1.GetDistanceFromSurfaceInMeters());

  return(distance_to_horizon);
}

double CEarth::GetEquatorialRadiusInMeters(void) const
{
  return(m_EquatorialRadiusInMeters);
}

double CEarth::GetPolarRadiusInMeters(void) const
{
  return(m_PolarRadiusInMeters);
}

double CEarth::GetLineOfSightDistanceFromCourse(const CEarthCoordinate& current_location, 
                                                const CEarthCoordinate& point_a, 
                                                const CEarthCoordinate& point_b) const
{
  // This function tells you how far off course you are from a straight line between
  // point_a and point_b.

  /*
  ** References:
  ** I got the formula from:
  ** Engineering Mathematics Handbook
  ** Jan J. Tuma, Ph.D.
  ** McGraw-Hill Book Company
  ** 1970
  ** Library of Congress Catalog Number 78-101174
  ** page 19, (a) Oblique triangle
  **
  ** Teach Yourself Trigonometry
  ** P. Abbott, B.A.
  ** English Universities Press Ltd.
  ** 102 Newgate Street
  ** London, E.C.I
  ** Originally published 1940
  ** I used the 1964 printing.
  ** Page 22, Figure 12 calls this "the altitude from the vertex A"
  */ 

  double distance_from_current_location_to_point_a = 0.0;
  double distance_from_current_location_to_point_b = 0.0;
  double distance_from_point_a_to_point_b          = 0.0;

  distance_from_current_location_to_point_a = GetLineOfSightDistance(current_location, point_a);
  distance_from_current_location_to_point_b = GetLineOfSightDistance(current_location, point_b);
  distance_from_point_a_to_point_b          = GetLineOfSightDistance(point_a,          point_b);

  double p = 0.0;

  p  = distance_from_current_location_to_point_a;
  p += distance_from_current_location_to_point_b;
  p += distance_from_point_a_to_point_b;
  p /= 2.0;

  double temp_double = 0.0;

  temp_double  = p;
  temp_double *= (double) (p - distance_from_current_location_to_point_a);
  temp_double *= (double) (p - distance_from_current_location_to_point_b);
  temp_double *= (double) (p - distance_from_point_a_to_point_b);

  double area = 0.0;

  area = CMath::SquareRoot(temp_double);

  double distance_from_course = 0.0;

  // The altitude from the vertex A is two times the area of the triangle divided by the baseline

  distance_from_course = (2.0 * area) / distance_from_point_a_to_point_b;

  return(distance_from_course);
}

double CEarth::GetLineOfSightDistance(const CEarthCoordinate& point_1, 
                                      const CEarthCoordinate& point_2) const
{
  // This function implements the Pythagoras method of computing the distance
  // between two points.
  // This is a line-of-sight algorithm. It does not take into acccount the 
  // curvature of the Earth. It is not a distance on the surface algorithm.
  // If you had a laser and connected the two points, this algorithm tells
  // you how long the laser beam is. 

  double distance     = 0.0;
  double x_coordinate = 0.0;
  double y_coordinate = 0.0;
  double z_coordinate = 0.0;

  x_coordinate = point_1.GetXCoordinateInMeters() - point_2.GetXCoordinateInMeters();
  y_coordinate = point_1.GetYCoordinateInMeters() - point_2.GetYCoordinateInMeters();
  z_coordinate = point_1.GetZCoordinateInMeters() - point_2.GetZCoordinateInMeters();

  // Square the coordinates
  x_coordinate *= x_coordinate;
  y_coordinate *= y_coordinate;
  z_coordinate *= z_coordinate;

  distance = CMath::SquareRoot(x_coordinate + y_coordinate + z_coordinate);

  return(distance);
}

double CEarth::GetLineOfSightDistance(const CEarthCoordinate& point_1,
                                      const CPolarCoordinate& point_2) const
{
  CEarthCoordinate earth_center_earth_fixed_point_2;

  Convert(point_2,earth_center_earth_fixed_point_2);

  return(GetLineOfSightDistance(point_1,earth_center_earth_fixed_point_2));
}

double CEarth::GetLineOfSightDistance(const CPolarCoordinate& point_1,
                                      const CEarthCoordinate& point_2) const
{
  CEarthCoordinate earth_center_earth_fixed_point_1;

  Convert(point_1,earth_center_earth_fixed_point_1);

  return(GetLineOfSightDistance(earth_center_earth_fixed_point_1,point_2));
}

double CEarth::GetLineOfSightDistance(const CPolarCoordinate& point_1,const CPolarCoordinate& point_2) const
{
  CEarthCoordinate earth_center_earth_fixed_point_1;
  CEarthCoordinate earth_center_earth_fixed_point_2;

  Convert(point_1,earth_center_earth_fixed_point_1);
  Convert(point_2,earth_center_earth_fixed_point_2);

  return(GetLineOfSightDistance(earth_center_earth_fixed_point_1,earth_center_earth_fixed_point_2));
}

double CEarth::GetSurfaceDistance(const CEarthCoordinate& point_1, 
                                  const CEarthCoordinate& point_2, 
                                  double * heading_from_point_1_to_point_2_p, 
                                  double * heading_from_point_2_to_point_1_p) const
{
  CPolarCoordinate polar_point_1;
  CPolarCoordinate polar_point_2;

  Convert(point_1,polar_point_1);
  Convert(point_2,polar_point_2);

  return(GetSurfaceDistance(polar_point_1,
                            polar_point_2,
                            heading_from_point_1_to_point_2_p, 
                            heading_from_point_2_to_point_1_p));
}

double CEarth::GetSurfaceDistance(const CEarthCoordinate& point_1, 
                                  const CPolarCoordinate& point_2, 
                                  double * heading_from_point_1_to_point_2_p, 
                                  double * heading_from_point_2_to_point_1_p) const
{
  CPolarCoordinate polar_point_1;

  Convert(point_1,polar_point_1);

  return(GetSurfaceDistance(polar_point_1,
                            point_2,
                            heading_from_point_1_to_point_2_p, 
                            heading_from_point_2_to_point_1_p));
}

double CEarth::GetSurfaceDistance(const CPolarCoordinate& point_1, 
                                  const CEarthCoordinate& point_2, 
                                  double * heading_from_point_1_to_point_2_p, 
                                  double * heading_from_point_2_to_point_1_p) const
{
  CPolarCoordinate polar_point_2;

  Convert(point_2,polar_point_2);

  return(GetSurfaceDistance(point_1, 
                            polar_point_2, 
                            heading_from_point_1_to_point_2_p, 
                            heading_from_point_2_to_point_1_p));
}

double CEarth::GetSurfaceDistance(const CPolarCoordinate& point_1, 
                                  const CPolarCoordinate& point_2, 
                                  double * heading_from_point_1_to_point_2_p, 
                                  double * heading_from_point_2_to_point_1_p) const
{
  // This is a translation of the Fortran routine INVER1 found in the
  // INVERS3D program at:
  // ftp://ftp.ngs.noaa.gov/pub/pcsoft/for_inv.3d/source/invers3d.for

  // The ton of variables used...

  double c           = 0.0;
  double c_value_1   = 0.0;
  double c_value_2   = 0.0;
  double c2a         = 0.0;
  double cosine_of_x = 0.0;
  double cy          = 0.0;
  double cz          = 0.0;
  double d           = 0.0;
  double e           = 0.0;
  double r_value     = 0.0;
  double s           = 0.0;
  double s_value_1   = 0.0;
  double sa          = 0.0;
  double sine_of_x   = 0.0;
  double sy          = 0.0;
  double tangent_1   = 0.0;
  double tangent_2   = 0.0;
  double x           = 0.0;
  double y           = 0.0;

  int loop_count = 0;

  double heading_from_point_1_to_point_2 = 0.0;
  double heading_from_point_2_to_point_1 = 0.0;

  // UpDown    == Latitude
  // LeftRight == Longitude

  double point_1_latitude_in_radians  = CMath::ConvertDegreesToRadians(point_1.GetUpDownAngleInDegrees());
  double point_1_longitude_in_radians = CMath::ConvertDegreesToRadians(point_1.GetLeftRightAngleInDegrees());
  double point_2_latitude_in_radians  = CMath::ConvertDegreesToRadians(point_2.GetUpDownAngleInDegrees());
  double point_2_longitude_in_radians = CMath::ConvertDegreesToRadians(point_2.GetLeftRightAngleInDegrees());

  r_value = 1.0 - m_Flattening;
  tangent_1 = (r_value * CMath::Sine(point_1_latitude_in_radians)) / CMath::Cosine(point_1_latitude_in_radians);
  tangent_2 = (r_value * CMath::Sine(point_2_latitude_in_radians)) / CMath::Cosine(point_2_latitude_in_radians);
  c_value_1 = 1.0 / CMath::SquareRoot((tangent_1 * tangent_1) + 1.0);
  s_value_1 = c_value_1 * tangent_1;
  c_value_2 = 1.0 / CMath::SquareRoot((tangent_2 * tangent_2) + 1.0);
  s = c_value_1 * c_value_2;

  heading_from_point_2_to_point_1 = s * tangent_2; // backward_azimuth
  heading_from_point_1_to_point_2 = heading_from_point_2_to_point_1 * tangent_1;

  x = point_2_longitude_in_radians - point_1_longitude_in_radians;

  bool exit_loop = false;

  while(exit_loop != true)
  {
    sine_of_x = CMath::Sine(x);
    cosine_of_x = CMath::Cosine(x);
    tangent_1 = c_value_2 * sine_of_x;
    tangent_2 = heading_from_point_2_to_point_1 - (s_value_1 * c_value_2 * cosine_of_x);
    sy = CMath::SquareRoot((tangent_1 * tangent_1) + (tangent_2 * tangent_2));
    cy = (s * cosine_of_x) + heading_from_point_1_to_point_2;
    y = CMath::ArcTangentOfYOverX(sy,cy);

    // Thanks to John Werner (werner@tij.wb.xerox.com) for
    // finding a bug where sy could be zero. Here's his fix:

    if ((s * sine_of_x ) == 0.0 && (sy == 0.0))
    {
      sa = 1.0;
    }
    else
    {
      sa = (s * sine_of_x) / sy;
    }

    c2a = ((-sa) * sa) + 1.0;
    cz = heading_from_point_1_to_point_2 + heading_from_point_1_to_point_2;

    if (c2a > 0.0)
    {
      cz = ((-cz) / c2a) + cy;
    }

    e = (cz * cz * 2.0) - 1.0;
    c = (((((-3.0 * c2a) + 4.0) * m_Flattening) + 4.0) * c2a * m_Flattening) / 16.0;
    d = x;
    x = ((((e * cy * c) + cz) * sy * c) + y) * sa;
    x = ((1.0 - c) * x * m_Flattening) + point_2_longitude_in_radians - point_1_longitude_in_radians;

    if (CMath::AbsoluteValue(d - x) > 0.00000000000000000000005)
    {
      exit_loop = false;
    }
    else
    {
      exit_loop = true;
    }
  }

  heading_from_point_1_to_point_2 = CMath::ArcTangentOfYOverX(tangent_1,tangent_2);

  double temp_degrees         = 0.0;
  double temp_minutes         = 0.0;
  double temp_seconds         = 0.0;
  double temp_decimal_degrees = 0.0;

  temp_decimal_degrees = CMath::ConvertRadiansToDegrees(heading_from_point_1_to_point_2);

  if (temp_decimal_degrees < 0.0)
  {
    temp_decimal_degrees += 360.0;
  }

  if (heading_from_point_1_to_point_2_p != NULL)
  {
    // The user passed us a pointer, don't trust it.
    // If you are using Visual C++ on Windows NT, the following
    // try/catch block will ensure you won't blow up when random
    // pointers are passed to you. If you are on a legacy operating
    // system like Unix, you are screwed.

    try
    {
      *heading_from_point_1_to_point_2_p = temp_decimal_degrees;
    }
    catch( ... )
    {
      // Do Nothing
    }
  }

  heading_from_point_2_to_point_1 = CMath::ArcTangentOfYOverX(c_value_1 * sine_of_x,
    ((heading_from_point_2_to_point_1 * cosine_of_x) - (s_value_1 * c_value_2))) + CMath::Pi();

  temp_decimal_degrees = CMath::ConvertRadiansToDegrees(heading_from_point_2_to_point_1);

  if (temp_decimal_degrees < 0)
  {
    temp_decimal_degrees += 360.0;
  }

  if (heading_from_point_2_to_point_1_p != NULL)
  {
    // The user passed us a pointer, don't trust it.
    // If you are using Visual C++ on Windows NT, the following
    // try/catch block will ensure you won't blow up when random
    // pointers are passed to you. If you are on a legacy operating
    // system like Unix, you are screwed.

    try
    {
      *heading_from_point_2_to_point_1_p = temp_decimal_degrees;
    }
    catch( ... )
    {
      // Do Nothing
    }
  }

  x = CMath::SquareRoot((((1.0 / r_value / r_value) - 1) * c2a) + 1.0) + 1.0;
  x = (x - 2.0) / x;
  c = 1.0 - x;
  c = (((x * x) / 4.0) + 1.0) / c;
  d = ((0.375 * (x * x)) - 1.0) * x;

  // 1998-09-01
  // Thanks go to Gerard Murphy (bjmillar@dera.gov.uk) for finding a typo here.

  x = e * cy;

  s = (1.0 - e) - e;

  double term_1 = 0.0;
  double term_2 = 0.0;
  double term_3 = 0.0;
  double term_4 = 0.0;
  double term_5 = 0.0;

  term_1 = (sy * sy * 4.0) - 3.0;
  term_2 = ((s * cz * d) / 6.0) - x;
  term_3 = term_1 * term_2;
  term_4 = ((term_3 * d) / 4.0) + cz;
  term_5 = (term_4 * sy * d) + y;

  s = term_5 * c * m_EquatorialRadiusInMeters * r_value;

  return( s );
}

void CEarth::m_ComputeEccentricitySquared(void)
{
  if (m_Flattening == 0.0)
  {
    m_EccentricitySquared = 0.0;
    return;
  }

  m_EccentricitySquared = (2.0 * m_Flattening) - (m_Flattening * m_Flattening);
  //printf( "Eccentricity Squared = %.23lf\n", m_EccentricitySquared );
}

void CEarth::m_ComputeFlattening(void)
{
  if (m_EquatorialRadiusInMeters == 0.0 || m_PolarRadiusInMeters == 0.0)
  {
    return;
  }

  m_Flattening = CMath::AbsoluteValue(m_EquatorialRadiusInMeters - m_PolarRadiusInMeters) / m_EquatorialRadiusInMeters;
  //printf( "Flattening = %.23lf\n", m_Flattening );
}

void CEarth::m_Initialize(void)
{
  m_EllipsoidID              = 0;
  m_PolarRadiusInMeters      = 0.0;
  m_EquatorialRadiusInMeters = 0.0;
  m_Flattening               = 0.0;
  m_EccentricitySquared      = 0.0;
}

void CEarth::SetEllipsoid(int ellipsoid_identifier)
{
  m_EllipsoidID = ellipsoid_identifier;

  switch(ellipsoid_identifier)
  {
    case Perfect_Sphere:
         m_EquatorialRadiusInMeters = 6378137.0;
         m_PolarRadiusInMeters      = 6378137.0;
         break;

    case Airy:
         m_EquatorialRadiusInMeters = 6377563.396;
         m_PolarRadiusInMeters      = 6356256.909237;
         break;

    case Austrailian_National:
         m_EquatorialRadiusInMeters = 6378160.0;
         m_PolarRadiusInMeters      = 6356774.719195;
         break;

    case Bessell_1841:
         m_EquatorialRadiusInMeters = 6377397.155;
         m_PolarRadiusInMeters      = 6356078.962818;
         break;

    case Bessel_1841_Nambia:
         m_EquatorialRadiusInMeters = 6377483.865;
         m_PolarRadiusInMeters      = 6356165.382966;
         break;

    case Clarke_1866:
         m_EquatorialRadiusInMeters = 6378206.4;
         m_PolarRadiusInMeters      = 6356583.799999;
         break;

    case Clarke_1880:
         m_EquatorialRadiusInMeters = 6378249.145;
         m_PolarRadiusInMeters      = 6356514.86955;
         break;

    case Everest:
         m_EquatorialRadiusInMeters = 6377276.345;
         m_PolarRadiusInMeters      = 6356075.41314;
         break;

    case Fischer_1960_Mercury:
         m_EquatorialRadiusInMeters = 6378166.0;
         m_PolarRadiusInMeters      = 6356784.283607;
         break;

    case Fischer_1968:
         m_EquatorialRadiusInMeters = 6378150.0;
         m_PolarRadiusInMeters      = 6356768.337244;
         break;

    case GRS_1967:
         m_EquatorialRadiusInMeters = 6378160.0;
         m_PolarRadiusInMeters      = 6356774.516091;
         break;

    case GRS_1980:
         m_EquatorialRadiusInMeters = 6378137.0;
         m_PolarRadiusInMeters      = 6356752.31414;
         break;

    case Helmert_1906:
         m_EquatorialRadiusInMeters = 6378200.0;
         m_PolarRadiusInMeters      = 6356818.169628;
         break;

    case Hough:
         m_EquatorialRadiusInMeters = 6378270.0;
         m_PolarRadiusInMeters      = 6356794.343434;
         break;

    case International:
         m_EquatorialRadiusInMeters = 6378388.0;
         m_PolarRadiusInMeters      = 6356911.946128;
         break;

    case Krassovsky:
         m_EquatorialRadiusInMeters = 6378245.0;
         m_PolarRadiusInMeters      = 6356863.018773;
         break;

    case Modified_Airy:
         m_EquatorialRadiusInMeters = 6377340.189;
         m_PolarRadiusInMeters      = 6356034.447939;
         break;

    case Modified_Everest:
         m_EquatorialRadiusInMeters = 6377304.063;
         m_PolarRadiusInMeters      = 6356103.038993;
         break;

    case Modified_Fischer_1960:
         m_EquatorialRadiusInMeters = 6378155.0;
         m_PolarRadiusInMeters      = 6356773.320483;
         break;

    case South_American_1969:
         m_EquatorialRadiusInMeters = 6378160.0;
         m_PolarRadiusInMeters      = 6356774.719195;
         break;

    case Topex_Poseidon_Pathfinder_ITRF: // Source is http://neptune.gsfc.nasa.gov/~krachlin/corr/refframe.html
         m_EquatorialRadiusInMeters = 6378136.3;
         m_PolarRadiusInMeters      = 6356751.6005629376;
         break;

    case WGS_60:
         m_EquatorialRadiusInMeters = 6378165.0;
         m_PolarRadiusInMeters      = 6356783.286959;
         break;

    case WGS_66:
         m_EquatorialRadiusInMeters = 6378145.0;
         m_PolarRadiusInMeters      = 6356759.769489;
         break;

    case WGS_72:
         m_EquatorialRadiusInMeters = 6378135.0;
         m_PolarRadiusInMeters      = 6356750.520016;
         break;

    case WGS_84:
         // Computed polar radius from the flattening value specified at
         // http://acro.harvard.edu/SSA/BGA/wg84figs.html
         // because it had the most digits after the decimal point.

         m_EquatorialRadiusInMeters = 6378137.0;
         m_PolarRadiusInMeters      = 6356752.3142451793;
         break;

    case Unknown:
    default:

         m_EllipsoidID = Unknown;

         m_Initialize();
         return;
  }

  m_ComputeFlattening();
  m_ComputeEccentricitySquared();
}

void CEarth::SetEllipsoidByRadii(double equatorial_radius,double polar_radius)
{
  m_EquatorialRadiusInMeters = equatorial_radius;
  m_PolarRadiusInMeters      = polar_radius;
  m_EllipsoidID              = Custom;

  m_ComputeFlattening();
  m_ComputeEccentricitySquared();
}

void CEarth::SetEllipsoidByEquatorialRadiusAndFlattening(double equatorial_radius,double flattening)
{
  m_EquatorialRadiusInMeters = equatorial_radius;
  m_Flattening               = flattening;
  m_EllipsoidID              = Custom;

  // We must compute the polar radius

  double temp_double = m_Flattening * m_EquatorialRadiusInMeters;

  m_PolarRadiusInMeters = m_EquatorialRadiusInMeters - temp_double;

  m_ComputeEccentricitySquared();
}

// WGS84 to Gauss-Krueger converter (recoding of Pascal sources from Ottmar Labonde)

void WGS84toDHDN::HelmertTransformation(double x,double y,double z,double& xo,double& yo,double& zo)
{
  xo=dx+(sc*(1*x+rotz*y-roty*z));
  yo=dy+(sc*(-rotz*x+1*y+rotx*z));
  zo=dz+(sc*(roty*x-rotx*y+1*z));
}

void WGS84toDHDN::BesselBLnachGaussKrueger(double b,double ll,double& Re,double& Ho)
{
  double l;
  double l0;
  double bg;
  double lng;
  double Ng;
  double k;
  double t;
  double eq;
  double Vq;
  double v;
  double nk;
  double X;
  double gg;
  double SS;
  double Y;
  double kk;
  double Pii;
  double RVV;

  bg=180*b/Pi;
  lng=180*ll/Pi;
  l0=3*round((180*ll/Pi)/3);
  l0=Pi*l0/180;
  l=ll-l0;
  k=cos(b);
  t=sin(b)/k;
  eq=(abes*abes-bbes*bbes)/(bbes*bbes);
  Vq=1+eq*k*k;
  v=sqrt(Vq);
  Ng=abes*abes/(bbes*v);
  nk=(abes-bbes)/(abes+bbes);
  X=((Ng*t*k*k*l*l)/2)+((Ng*t*(9*Vq-t*t-4)*k*k*k*k*l*l*l*l)/24);

  gg=b+(((-3*nk/2)+(9*nk*nk*nk/16))*sin(2*b)+15*nk*nk*sin(4*b)/16-35*nk*nk*nk*sin(6*b)/48);
  SS=gg*180*cbes/Pi;
  Ho=(SS+X);
  Y=Ng*k*l+Ng*(Vq-t*t)*k*k*k*l*l*l/6+Ng*(5-18*t*t+t*t*t*t)*k*k*k*k*k*l*l*l*l*l/120;
  kk=500000;
  Pii=Pi;
  RVV=round((180*ll/Pii)/3);
  Re=RVV*1000000+kk+Y;
}

void WGS84toDHDN::BLRauenberg (double x,double y,double z,double& b,double& l,double& h)
{
  double f;
  double f1;
  double f2;
  double ft;
  double p;

  f=Pi*50/180;
  p=Z/sqrt(x*x+y*y);
  
  do
  {
    f1=neuF(f,x,y,p);
    f2=f;
    f=f1;
    ft=180*f1/Pi;
  }
  while(!(fabs(f2-f1)<10E-10));
  
  b=f;
  l=atan(y/x);
  h=sqrt(x*x+y*y)/cos(f1)-a/sqrt(1-eq*sin(f1)*sin(f1));
}

double WGS84toDHDN::neuF(double f,double x,double y,double p)
{
  double zw;
  double nnq;

  zw=a/sqrt(1-eq*sin(f)*sin(f));
  nnq=1-eq*zw/(sqrt(x*x+y*y)/cos(f));
  return(atan(p/nnq));
}

double WGS84toDHDN::round(double src)
{
  double theInteger;
  double theFraction;
  double criterion = 0.5;

  theFraction = modf(src,&theInteger);

  if (!(theFraction < criterion))
  {
    theInteger += 1; 
  } 

  return theInteger;
}

WGS84toDHDN::WGS84toDHDN()
{
  awgs = 6378137.0;        // WGS84 Semi-Major Axis = Equatorial Radius in meters
  bwgs = 6356752.314;      // WGS84 Semi-Minor Axis = Polar Radius in meters 
  abes = 6377397.155;      // Bessel Semi-Major Axis = Equatorial Radius in meters
  bbes = 6356078.962;      // Bessel Semi-Minor Axis = Polar Radius in meters
  cbes = 111120.6196;      // Bessel ???
  dx   = -585.7;           // Translation Parameter 1
  dy   = -87.0;            // Translation Parameter 2 
  dz   = -409.2;           // Translation Parameter 3
  rotx = 2.540423689E-6;   // Rotation Parameter 1
  roty = 7.514612057E-7;   // Rotation Parameter 2
  rotz = -1.368144208E-5;  // Rotation Parameter 3
  sc   = 1/0.99999122;     // Scaling Factor 
  Pi   = 3.1415926535897932384626433832795028841971693993751058209749445923078164;
  l1 = 0.0;
  b1 = 0.0;
  l2 = 0.0;
  b2 = 0.0;
  h1 = 0.0;
  h2 = 0.0;
  R = 0.0;
  H = 0.0;
  a = 0.0;
  b = 0.0;
  eq = 0.0;
  N = 0.0;
  Xq = 0.0;
  Yq = 0.0;
  Zq = 0.0;
  X = 0.0;
  Y = 0.0;
  Z = 0.0;
}

WGS84toDHDN::~WGS84toDHDN()
{
  l1 = 0.0;
  b1 = 0.0;
  l2 = 0.0;
  b2 = 0.0;
  h1 = 0.0;
  h2 = 0.0;
  R = 0.0;
  H = 0.0;
  a = 0.0;
  b = 0.0;
  eq = 0.0;
  N = 0.0;
  Xq = 0.0;
  Yq = 0.0;
  Zq = 0.0;
  X = 0.0;
  Y = 0.0;
  Z = 0.0;
}

// Precision is +- 1 meter for Germany
void WGS84toDHDN::Convert(double latIn,double lonIn,double heightIn,double& ROut,double& HOut)
{
  b1 = latIn;
  l1 = lonIn;
  h1 = heightIn;
    
  l1=Pi*l1/180;
  b1=Pi*b1/180;
  a=awgs; 
  b=bwgs;
  eq=(a*a-b*b)/(a*a);
  N=a/sqrt(1-eq*sin(b1)*sin(b1));
  Xq=(N+h1)*cos(b1)*cos(l1);
  Yq=(N+h1)*cos(b1)*sin(l1);
  Zq=((1-eq)*N+h1)*sin(b1);
  HelmertTransformation(Xq,Yq,Zq,X,Y,Z);
  a=abes;
  b=bbes;
  eq=(a*a-b*b)/(a*a);
  BLRauenberg(X,Y,Z,b2,l2,h2);
  BesselBLnachGaussKrueger(b2,l2,R,H);

  ROut = R;
  HOut = H;
}


